; -*- Scheme -*-
;; Modified 12/95 Matthew Flatt for DrScheme2
;;  Stripped out stuff not needed for syntax-checking 
;;
;; Modified 7/95 Dan Grossman for DrScheme
;;all emacs related functions replaced.
;;aries:transmit-to-emacs replaced by aries:transmit-to-mred
;;no other specifications changed EXCEPT NO DOWNCASING OF SYMBOLS!!!
;;   AND COUNT TABS AS ONE COLUMN NUMBER: DON'T FORGET TO KEEP THIS CHANGE!!!
;; Aries
;; (Aries: Rice's Interactive Environment for Scheme)
;;
;; Shriram Krishnamurthi
;; shriram@cs.rice.edu
;;
;; Version 1.0  ;; (NOT REALLY SEE MODIFICATION NOTE ABOVE)
;; February 17, 1995

(define aries:checking? #f) ;;to distinguish using checker and aries.

(define aries:top-level #t)

(define-structure (aries:token text type
			       start-line start-column
			       current-line current-column))

(define-structure (aries:lexeme object type start-line start-column))

(define-structure (aries:sequence contents type
				  start-paren end-paren))

(define-structure (aries:paren text line column))

(define-structure (aries:error-report line column))

(define aries:quote-prefix-list
  (list 'quote 'quasiquote 'unquote 'unquote-splicing))

(define aries:alphabet-list
  (list #\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
	#\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z))

(define aries:digit-list
  (list #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))

(define aries:identifier-initial-list
  (append aries:alphabet-list
	  (list #\! #\$ #\% #\& #\* #\/ #\: #\< #\= #\> #\? #\~ #\_ #\^)))

(define aries:identifier-subsequent-list
  (append aries:identifier-initial-list
	  aries:digit-list
	  (list #\. #\+ #\-)))

(define aries:whitespace-list
  (list #\space #\newline #\tab #\return))

(define aries:open-parens
  (list #\( #\[)) ;; we're being generous compared to R4RS

(define aries:close-parens
  (list #\) #\])) ;; even more generosity

(define aries:delimiter-list
  (append aries:whitespace-list
	  aries:open-parens aries:close-parens
	  (list #\" #\;
		#\' #\` #\,))) ;; this line added to stuff in R4RS

(define aries:primitive-list   ;'list removed because nothing to check!
  (list 'car 'cdr 'cons 'null? 'add1 'sub1 'zero? '+ '-
	'not 'atom? 'list? 'cons? 'eq? 'number? '* '/ '> '< '>= '<=
	'= 'symbol? 'pair? 'map))

(define aries:default-vector-value
  (let ([default-value (make-aries:lexeme 0 'number -1 -1)])
    (lambda args
      (cond
       [(null? args) default-value]
       [(pair? (cdr args)) (error 'aries:default-vector-value
				  "too many arguments")]
       [else
	(let ([new-value (car args)])
	  (set! default-value
		(make-aries:lexeme new-value
				   (cond
				    [(symbol? new-value) 'symbol]
				    [(string? new-value) 'string]
				    [(number? new-value) 'number]
				    [(char? new-value) 'char]
				    [(boolean? new-value) 'boolean]
				    [else (error 'aries:default-vector-value
						 "invalid type of value")])
				   -1 -1)))]))))

(define aries:trace-primitives #f)

(define aries:current-file-number #f)

;; (define aries:fifo-port #f)  ;;not used in Mr Ed implementation
;; (define aries:s2e-filename #f)

(define aries:resume #f)

(define aries:void
  (let ([void-object (if #f #f)])
    (lambda ()
      void-object)))

(define aries:matched-parens?
  (lambda (open close)
    (or (and (char=? #\( open) (char=? #\) close))
	(and (char=? #\[ open) (char=? #\] close)))))

(define aries:ormap
  (lambda (predicate? list)
    (if (null? list)
	#f
	(if (predicate? (car list))
	    #t
	    (aries:ormap predicate? (cdr list))))))

(define aries:andmap
  (lambda (predicate? list)
    (if (null? list)
	#t
	(if (predicate? (car list))
	    (aries:andmap predicate? (cdr list))
	    #f))))

(define aries:string->char
  (let ([string->char-table '(("#\\newline" #\newline)
			      ("#\\nl"      #\newline)
			      ("#\\space"   #\space)
			      ("#\\tab"     #\tab))])
    (lambda (s)
      (let loop
	  ([table string->char-table])
	(if (null? table)
	    (and (= (string-length s) 3)
		 (char=? (string-ref s 0) #\#)
		 (char=? (string-ref s 1) #\\)
		 (string-ref s 2))
	    (if (string-ci=? s (car (car table)))
		(cadr (car table))
		(loop (cdr table))))))))

(define aries:scanner
  (lambda (port)
    (let ([current-line-number   1]  ;; since first line is 1
	  [current-column-number 0]  ;; will be incremented at first read
	  [start-line-number     'error]
	  [start-column-number   'error]
	  [c                     'error]
	  [next-c                'error]
	  [first-time            #f])
      (letrec
	  ([advance-input!
	    (lambda ()
	      (set! c (read-char port))
	      (set! next-c (peek-char port))
	      (if (not (eof-object? c))
		  (cond
		   [(char=? c #\newline)
		    (set! current-line-number (+ 1 current-line-number))
		    (set! current-column-number 0)]
;		   [(char=? c #\tab)
;		    (set! current-column-number
;			  (+ current-column-number
;			     (- 8 (remainder current-column-number 8))))]
		   [else
		    (set! current-column-number
			  (+ 1 current-column-number))])
		  'endif))]
	   [record-start-position!
	    (lambda ()
	      (set! start-line-number current-line-number)
	      (set! start-column-number current-column-number))]
	   [char-equal?
	    (lambda (c1 c2)
	      (if (or (eof-object? c1) (eof-object? c2))
		  #f
		  (char-ci=? c1 c2)))]
	   [char-contained?
	    (lambda (c l)
	      (aries:ormap (lambda (elt) (char-equal? c elt)) l))]
	   [skip-comment!
	    (lambda ()
	      (let loop
		  ([in-comment #f])
		(if in-comment
		    (begin
		      (advance-input!)
		      (if (or (eof-object? c)(char-equal? c #\newline))
			  (aries:void)
			  (loop #t)))
		    (if (char-equal? c #\;)
			(loop #t)
			(aries:void)))))]
	   [skip-atmosphere!
	    (lambda ()
	      (skip-comment!)
	      (if (char-contained? c aries:whitespace-list)
		  (begin
		    (advance-input!)
		    (skip-atmosphere!))
		  'endif))])
	(advance-input!)
	(lambda (state paren-file text)
	  (skip-atmosphere!)
	  (record-start-position!)
	  (set! first-time #t)
	  (let reader
	      ([state      state]
	       [paren-file paren-file]
	       [text       text])
	    (letrec
		([return-unread
		  (lambda (type paren-file text)
		    (values
		     (make-aries:token text type
				start-line-number start-column-number
				current-line-number current-column-number)
		     paren-file))]
		 [return
		  (lambda (type paren-file text)
		    (advance-input!)
		    (return-unread type paren-file text))]
		 [in-state?
		  (lambda (check-state)
		    (eq? state check-state))]
		 [match-input
		  (lambda (input)
		    (let matcher
			([input (string->list input)])
		      (if (null? input)
			  #t
			  (if (char-equal? c (car input))
			      (begin
				(advance-input!)
				(matcher (cdr input)))
			      #f))))])
	      (if first-time
		  (set! first-time #f)
		  (advance-input!))
	      (cond

	       ;; eof
	       [(eof-object? c)
		(return 'eof paren-file text)]

	       ;; string
	       [(and (in-state? 'ether)
		     (char-equal? c #\"))
		(reader 'string paren-file text)]
	       [(and (in-state? 'string)
		     (char-equal? c #\\))
		;; don't need to cons the #\\ on since
		;; list->string will take care of it for us;
		;; also note we let anything be quoted inside a string
		(advance-input!)
		(reader state paren-file (cons c text))]
	       [(and (in-state? 'string)
		     (char-equal? c #\"))
		(return 'string paren-file text)]
	       [(in-state? 'string)
		(reader state paren-file (cons c text))]

	       ;; list/vector
	       [(and (in-state? 'ether)
		     (char-contained? c aries:open-parens))
		(return 'list
			(cons (make-aries:paren c
						start-line-number
						start-column-number)
			      paren-file)
			text)]
	       [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c aries:open-parens))
		(advance-input!)
		(return 'vector
			(cons (make-aries:paren c
						start-line-number
						start-column-number)
			      paren-file)
			text)]
	       [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c aries:digit-list))
		(reader 'vector-w/-size paren-file text)]
	       [(and (in-state? 'vector-w/-size)
		     (char-contained? c aries:digit-list))
		(reader 'vector-w/-size paren-file (cons c text))]
	       [(and (in-state? 'vector-w/-size)
		     (char-contained? c aries:open-parens))
		(return 'vector
			(cons (make-aries:paren c
						start-line-number
						start-column-number)
			      paren-file)
			text)]
	       [(in-state? 'vector-w/-size)
		(aries:report-static-error
		 "Non-digit character in vector size prefix"
		 (make-aries:error-report current-line-number
					  current-column-number))]
	       [(and (in-state? 'ether)
		     (char-contained? c aries:close-parens))
		(if (null? paren-file)
		    (aries:report-static-error
		     "No matching opening parenthesis"
		     (make-aries:error-report start-line-number
					      start-column-number))
		    (let* ([to-match (car paren-file)]
			   [open (aries:paren-text to-match)])
		      (if (aries:matched-parens? open c)
			  (return 'endseq (cdr paren-file) text)
			  (aries:report-static-error
			   "Mismatched parenthesis"
			   (make-aries:error-report start-line-number
						    start-column-number)))))]

	       ;; boolean
	       [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (or (char-equal? next-c #\t)
			 (char-equal? next-c #\f)))
		(advance-input!)
		;; remember: consing must be done in reverse order
		(return 'boolean paren-file (cons c (cons #\# text)))]

	       ;; period
	       [(and (in-state? 'ether)
		     (char-equal? c #\.)
		     (char-contained? next-c aries:delimiter-list))
		(return 'period paren-file (cons c text))]

	       ;; quote/quasiquote/unquote/unquote-splicing
	       [(and (in-state? 'ether)
		     (or (char-equal? c #\')
			 (char-equal? c #\`)
			 (char-equal? c #\,)))
		(case c
		  [(#\') (return 'quote paren-file (cons c text))]
		  [(#\`) (return 'quasiquote paren-file (cons c text))]
		  [(#\,) (return
			  (if (char-equal? next-c #\@)
			      (begin
				(advance-input!)
				'unquote-splicing)
			      'unquote)
			  paren-file
			  (cons c text))])]

	       ;; char/number/symbol
	       [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-equal? next-c #\\))
		(advance-input!)
		(if (char-contained? next-c aries:alphabet-list)
		    (reader 'until-delim paren-file (cons c (cons #\# text)))
		    (begin
		      (advance-input!)
		      (return 'char paren-file
			      (cons c (cons #\\ (cons #\# text))))))]
	       [(in-state? 'ether)
		(reader 'until-delim paren-file (cons c text))]
	       [(in-state? 'until-delim)
		(if (char-contained? c aries:delimiter-list)
		    (let ([text-string (list->string (reverse text))])
		      (cond
		       [(string->number text-string)
			(return-unread 'number paren-file text)]
		       [(aries:string->char text-string)
			(return-unread 'char paren-file text)]
		       [(and (>= (string-length text-string) 2)
			     (char-equal? (string-ref text-string 0) #\#)
			     (char-equal? (string-ref text-string 1) #\\))
			(aries:report-static-error
			 "Invalid character syntax"
			 (make-aries:error-report start-line-number
					    start-column-number))]
		       [(string->symbol text-string)
			(return-unread 'symbol paren-file
				       text)]
		       [else
			(error 'aries:scanner
			       "Should not get here ~s" text-string)]))
		    (reader 'until-delim paren-file (cons c text)))]

	       [else
		(error 'aries:scanner
		       "Should not get here ~s ~s" state c)]))))))))

(define aries:make-eof-lexeme
  (lambda (line column)
    (make-aries:lexeme (aries:void) 'eof line column)))
(define aries:eof-lexeme?
  (lambda (lexeme)
    (and (aries:lexeme? lexeme) (eq? 'eof (aries:lexeme-type lexeme)))))

(define aries:make-endseq-lexeme
  (lambda (line column)
    (make-aries:lexeme (aries:void) 'endseq line column)))
(define aries:endseq-lexeme?
  (lambda (lexeme)
    (and (aries:lexeme? lexeme) (eq? 'endseq (aries:lexeme-type lexeme)))))

(define aries:make-period-lexeme
  (lambda (line column)
    (make-aries:lexeme (aries:void) 'period line column)))
(define aries:period-lexeme?
  (lambda (lexeme)
    (and (aries:lexeme? lexeme) (eq? 'period (aries:lexeme-type lexeme)))))

(define aries:read
  (lambda filename
    (let ([getter (aries:scanner (if (null? filename)
				     (current-input-port)
				     (open-input-file (car filename))))])
      (lambda ()
	(letrec

	    ([read-one
	      (lambda (paren-file sequence-type only-endseq-okay?)
		(call-with-values
		 (lambda () (getter 'ether paren-file '()))
		 (lambda (token paren-file)
		  (if (and only-endseq-okay?
			   (not (eq? (aries:token-type token) 'endseq)))
		      (aries:report-static-error "Not end of sequence" token)
		      'endif)
		  (case (aries:token-type token)
		    [(eof) (if (null? paren-file)
			       (values (aries:make-eof-lexeme
					      (aries:token-start-line token)
					      (aries:token-start-column token))
					     paren-file)
			       (aries:report-static-error
				"Unclosed parenthesis" (car paren-file)))]
		    [(list) (call-with-values
			     (lambda ()
			       (read-sequence paren-file 'list))
			     (lambda (sequence last-paren new-paren-file)
			       (values
				(make-aries:sequence
				 sequence 'list
				 (car paren-file) last-paren)
				new-paren-file)))]
		    [(vector)
		     (call-with-values
		      (lambda () (read-sequence paren-file 'vector))
		      (lambda (sequence last-paren new-paren-file)
			(let ([length-specified
			       (string->number
				(list->string
				 (reverse (aries:token-text token))))])
			  (cond
			   [length-specified
			    (let ([final-element
				   (let loop ([sequence sequence]
					      [length-so-far 0]
					      [last-one
					       (aries:default-vector-value)])
				     (cond
				      [(null? sequence)
				       last-one]
				      [(>= length-so-far length-specified)
				       (aries:report-static-error
					"Vector longer than specified length"
					(car sequence))]
				      [else
				       (loop (cdr sequence) (+ 1 length-so-far)
					     (car sequence))]))])
			      (let ([v (make-vector length-specified
						    final-element)])
				(values
				 (make-aries:sequence
				  (let loop ([index 0] [sequence sequence])
				    (if (null? sequence)
					v
					(begin
					  (vector-set! v index (car sequence))
					  (loop (+ 1 index) (cdr sequence)))))
				  'vector (car paren-file) last-paren)
				 new-paren-file)))]
			   [else
			    (values
			     (make-aries:sequence
			      (apply vector sequence) 'vector
			      (car paren-file) last-paren)
			     new-paren-file)]))))]
		    [(endseq) (if sequence-type
				  (values (aries:make-endseq-lexeme
						 (aries:token-start-line token)
						 (aries:token-start-column token))
						paren-file)
				  (aries:report-static-error
				   "Extra parenthesis" token))]
		    [(quote quasiquote unquote unquote-splicing)
		     (call-with-values
		      (lambda () (read-one paren-file sequence-type #f))
		      (lambda (result paren-file)
			(values
			 (list (make-aries:lexeme
				(aries:token-type token)
				(aries:token-type token)
				(aries:token-start-line token)
				(aries:token-start-column token))
			       result)
			 paren-file)))]
		    [(period) 
		     (case sequence-type
		       [(list) (values
				(aries:make-period-lexeme
				 (aries:token-start-line token)
				 (aries:token-start-column token))
				paren-file)]
		       [(vector) (aries:report-static-error
				  ". inside vector" token)]
		       [else 
			(aries:report-static-error
			 ". outside list" token)])]
		    [(string boolean char symbol number)
		     (values
		      (make-aries:lexeme
		       (let ([text (list->string
				    (reverse (aries:token-text token)))])
			 (case (aries:token-type token)
			   [(string) text]
			   [(boolean) (string-ci=? text "#t")]
			   [(char) (aries:string->char text)]
			   [(symbol) (string->symbol text)]
			   [(number) (string->number text)]))
		       (aries:token-type token)
		       (aries:token-start-line token)
		       (aries:token-start-column token))
		      paren-file)]
		    [else
		     (error 'read-one
			    "Should not get here ~s" token)]))))]

	     [read-sequence
	      (lambda (paren-file sequence-type)
		(let loop
		    ([paren-file paren-file]
		     [accum '()])
		  (call-with-values
		   (lambda () (read-one paren-file sequence-type #f))
		   (lambda (lexeme paren-file)
		     (cond
		      [(aries:endseq-lexeme? lexeme)
		       (values (reverse accum) lexeme paren-file)]
		      [(aries:period-lexeme? lexeme)
		       (if (null? accum)
			   (aries:report-static-error
			    "No datum before ." lexeme)
			   'endif)
		       (call-with-values
			(lambda () (read-one paren-file sequence-type #f))
			(lambda (this+1 paren-file)
			  (cond
			   [(aries:period-lexeme? this+1)
			    (aries:report-static-error
			     "Adjacent periods" this+1)]
			   [(aries:endseq-lexeme? this+1)
			    (aries:report-static-error
			     "Nothing following period" this+1)]
			   [(aries:eof-lexeme? this+1)
			    (aries:report-static-error "Premature EOF" this+1)]
			   [(or (aries:lexeme? this+1)
				(aries:sequence? this+1))
			    (call-with-values
			     (lambda () (read-one paren-file sequence-type #t))
			     (lambda (this+2 paren-file)
			       (cond
				[(aries:endseq-lexeme? this+2)
				 (values (append (reverse accum) this+1)
					       this+2
					       paren-file)]
				[else
				 (aries:report-static-error
				  "More than one datum after ." this+2)])))]
			   [else
			    (aries:report-static-error
			     "Invalid datum after ." this+1)])))]
		      [(aries:eof-lexeme? lexeme)
		       (aries:report-static-error
			"Unclosed parenthesis" (car paren-file))]
		      [(or (aries:lexeme? lexeme)
			   (aries:sequence? lexeme)
			   (pair? lexeme) (null? lexeme)
			   (vector? lexeme))
		       (loop paren-file (cons lexeme accum))]
		      [else
		       (error 'read-sequence
			      "Should not get here ~s" lexeme)])))))])

	  (read-one '() #f #f))))))

;;use of globals is a kludge, but keeps interface with Shriram's stuff the
;;same -- or minimally different.  It's ugly, ugly, ugly, but it works. 
;;check-buffer sets aries:checking? and then unsets it.
;;The idea is this: the syntax checker uses aries:reader, but we want
;;report-static-error to do a very different thing. So we have 
;;report-static-error just set some globals (parse-error and parse-place)
;;instead of calling transmit-to-mred.

(define aries:report-static-error
  (lambda (name object)
    (call-with-values
     (lambda ()
       (cond
	[(aries:token? object)
	 (values aries:token-start-line aries:token-start-column)]
	[(aries:lexeme? object)
	 (values aries:lexeme-start-line aries:lexeme-start-column)]
	[(aries:error-report? object)
	 (values aries:error-report-line aries:error-report-column)]
	[(aries:paren? object)
	 (values aries:paren-line aries:paren-column)]
	[else
	 (error 'aries:report-static-error "Should not get here ~s" object)]))
     (lambda (row-extractor column-extractor)
       (if aries:checking?
	   (begin  (set! parse-error name)
		   (set! parse-place (cons (column-extractor object)
					   (row-extractor object))))
	   (aries:transmit-to-mred 'static-error name
				   (row-extractor object)
				   (column-extractor object)
				   aries:current-file-number '()))
       (aries:top-level (aries:void))))))

;; takes the list messages, puts it into a string and writes it to
;; Mr Ed. The one-list loop takes a nested list (one-level only) and puts
;; its contents directly in the main list. It turns out Mr Ed can better
;; parse the string it gets when there are not nested lists. Only need
;; to "unnest" one level because that is all the nesting this function
;; ever gets (if that changes, then this can be easily changed).


(define aries:transmit-to-mred
  (lambda messages
    (let ([msg-str
	   (aries:expr->string 
	    (let one-list ([left messages])
	      (cond [(null? left) ()]
		    [(list? (car left))
		     (append (car left)(one-list (cdr left)))]
		    [else (cons (car left)(one-list (cdr left)))])))])
      (newline)
      (if aries:wx-ipc-connection
	  (begin (wx-ipc-write aries:wx-ipc-connection msg-str)(aries:void))
	  (begin (aries:set-info msg-str)(aries:void))))))

      



