;; Dan Grossman
;; Uses Aries Scheme Parser to check questionable syntax for Scheme
;; in Mr. Ed.
;; 7/95

;;position-msg-y contains a lot of  margin constants that may need to be 
;; parameterized for other environments/platforms. I just don't know.

;; idea is to derive classes from this one with different keyword-checks-list
;; and all-checks-list to create a custom checker. To build on lists
;; have derived class construction invoke add  and remove  functions.

;;these need to stay global. They are used to pass messages from aries
;;to check. (this is pretty much an error-number style hack).

(define parse-error #f)
(define parse-place #f)

;;structure for elements of keyword-checks-list.
(define-structure (keyword-check keywords functions))

(define std-check%
  (make-class
   '()
   (public
    [name "std-check"])

;;;this is the "laundry list" of functions that can be called:
; Preceding comments indicate which keywords they are aimed at.
; All functions must take 3 arguments: the lexeme or sequence,
;  the parent-sequence and a boolean first-in-seq?
; Predefined lists included too! Please add to and change them.
; Generally assume no local bindings of keywords
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   (public
;;;;;;LISTS!
    ; should be used with syntax-first in the lex-checks-list
    [syntax-list
     (list 'define 'lambda 'if 'cond 'else 'let 'let* 'letrec 'local
	   'begin 'begin0 'or 'and 'when 'unless 'quote 'delay)]
;;; these are not "complete"
    [one-arg-words
     (list 'car 'cdr 
	   'caar 'cadr 'cdar 'cddr
	   'caaar 'caadr 'cadar 'caddr 'cdaar 'cdadr 'cddar 'cdddr
	   'caaaar 'caaadr 'caadar 'caaddr 'cadaar 'cadadr 'caddar 'cadddr
	   'cdaaar 'cdaadr 'cdadar 'cdaddr 'cddaar 'cddadr 'cdddar 'cddddr
	   'null? 'zero? 'pair? 'cons? 'atom? 
	   'list? 'symbol? 'odd? 'even? 'string?
	   'positive? 'negative?
	   'sybmol->string 'string->symbol
	   'add1 'sub1 'abs 'length 'reverse
	   'force
	   'quote 'quasiquote)]
    [two-arg-words(list 'cons 'define 'eq? 'eqv? 'equal? 'quotient 'remainder
			'modulo  'set! 'set-car! 'set-cdr! 'list-tail 'list-ref
			'memq 'memv 'member 'assq 'assv 'assoc
			'char=? 'char<? 'char>? 'char<=? 'char>=?)]
    [two-or-three-arg-words (list 'if)]
    [min-one-arg-words (list '- '\ )]
    [min-two-arg-words
     (list 'lambda '> '< '>= '<= '= '+ '* 'or 'and 'max 'min)]
    [let-args-words (list 'let 'letrec 'let* 'local)]
    [quote-words (list 'quote 'quasiquote)]
    ;; use for beginners unlikely to intend an if without an else:
    [three-arg-words-beginner (list 'if)]
;;;;;;;FUNCTIONS!
    ;;will return false and issure a message if the contents
    ;;are not in a list (probably because there's a dot: .)
    ;;this is a helper function to avoid errors in accessing contents.
    [get-seq-contents-list
     (lambda (seq)
       (let ([contents (aries:sequence-contents seq)])
	 (if (list? contents)
	     contents
	     (begin (produce-message
		     "Something here does not have the form of a list.")
		    #f))))]
    ;;may be useful to point to
    [always-true
     (lambda (lex parent-seq first-in-seq?)
       #t)]
    ;;make sure a syntactic keyword comes first in a sequence _OR_ that
    ;;"quote" is the first in the sequence.
    ;;designed for lex-checks-list. Uses syntax-list defined above.
    ;;example of false: (if #t if #f)
    [syntax-first
     (lambda (lex parent-seq first-in-seq?)
       (if (or first-in-seq?
	       (not (member (aries:lexeme-object lex) syntax-list))
	       (let ([parent-contents (get-seq-contents-list parent-seq)])
		 (and parent-contents 
		      (aries:lexeme? (car parent-contents))
		      (member (aries:lexeme-object (car parent-contents))
			      quote-words)))) 
	   #t
	   (begin (produce-message
		   (string-append  
		    "Warning: "
		    (symbol->string (aries:lexeme-object lex))
		    " does not come right after an open parenthesis."))
		  #f)))]   
    ;;make sure define is followed by a symbol
    ;;example of false: (define 'foo 3)
    ;;if you use the form (define (x . y) blah) DON'T use this; it will return
    ;;false
    ;;this assumes define is first in sequence, so use syntax-first first.
    [def-symbol 
      (lambda (lex parent-seq first-in-seq?)
	(let ([contents (get-seq-contents-list parent-seq)])
	  (cond [(not contents) #f]
		[(null? (cdr contents))
		 (begin (produce-message "define not followed by anything.")
			#f)]
		[(aries:lexeme? (cadr contents)) #t]
		[else (begin (produce-message 
			      "Warning: define is not followed by a symbol.")
			     #f)])))]
    ;;The following set of functions use num-args to make sure the number
    ;;of arguments following a piece of syntax is correct.
    ;;example of false (if 1 2 3 4)
    ;;because some syntax like "if" can take a finite variety of num-args
    ;;the acceptable numbers are all passed as arguments.
    ;;this assumes the syntax is first in sequence.
    [num-args
     (lambda (lex parent-seq . correct-nums)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (if (not contents) 
	     #f
	     (let ([actual-num (sub1 (length contents))])
	       (if (member actual-num correct-nums)
		   #t
		   (begin (produce-message
			   (string-append 
			    (symbol->string (aries:lexeme-object lex))
			    " is followed by " (number->string actual-num)
			    " expression(s)."))
			  #f))))))]
    [one-arg
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq? 
	   (num-args lex parent-seq 1)
	   #t))]
    [two-args
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq?
	   (num-args lex parent-seq 2)
	   #t))]
    [two-or-three-args
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq? 
	   (num-args lex parent-seq 2 3)
	   #t))]
    [three-args
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq?
	   (num-args lex parent-seq 3)
	   #t))]
    ;; same as num-args with the obvious difference
    [at-least-num-args
     (lambda (lex parent-seq minimum-num)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (if (not contents)
	     #f
	     (let ([actual-num (sub1 (length contents))])
	       (if (>= actual-num minimum-num)
		   #t
		   (begin (produce-message
			   (string-append 
			    (symbol->string (aries:lexeme-object lex))
			    " is followed by " (number->string actual-num)
			    " expression(s)."))
			  #f))))))]
    [min-two-args
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq?
	   (at-least-num-args lex parent-seq 2)))]
    [min-one-arg
     (lambda (lex parent-seq first-in-seq?)
       (if first-in-seq?
	   (at-least-num-args lex parent-seq 1)))]
    ;;Make sure the bindings in a let type form are in parentheses.
    ;;false example (let [x 3]) or (let ([x 3] y 4)), etc.
    ;;Also makes sure not binding to a symbol (eg. (let (['x 3])))
    ;;Also makes sure binding to something (eg. ([x])
    ;;assumes let is first-in-sequence so run syntax-first first.
    ;;allows: (let loop ([x 3])) or (let ([x 3])) Use let-beginner to disallow
    ;; the named let version
    [bindings-in-parens
     (lambda (lex parent-seq first-in-seq?)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (if 
	  (not contents)
	  #f
	  (let* ([next-obj (cadr contents)]
		 [bindings-obj (if (aries:lexeme? next-obj)
				   (if (pair? (cdr contents))
				       (caddr contents)
				       ())
				   next-obj)])
	    (if (not (aries:sequence? bindings-obj))
		(begin (produce-message "Bindings list is not a list.")
		       #f)
		(let bindings([seqs-left(aries:sequence-contents bindings-obj)]
			      [binding-num 1])
		  (cond 
		   [(null? seqs-left) #t]
		   [(aries:sequence? (car seqs-left))
		    (let ([binding-contents 
			   (get-seq-contents-list (car seqs-left))])
		      (cond
		       [(not binding-contents) #f]
		       [(or(null? binding-contents)
			   (not(aries:lexeme? (car binding-contents)))
			   (not(eq?(aries:lexeme-type(car binding-contents)) 
				   'symbol)))
			(begin
			  (produce-message
			   (string-append 
			    "Binding number " (number->string binding-num)
			  " in the bindings list is not binding to a symbol."
			  (string #\newline)
			  "Rest of bindings cannot be checked."))
			  #f)]
		       [(null? (cdr binding-contents))
			(begin
			  (produce-message
			   (string-append
			    "Binding number " (number->string binding-num)
			    " in the bindings list is not binding to anything."
			    (string #\newline)
			    "Rest of bindings cannot be checked."
			    )) #f)]
		       [else (bindings (cdr seqs-left)(add1 binding-num))]))]
		   [else (begin 
			   (produce-message 
			    (string-append
			     "Binding number "
			     (number->string binding-num)
			     " of this let-form is not in parentheses. "))
			   #f)])))))))]
    ;;DON'T use this if you have named let forms:
    ;;returns false on (let arg (body))
    [let-beginner
     (lambda (lex parent-seq first-in-seq?)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (if (not contents)
	     #f
	     (if (or (null? contents)(null? (cdr contents)) 
		     (not (aries:sequence? (cadr contents))))
		 (begin (produce-message 
			 (string-append 
			  (symbol->string (aries:lexeme-object lex))
			  " is not followed by a list of bindings."))
			#f)))))]
    ;;DON'T use this if you have functions which take arbitrary numbers of
    ;; arguments.
    ;; i.e. it will return false on (lambda x 3)      
    [lambda-beginner
     (lambda (lex parent-seq first-in-seq?)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (cond [(not contents) #f]
	       [(null? (cdr contents))
		(begin (produce-message "lambda has no parameters or body.")
		       #f)]
	       [(not (aries:sequence? (cadr contents)))
		(begin (produce-message 
			"lambda is not followed by a list of parameters.")
		       #f)]
	       [else #t])))]
    ;; cond is a very common and peculiar form. This will return false on
    ;; several things including implied begins.
    [cond-check
     (lambda (lex parent-seq first-in-seq?)
       (let ([contents (get-seq-contents-list parent-seq)])
	 (cond 
	  [(not contents) #f]
	  [(null? (cdr contents))
	   (begin (produce-message "cond has no clauses.")
		  #f)]
	  [else
	   (let clauses ([clauses-left (cdr contents)][clause-num 1])
	     (cond [(null? clauses-left) #t]
		   [(not (aries:sequence? (car clauses-left)))
		    (begin (produce-message
			    (string-append
			     "Clause number "
			     (number->string clause-num)
			     " of this cond is not in parentheses."
			     (string #\newline)
			     "Cannot check other clauses."))
			   #f)]
		   [else 
		    (let ([clause-contents 
			   (get-seq-contents-list (car clauses-left))])
		      (cond [(not contents) #f]
			    [(null? clause-contents)
			     (begin (produce-message
				     (string-append
				      "Clause number "
				      (number->string clause-num)
				      " is an empty list."
				      (string #\newline)
				      "Cannot check other clauses."))
				    #f)]
			    [(null? (cdr clause-contents))
			     (begin (produce-message
				     (string-append
				      "Clause number "
				      (number->string clause-num)
				      " only has a test (no response)."
				      (string #\newline)
				      "Cannot check other clauses."))
				    #f)]
			    [(not (null? (cddr clause-contents)))
			     (begin 
			       (produce-message
				(string-append
				 "Clause number "
				 (number->string clause-num)
				 " has more than one response (implied begin)."
				 (string #\newline)
				 "Cannot check other clauses."))
			       #f)]
			    [else (clauses (cdr clauses-left)
					   (add1 clause-num))]))]))])))])

   (public
;;; used to create and process messages.
    [produce-message
     (lambda (str)
       (set! the-message str))]
    [the-message #f]

;;; functions to show what produced a #f.
    [buffer-place
     (lambda (edit buffer-pos col line)
       (let ([last-pos (send edit last-position)])
	 (let get-row-loop ([rows-left line][pos buffer-pos])
	   (if (zero? rows-left)
	       (+ pos col)
	       (get-row-loop (sub1 rows-left)
			     (add1
			      (send edit find-string 
				    (string #\newline) 1 pos last-pos)))))))]
    [show-lexeme  
     (lambda (canvas edit buffer-pos lex) 
       (let* ([start-col (sub1 (aries:lexeme-start-column lex))]
	      [start-line (sub1 (aries:lexeme-start-line lex))]
	      [buff-start (buffer-place edit buffer-pos start-col start-line)]
	      [buff-end (+ buff-start (string-length
				       (symbol->string 
					(aries:lexeme-object lex))))])
	 (set-position-middle  edit buff-start buff-end)
	 (show-message-cont canvas edit)))]
    [show-seq
     (lambda (canvas edit buffer-pos seq) 
       (let* ([start-col (sub1 (aries:paren-column
				(aries:sequence-start-paren seq)))]
	      [start-line (sub1 (aries:paren-line
				 (aries:sequence-start-paren seq)))]
	      [buff-start (buffer-place edit buffer-pos start-col start-line)]
	      [end-col (aries:lexeme-start-column      ;;;end-paren is a lexeme
			(aries:sequence-end-paren seq))]
	      [end-line (sub1 (aries:lexeme-start-line
			       (aries:sequence-end-paren seq)))]
	      [buff-end (buffer-place edit buff-start
				      (if (= start-line end-line)
					  (- end-col start-col)
					  end-col)
				      (- end-line start-line))])
	 (set-position-middle  edit buff-start buff-end)
	 (show-message-cont canvas edit)))]
    [position-msg-y ;; lost of arbitrary margin numbers within
     (lambda (canvas edit)
       (let ([top-sel-holder (box 0)]
	     [bot-sel-holder (box 0)]
	     [top-view-holder (box 0)]
	     [bot-view-holder (box 0)])
	 (send edit position-location 
	       (send edit get-end-position) () bot-sel-holder #f)
	 (send edit local-to-global (box 0) bot-sel-holder)
	 (send canvas client-to-screen (box 0) bot-sel-holder)
	 (send canvas get-size (box 0) bot-view-holder)
	 (send canvas client-to-screen (box 0) bot-view-holder)
	 (if (< (unbox bot-sel-holder)(- (unbox bot-view-holder) 150))
	     (+ 10 (unbox bot-sel-holder))
	     (begin
	       (send edit position-location 
		     (send edit get-start-position) () top-sel-holder #f)
	       (send edit local-to-global (box 0) top-sel-holder)
	       (send canvas client-to-screen (box 0) top-sel-holder)
	       (send canvas client-to-screen (box 0) top-view-holder)
	       (if (> (- (unbox top-sel-holder) 150)(unbox top-view-holder))
		   (- (unbox top-sel-holder) 175)
		   (- (unbox bot-view-holder) 175))))))]
    [position-msg-x
     (lambda (canvas)
       (let ([edge (box 0)]
	     [width (box 0)])
	 (send canvas client-to-screen edge (box 0))
	 (send canvas get-size width (box 0))
	 (+ (unbox edge) (round (/ (unbox width) 3)))))] ; 3 is arbitrary
    [show-message-cont       ;;;the-message is global
     (lambda (canvas edit)
       (let ([top (position-msg-y canvas edit)]
	     [left (position-msg-x canvas)])
	 (if (mred:get-choice the-message "Continue" "Cancel" 
			      "Problem?" left top)
	     wx:const-yes
	     wx:const-no)))]
    [show-message-no-cont
     (lambda (canvas edit)
       (let ([top (position-msg-y canvas edit)]
	     [left (position-msg-x canvas)])
	 (wx:message-box the-message "Problem?" wx:const-cancel
			 canvas left top)
	 wx:const-ok))]
    [set-position-middle 
     (lambda (edit start end)
       (let ([curr-top-vis (box 0)]
	     [curr-bot-vis (box 0)])
	 (send edit get-visible-line-range curr-top-vis curr-bot-vis)
	 (let ([amt (quotient (- (unbox curr-bot-vis)(unbox curr-top-vis)) 4)]
	       [new-middle (send edit position-line start)])
	   (send edit scroll-to-position 
		 (send edit line-start-position (- new-middle amt)) #f
		 (send edit line-start-position (+ new-middle (* 3 amt))))
	   (send edit set-position  start end))))]
       	     
;;; actual buffer functions:  
    [do-lexeme
     (lambda (canvas edit buffer-pos lex parent-seq first-in-seq?)
       (and
	(let lex-check-loop ([lex-list lex-checks-list])
	  (if (null? lex-list)
	      #t
	      (if (or ((car lex-list) lex parent-seq first-in-seq?)
		      (=(show-lexeme canvas edit buffer-pos lex) wx:const-yes))
		  (lex-check-loop (cdr lex-list))
		  #f)))
	(let find-keyword-loop ([keyword-list keyword-checks-list])
	  (cond [(null? keyword-list) #t]
		[(member (aries:lexeme-object lex)
			 (keyword-check-keywords (car keyword-list)))
		 (and 
		  (let keyword-check-loop 
		      ([func-list (keyword-check-functions(car keyword-list))])
		    (cond [(null? func-list) #t]
			  [((car func-list) lex parent-seq first-in-seq?)
			   (keyword-check-loop (cdr func-list))]
			  [else 
			   (if (= (show-seq canvas edit buffer-pos parent-seq) 
				  wx:const-yes)
			       (keyword-check-loop (cdr func-list))
			       #f)]))
		  (find-keyword-loop (cdr keyword-list)))]
		[else (find-keyword-loop (cdr keyword-list))]))))]
    [do-seq
     (lambda (canvas edit buffer-pos seq parent-seq first-in-seq?)
      ;; maybe put a continuation in here for quote problem.
       (and
	(let check-seq-loop ([seq-list seq-checks-list])
	  (cond 
	   [(null? seq-list) #t]  
	   [((car seq-list) seq parent-seq first-in-seq?)
	    (check-seq-loop (cdr seq-list))]
	   [else (if (=(show-seq canvas edit buffer-pos seq) wx:const-yes)
		     (check-seq-loop (cdr seq-list))
		     #f)])) 
	(if (null? seq) #t    ;;(empty list)
	    (let check-contents-loop 
		([remaining (aries:sequence-contents seq)]
		 [first-in-seq? #t])
	      (cond 
	       [(null? remaining) #t]
	       [(not (pair? remaining)) #t];;else handled well by parser
	       [(aries:sequence? (car remaining))
		(if (do-seq canvas edit buffer-pos
			    (car remaining) seq first-in-seq?)
		    (check-contents-loop (cdr remaining) #f)
		    #f)]
	       [(aries:lexeme? (car remaining))
		(if (do-lexeme canvas edit buffer-pos
			       (car remaining) seq first-in-seq?)
		    (check-contents-loop (cdr remaining) #f)
		    #f)]
	       ;;skip over quotes and perhaps other junk.
	       [else (check-contents-loop (cdr remaining) #f)])))))]
    
    [handle-parse-error
     (lambda (canvas edit buffer-pos)
       (produce-message (string-append "parse error: " parse-error))
       (let ([buff-start (buffer-place edit buffer-pos (sub1 (car parse-place))
				     (sub1 (cdr parse-place)))])
	 (set-position-middle  edit (sub1 buff-start) buff-start))
       (show-message-no-cont canvas edit))]
    [parse-one-top-level
     (lambda (canvas edit buffer-pos str)
       (let ([curr-input-port (current-input-port)]
	     [continue? #t])
	 (current-input-port (open-input-string str))
	 (let/ec x
	     (call-with-values
	      (begin (set! aries:top-level x)
		     (set! parse-error #f)(set! aries:checking? #t) (aries:read))
	      (lambda (expr paren-file)
		(set! aries:checking? #f)
		(if (aries:sequence? expr)
		    (set! continue?
			  (let ([cts (aries:sequence-contents expr)])
			    (if (and (pair? cts)
				     (aries:lexeme? (car cts))
				     (eq? (aries:lexeme-object (car cts))
					  'define-schema))
				#t
				(do-seq canvas edit buffer-pos expr #f #t))))
		    (begin 
		      (let ([buffer-pos (mred:skip-whitespace edit 
							      buffer-pos 1)])
			(set-position-middle  edit buffer-pos 
					      (add1 buffer-pos)))
		      (produce-message
		       "A top-level s-expression is not in parentheses.")
		      (set! continue?
			    (=(show-message-cont canvas edit) wx:const-yes)))))))
	 (begin0
	   (if parse-error
	       (begin 
		 (handle-parse-error canvas edit buffer-pos)
		 (set! parse-error #f)    
		 (send canvas force-display-focus  #f)
		 #f)
	       continue?)
	   (current-input-port curr-input-port))))]
    [check-buffer                           
     (opt-lambda (canvas 
		  [start-pos 0]
		  [end-pos (send (send canvas get-media) last-position)])
       (send canvas force-display-focus  #t)   
       (begin0
	(let ([edit (send canvas get-media)])		 
	  (let top-level-loop ([buffer-pos start-pos])
	    (if (= (mred:skip-whitespace edit buffer-pos 1) end-pos) 
		#t
		(let ([next-top-level (mred:scheme-forward-match 
				       edit buffer-pos end-pos)])
		  (cond 
		    [(and next-top-level (> next-top-level end-pos)) #t] 
		    [next-top-level
		     (and (parse-one-top-level
			   canvas edit buffer-pos
			   (send edit get-text buffer-pos next-top-level))
			  (top-level-loop next-top-level))]
		    [else
		     (begin 
		       (let ([buffer-pos (mred:skip-whitespace edit buffer-pos 1)])
			 (set-position-middle  edit buffer-pos 
			       (add1 buffer-pos)))
		       (produce-message 
			"This top-level definition does not balance parens.")
		       (show-message-no-cont canvas edit)
		       #f)])))))
	(send canvas force-display-focus  #f)))])
   (public
    [keyword-checks-list]
    [seq-checks-list]
    [lex-checks-list])
   (public
    [generate-keyword-checks-list
     (lambda (the-list)
       (set! keyword-checks-list the-list))]
    [generate-seq-checks-list
     (lambda (the-list)
       (set! seq-checks-list the-list))]
    [generate-lex-checks-list
     (lambda (the-list)
       (set! lex-checks-list the-list))])
  (lambda ()
    (generate-keyword-checks-list ())
    (generate-seq-checks-list ())
    (generate-lex-checks-list ()))))

