
(defmacro define-param (name value)
  `(unless (defined? (quote ,name))
     (define ,name ,value)))

; See setup.ss:
(define-param drscheme:use-setup? #f)

; User-settable
(define-param drscheme:verify-quit? #t)
(define-param drscheme:advanced? #f)
(define-param drscheme:wrap-program? #t)
(define-param drscheme:always-check? #t)
(define-param drscheme:print-as-input? #t)
(define-param drscheme:switch-to-in-main? #f)
(define-param drscheme:use-common-file-dialogs? (eq? wx:platform 'unix))
(define-param drscheme:allow-x-selection? #t)
(define-param drscheme:allow-analysis? #t)
(define-param drscheme:allow-stepper? #t)
(define-param drscheme:allow-mrslatex? #f)

(define-param drscheme:mzscheme-type 'wx)

(define-param drscheme:path "/usr/local/lib/plt/drscheme")

(define-param drscheme:mrspidey-home-directory
  (build-path drscheme:plt-home-directory "mrspidey"))
(define-param drscheme:mrspidey-path
  (build-path drscheme:mrspidey-home-directory "mrspidey.zo"))

(define-param drscheme:donkey-home-directory
  (build-path drscheme:plt-home-directory "donkey"))
(define-param drscheme:donkey-path
  (build-path drscheme:donkey-home-directory "setup.ss"))

; Use the standard get & put dialogs
(when drscheme:use-common-file-dialogs?
      (set! mred:get-file mred:common-get-file)
      (set! mred:put-file mred:common-put-file))

; Hook for adding to the menu bar:
(define-param drscheme:create-menu-bar-hook void)

; Mode flags:
(define-param drscheme:allow-execute? #t)
(define-param drscheme:start-show-mode 'both)

; Startup files:
(define-param drscheme:startup-files '())

; Called when DrScheme quits:
(define-param drscheme:exit
  (lambda ()
    #f))

(define drscheme:do-help
  (lambda ()
    (mred:open-hyper-view (build-absolute-path drscheme:path
					       "help"
					       "index.htx"))))

(unless drscheme:allow-x-selection?
  (wx:media-set-x-selection-mode #f))

(when drscheme:allow-analysis?
      (load drscheme:mrspidey-path))
(when drscheme:allow-stepper?
      (load drscheme:donkey-path))

(define drscheme:mzscheme-console-edit%
  (class-asi drscheme:spawned-process-console-edit%
    (inherit display-delta set-display-delta
	     this-out-write insert-prompt
	     initialize-console send-scheme
	     exported-eval)
    (rename [super-display-result display-result])
    (public
      [scheme-type drscheme:mzscheme-type]
      [use-aries? #t]
      [constant-globals? #f]
      [set!-undefined? #t]
      [autoprompting? #t]
      ;; [get-prompt (lambda () (make-object recp:prompt-snip%))]
      [get-prompt (lambda () 
		    (if drscheme:print-as-input?
			"|- "
			"> "))]
      [display-result
       (lambda (r)
	 (super-display-result
	  (if (or (void? r) (not drscheme:print-as-input?))
	      r
	      (print-convert r))))]
      [do-initialize
       (lambda ()
	 (let ([delta (make-object wx:style-delta%
				   wx:const-change-family
				   wx:const-decorative)])
	   (send delta set-delta wx:const-change-size 10)
	   ; Visible initialization
	   (let ([dd display-delta])
	     (dynamic-wind
	      (lambda ()
		(set-display-delta delta))
	      (lambda ()
		(this-out-write (string-append
				 "Welcome to DrScheme"
				 newline-string))
		(initialize-console)
		(send-scheme '(constant-name-all-globals 
			       'zodiac:sc-to-monitor
			       'graphics-flag
			       'GLOBAL-VIEWPORT-LIST
			       'zodiac:back-boxes
			       'zodiac:dt-global-env)))
	      (lambda ()
		(set-display-delta dd)))))
	 (insert-prompt))])))

      
(define drscheme:console-edit% drscheme:mzscheme-console-edit%)
      
(define drscheme:project-frame%
  (class-asi drscheme:scheme-project-frame%
    (inherit group console)
    (rename [super-on-close on-close])
    (public
      [on-close
       (lambda ()
	 (let ([c console])
	   (send group remove-frame c)
	   (send (ivar group buffers) kill-buffer (send c get-program-edit))
	   (set! console #f)
	   (let ([ok? (super-on-close)])
	     (set! console c)
	     (send group insert-frame c)
	     (if ok?
		 (send console project-closed))
	     ok?)))])))

(define drscheme:frame%
  (let-enumerate
      ((FILE-LOAD
	FILE-LOAD-PROJECT
	FILE-NEW
	FILE-NEW-PROJECT
	FILE-SAVE
	FILE-SAVE-AS
	FILE-RELOAD
	FILE-OTHER
	FILE-PRINT-PROGRAM
	FILE-PRINT-TRANSCRIPT
	FILE-QUIT 
	
	EDIT-UNDO
	EDIT-REDO
	EDIT-CLEAR
	EDIT-COPY
	EDIT-CUT
	EDIT-PASTE
	EDIT-FIND

	SCHEME-INDENT-SELECTION
	SCHEME-INDENT-BUFFER
	SCHEME-COMMENT-OUT
	SCHEME-UN-COMMENT-OUT

	HW-SETUP
	LAB-SETUP
          
	SHOW-EDIT
	SHOW-SCHEME
	SHOW-SPLIT
          
	END-OF-MENUS))
    (class mred:frame% ()
      (inherit canvas add-canvas remove-canvas
	       show menu-bar% make-menu
	       open-file check-saved active-edit active-canvas
	       TOP-MARGIN on-simple-size)
      (rename [super-on-menu-command on-menu-command]
	      [super-on-close on-close])
      (private
        file-menu
	show-menu
	show-both
	[canvas-show-mode 'both])
      (public
	[forced-quit? #f]
	[checker 
	 (make-object (if drscheme:advanced?
			  advance-check%
			  beginner-check%))]
	[get-canvas%
	 (lambda ()
	   (class-asi drscheme:scheme-project-editor-canvas%
	     (rename [super-make-panel make-panel])
	     (public
	       [style-flags wx:const-mcanvas-no-h-scroll]
	       [use-panel? #t]
	       [make-panel
		(lambda (frame x y)
		  (let ([panel (super-make-panel frame x y)])
		    (make-object wx:button% panel 
				 (lambda args (send console-edit break))
				 "Stop Executing")
		    (make-object wx:button% panel
				 (lambda args
				   (check-syntax))
				 "Check Syntax")
		    (make-object wx:button% panel
				 (lambda args
				   (drscheme:do-help))
				 "Help")
		    panel))])))]
	[get-edit%
	 (let ([% (class-asi mred:scheme-mode-edit%
		    (public
		      [auto-set-wrap? drscheme:wrap-program?]))])
	   (lambda ()
	     %))]
	     
	[get-program-edit
	 (lambda ()
	   (send program-canvas get-media))]
	
	[get-buffer-group
	 (lambda ()
	   (ivar (ivar project group) buffers))]
	
	[break (lambda () (send console-edit break))]

	project
	[has-project? #f]
	[is-project?
	 (lambda (name)
	   (let ([ext (filename-extension name)])
	     (and ext (string=? ext "spj"))))]
	[program-in-project?
	 (lambda ()
	   (and has-project?
		(let ([name (send (get-program-edit) get-filename)])
		  (and (string? name)
		       (send project file-in-project? 
			     (normalize-path name))))))]
	[open-project
	 (lambda (name)
	   (send (ivar project group) remove-frame this)
	   (send (get-buffer-group) kill-buffer (get-program-edit))
	   (let ([ok? (send project read-project name)])
	     (send (ivar project group) insert-frame this)
	     (when ok?
	       (send (get-buffer-group) add-buffer 
		     'file #f (get-program-edit))
	       (set! has-project? name)
	       (send project show name)
	       (if name
		   (let ([last-file #f])
		     (send project for-each-file
			   (lambda (f) (set! last-file f)))
		     (if last-file
			 (change-to-file last-file)))))))]
	[project-closed
	 (lambda ()
	   (set! has-project? #f)
	   (send (ivar project group) clear)
	   (send (ivar (ivar project group) buffers) add-buffer 
		 'file #f (get-program-edit))
	   (send (ivar project group) insert-frame this)
	   (send (ivar project group) set-frame-title-prefix "DrScheme"))]
	[change-to-file
	 (lambda (name)
	   (let ([buffer (send program-canvas get-media)])
	     (if (open-file name program-canvas #f)
		 (release-buffer buffer))))]
	
	[release-buffer
	 (lambda (buffer)
	   ; If this buffer is not displayed anywhere, drop it
	   (let ([name (if has-project?
			   (send buffer get-filename)
			   #f)])
	     (unless (and has-project?
			  (string? name)
			  (send project file-in-project? 
				(normalize-path name)))
	       (send (get-buffer-group) kill-buffer buffer)
	       (send buffer remove-autosave))))]

	[set-show-mode
	 (lambda (which)
	   (if (not (eq? which canvas-show-mode))
	       (cond
		; moving TO console-only? Show the special panel
		[(eq? which 'scheme)
		 (set! TOP-MARGIN (+ TOP-MARGIN SCHEME-ONLY-PANEL-HEIGHT))
		 (send scheme-only-panel set-size 0 0 -1 -1)]
		; moving FROM console-only? Hide the special panel
		[(eq? canvas-show-mode 'scheme)
		 (set! TOP-MARGIN (- TOP-MARGIN SCHEME-ONLY-PANEL-HEIGHT))
		 (send scheme-only-panel set-size 50000 50000 -1 -1)
		 '(on-simple-size)]
		[else #f]))
	   (set! canvas-show-mode which)
	   (if (or (eq? which 'program)
		   (eq? which 'both))
	       (add-canvas program-canvas #t)
	       (remove-canvas program-canvas))
	   (if (or (eq? which 'scheme)
		   (eq? which 'both))
	       (add-canvas console-canvas)
	       (remove-canvas console-canvas))
	   (map 
	    (lambda (id)
	      (send file-menu enable id (not (eq? which 'scheme))))
	    (list FILE-LOAD
		  FILE-NEW
		  FILE-SAVE
		  FILE-SAVE-AS 
		  FILE-RELOAD 
		  FILE-PRINT-PROGRAM))
	   (send file-menu enable FILE-PRINT-TRANSCRIPT 
		 (not (eq? which 'program))))]
	[check-console-shown
	 (lambda ()
	   (when (eq? canvas-show-mode 'program)
	     (send show-menu dispatch show-both)))]

	[next-menu-id
	 (lambda ()
	   END-OF-MENUS)]
	
	[on-menu-command
	 (lambda (op)
	   (cond
	     [(= op FILE-LOAD)
	      (if (or (program-in-project?)
		      (check-saved program-canvas "Open File"))
		  (let ([name (mred:get-file () "Select a File")])
		    (when name
		      (if (is-project? name)
			  (if (or (not has-project?)
				  (send project check-project-all-saved
					"Open Project"))
			      (open-project name))
			  (change-to-file name)))))]
	     [(= op FILE-LOAD-PROJECT)
	      (if (if has-project?
		      (send project check-project-all-saved "Open Project")
		      (check-saved program-canvas "Open Project"))
		  (let ([name 
			 (mred:get-file 
			  () "Select a Project"
			  ".*\\.spj"
			  "Project names use the extension \".spj\".")])
		    (when name
			  (open-project name))))]
	     [(= op FILE-NEW) 
	      (if (or (program-in-project?)
		      (check-saved program-canvas "New Program"))
		  (change-to-file #f))]
	     [(= op FILE-NEW-PROJECT)
	      (if (if has-project?
		      (send project check-project-all-saved "Open Project")
		      (check-saved program-canvas "Open Project"))
		  (let ([name (mred:put-file
			       () () #f "New Project Name"
			       ".*\\.spj"
			       "Project names must end with \".spj\".")])
		    (if name
			(open-project name))))]
	     [(= op FILE-SAVE) (send (get-program-edit) save-file)]
	     [(= op FILE-SAVE-AS) (send (get-program-edit) save-file "")]
	     [(= op FILE-RELOAD) (send (get-program-edit) load-file)]
	     [(= op FILE-OTHER) (send (ivar (ivar project group) buffers)
				      pick program-canvas)]
	     [(= op FILE-PRINT-PROGRAM) (send (get-program-edit) print ())]
	     [(= op FILE-PRINT-TRANSCRIPT) (send console-edit print ())]
	     
	     [(= op FILE-QUIT)
	      (when (on-close)
		(set! drscheme:frame #f)
		(show #f))]

	     [(= op SCHEME-INDENT-SELECTION)    
	      (send (ivar (active-edit) mode) tabify-selection (active-edit))]
	     [(= op SCHEME-INDENT-BUFFER)
	      (send (ivar (active-edit) mode) tabify-all (active-edit))]
	     [(= op SCHEME-COMMENT-OUT)
	      (send (ivar (active-edit) mode) comment-out-selection (active-edit))]
	     [(= op SCHEME-UN-COMMENT-OUT)
	       (send (ivar (active-edit) mode) uncomment-selection (active-edit))]

	     [(= op HW-SETUP) (drscheme:do-setup "hw")]
	     [(= op LAB-SETUP) (drscheme:do-setup "lab")]
	     
	     [(= op EDIT-UNDO)
	      (send (active-edit) do-edit wx:const-edit-undo)]
	     [(= op EDIT-REDO)
	      (send (active-edit) do-edit wx:const-edit-redo)]
	     [(= op EDIT-CLEAR)
	      (send (active-edit) do-edit wx:const-edit-clear)]
	     [(= op EDIT-COPY)
	      (send (active-edit) do-edit wx:const-edit-copy)]
	     [(= op EDIT-CUT)
	      (send (active-edit) do-edit wx:const-edit-cut)]
	     [(= op EDIT-PASTE)
	      (send (active-edit) do-edit wx:const-edit-paste)]
	     [(= op EDIT-FIND)
	      (let ([bx (box 0)]
		    [by (box 0)]) 
		(send (get-program-edit) position-location 
		      (send (active-edit) get-start-position) 
		      bx by)
		(make-object mred:find-string-frame% 
			     (active-canvas) ()
			     (unbox bx) (unbox by)
			     (list 'ignore-case)))]
	     
	     [else (super-on-menu-command op)]))]
	[make-menu-bar
	 (lambda ()
	   (let ([mb (make-object menu-bar%)]
		 [edit-menu (make-menu)]
		 [scheme-menu (make-menu)])
	     (set! file-menu (make-menu))
	     (set! show-menu (make-menu))

	     (send* file-menu 
		    (append FILE-LOAD "Open...")
		    (append FILE-LOAD-PROJECT "Open Project...")
		    (append FILE-NEW "New Program")
		    (append FILE-NEW-PROJECT "New Project...")
		    (append-separator)
		    (append FILE-SAVE "Save")
		    (append FILE-SAVE-AS "Save As...")
		    (append-separator)
		    (append FILE-RELOAD "Revert"))
	     (if drscheme:switch-to-in-main?
		 (send file-menu append FILE-OTHER "Switch To..."))
	     (send* file-menu
		    (append-separator)
		    (append FILE-PRINT-PROGRAM "Print Program...")
		    (append FILE-PRINT-TRANSCRIPT 
			    "Print Scheme Transcript...")
		    (append-separator)
		    (append FILE-QUIT "Quit"))
	     (send* edit-menu
		    (append EDIT-UNDO "Undo")
		    (append EDIT-REDO "Redo")
		    (append-separator)
		    (append EDIT-CLEAR "Clear")
		    (append EDIT-COPY "Copy")
		    (append EDIT-CUT "Cut")
		    (append EDIT-PASTE "Paste")
		    (append-separator)
		    (append EDIT-FIND "Find"))
	     (send* scheme-menu
		    (append  SCHEME-INDENT-SELECTION "Indent")
		    (append  SCHEME-INDENT-BUFFER "Indent All")
		    (append  SCHEME-COMMENT-OUT "Comment Out")
		    (append  SCHEME-UN-COMMENT-OUT "Uncomment"))
	     (when drscheme:use-setup?
	       (send* scheme-menu
		      (append-separator)
		      (append  HW-SETUP "Setup Homework...")
		      (append  LAB-SETUP "Setup Lab...")))
	     (set! show-both
		   (caddr
		    (send show-menu append-check-set
			  '(("Program Only" . program) 
			    ("Scheme Only" . scheme) 
			    ("Both" . both))
			  set-show-mode
			  (case drscheme:start-show-mode
			    [program 0]
			    [scheme 1]
			    [both 2]))))
	     (send* mb
		    (append file-menu "File")
		    (append edit-menu "Edit")
		    (append show-menu "Show")
		    (append scheme-menu "Scheme"))

	     (when drscheme:allow-analysis?
		   (mrspidey:create-menu-bar-hook mb))
	     (when drscheme:allow-stepper?
		   (donkey:create-menu-bar-hook mb))
	     
	     (drscheme:create-menu-bar-hook mb)

	     mb))]
	
	[on-close
	 (lambda ()
	   (if (and (or forced-quit?
			(not drscheme:verify-quit?)
			(mred:get-choice
			 "Are you sure you want to quit DrScheme?"
			 "Quit" "Cancel"))
		    (send (get-buffer-group) check-buffers "Quit")
		    (or (not has-project?)
			(send project check-project-saved "Quit"))
		    (super-on-close))
	       (begin
		 (send (ivar project group) remove-frame this)
		 (send (ivar project group) close-all)
		 (send project show #f)
		 (send console-canvas set-media '())
		 (shut-down)
		 (drscheme:exit)
		 #t)
	       #f))]
	
	[check-syntax
	 (opt-lambda ([canvas program-canvas][ok-msg? #t])
	   (if (mred:local-busy-cursor
		console-canvas
		(lambda ()
		  (send checker check-buffer canvas)))
	       (begin
		 (if ok-msg?
		     (let ([top (box 0)]
			   [left (box 0)])
		       (send canvas client-to-screen top left)
		       (wx:message-box "Syntax looks OK" 
				       "OK" 
				       wx:const-ok
				       canvas
				       (+ 20 (unbox top))
				       (+ 20 (unbox left)))))
		 #t)
	       #f))]
	[slatex-buffer
	 (opt-lambda (buffer [group (ivar project group)])
	    (when (or (not drscheme:always-check?)
		      (let ([canvas (send buffer get-canvas)])
			(and canvas
			     (check-syntax canvas #f))))
		  (check-console-shown)
		  (mred:local-busy-cursor
		   console-canvas
		   (lambda ()
		     (send console-edit slatex-buffer buffer group)))))]
	[load-buffer-into-scheme
	 (opt-lambda (buffer [group (ivar project group)])
	    (when (or (not drscheme:always-check?)
		      (let ([canvas (send buffer get-canvas)])
			(and canvas
			     (check-syntax canvas #f))))
		  (check-console-shown)
		  (send console-canvas set-focus)
		  (mred:local-busy-cursor
		   console-canvas
		   (lambda ()
		     (send console-edit load-buffer-into-scheme buffer group)))))]
	[slatex-frame-buffer
	 (lambda (buffer)
	   (thread
	    (lambda ()
	      (slatex-buffer buffer))))]
	[load-frame-buffer-into-scheme
	 (lambda (buffer)
	   (thread
	    (lambda ()
	      (load-buffer-into-scheme buffer))))]
	[running? #t] ; For project to know the console is OK
	[slatex-file
	 (opt-lambda (file [group (ivar project group)])
	   (check-console-shown)
	   (send console-canvas set-focus)
	   (mred:local-busy-cursor
	    console-canvas
	    (lambda ()
	      (send console-edit slatex-file file group))))]
	[load-file-into-scheme
	 (opt-lambda (file [group (ivar project group)])
	   (check-console-shown)
	   (send console-canvas set-focus)
	   (mred:local-busy-cursor
	    console-canvas
	    (lambda ()
	      (send console-edit load-file-into-scheme file group))))]
	[shut-down
	 (lambda ()
	   (send console-edit shut-down))]
	[on-forced-quit
	 (lambda ()
	   (set! forced-quit? #t)
	   (let loop ()
	     (if (not (on-close))
		 (loop)))
	   (show #f))]
	[on-quit
	 (lambda ()
	   (if (on-close)
	       (show #f)))]
	[set-title-prefix void])
      
      (sequence
	(set! project (make-object drscheme:project-frame% #f #f this #t))
	(send (ivar project group) set-frame-title-prefix "DrScheme")
				   
	(super-init #f #f (ivar project group)))

      (public
        [scheme-only-panel (make-object wx:panel% this 0 0 -1 -1)])

      (sequence
	(make-object wx:button% scheme-only-panel
		     (lambda args
		       (send console-edit break))
		     "Stop Executing")
	(make-object wx:button% scheme-only-panel
		     (lambda args
		       (drscheme:do-help))
		     "Help")
	(send scheme-only-panel new-line)
	(send scheme-only-panel fit))

      (public
        [SCHEME-ONLY-PANEL-HEIGHT (send scheme-only-panel get-height)])

      (sequence
	(send scheme-only-panel set-size 50000 50000 5000 -1))
	
       (public
	[program-canvas canvas]
	[console-edit (make-object drscheme:console-edit%)])

       (sequence
	(send (send canvas get-media) set-mode (make-object mred:scheme-mode%))
	(send console-edit set-auto-set-wrap #t))

       (public
	[console-canvas
	 (make-object mred:simple-frame-canvas% this
		      -1 -1 -1 -1 "" 
		      wx:const-mcanvas-no-h-scroll
		      100)])

       (sequence
	(send console-canvas set-media console-edit)
	(add-canvas console-canvas)
	
	(set-show-mode drscheme:start-show-mode)

	(show #t)
	
	(wx:flush-display) ; so that the frame is really shown
	(send program-canvas set-focus)
	
	(mred:local-busy-cursor
	 console-canvas
	 (lambda ()
	   (send console-edit do-initialize)
	   (map load-file-into-scheme drscheme:startup-files)))))))

(define drscheme:frame #f)

(define (drscheme:go) 
  (let ([f (make-object drscheme:frame%)])
    (set! drscheme:frame f)
    f))

