
; Console for spawned processes

(define drscheme:plt-home-directory
  (if (defined? 'drscheme:plt-home-directory)
      (normalize-path drscheme:plt-home-directory)
      mred:plt-home-directory))

(define drscheme:zodiac-directory 
  (if (defined? 'drscheme:zodiac-directory)
      (normalize-path drscheme:zodiac-directory)
      (build-absolute-path drscheme:plt-home-directory "zodiac")))

(define drscheme:rice-directory
  (if (defined? 'drscheme:rice-directory)
      (normalize-path drscheme:rice-directory)
      (build-absolute-path drscheme:plt-home-directory "rice")))

(define drscheme:mrslatex-directory
  (if (defined? 'drscheme:mrslatex-directory)
      (normalize-path drscheme:mrslatex-directory)
      (build-absolute-path drscheme:plt-home-directory "drscheme" "mrslatex")))

(define zodiac:home-directory drscheme:zodiac-directory)

(define drscheme:zodiac-loaded? #f)
(define drscheme:rice-loaded? #f)

(define drscheme:load-zodiac
  (lambda ()
    (unless drscheme:zodiac-loaded?
      (load (build-absolute-path drscheme:zodiac-directory "load.ss"))
      (fetch "mzscheme")
      (set! drscheme:zodiac-loaded? #t))))

(define drscheme:load-rice
  (opt-lambda ([w/graphics? #t])
    (drscheme:load-zodiac)
    (unless drscheme:rice-loaded?
      (aires)
      (rice)
      (load (build-absolute-path drscheme:rice-directory "ricedefs.ss"))
      (load (build-absolute-path drscheme:rice-directory "x-rice.ss"))
      (if w/graphics?
	  (load (build-absolute-path drscheme:rice-directory "graphics.ss")))
      (set! drscheme:rice-loaded? #t))))

(define drscheme:zodiac-eval
  (lambda (e)
    (eval (zodiac:expr-stripper (zodiac:expand e)))))

(define drscheme:zodiac-load
  (lambda (f)
    (let ([p (open-input-file f)])
      (dynamic-wind
       void
       (lambda ()
	 (let loop ()
	   (let ([v (read p)])
	     (unless (eof-object? v)
	       (drscheme:zodiac-eval v)
	       (loop)))))
       (lambda () (close-intput-file f))))))

(define drscheme:process-buffer-registry%
 (let-struct aries-file (buffer-box filename group)
  (class-asi '()
    (private
     [id-table (make-hash-table)]
     [aries-file-id-count 0])
    (public
     [make-aries-id
      (lambda (buffer filename group)
	(let/ec found-it
	  (hash-table-for-each
	   id-table
	   (lambda (id aries-file)
	     (let ([b (weak-box-value (aries-file-buffer-box aries-file))]
		   [g (aries-file-group aries-file)]
		   [f (aries-file-filename aries-file)])
	       (if (and (eq? group g)
			(or (and b (eq? b buffer))
			    (and filename (equal? filename f))))
		   (found-it id)))))
	  (let ([id (begin0
		     (string->symbol
		      (string-append "aries"
				     (number->string aries-file-id-count)))
		     (set! aries-file-id-count 
			   (add1 aries-file-id-count)))])
	    (hash-table-put! id-table id 
			     (make-aries-file (make-weak-box buffer) 
					      filename
					      group))
	    id)))]
     [file-name-from-id
      (lambda (id)
	(if (string? id)
	    id
	    (let ([v (hash-table-get id-table id (lambda () #f))])
	      (if v
		  (let ([fn (aries-file-filename v)])
		    (if fn
			fn
			(let ([buffer 
			       (weak-box-value (aries-file-buffer-box v))])
			  (if buffer
			      (let ([fn (send buffer get-filename)])
				(if (null? fn)
				    "Untitled"
				    fn))
			      "Unknown"))))
		  "Unknown"))))]
     [hilite-in-id
      (lambda (id position)
	(unless (string? id)
	  (let ([v (hash-table-get id-table id (lambda () #f))])
	    (if v
		(let ([buffer (weak-box-value (aries-file-buffer-box v))])
		  (if buffer
		      (hilite-in-buffer buffer position
					(aries-file-group v))
		      (set-aries-file-buffer-box!
		       v
		       (make-weak-box
			(let ([fn (aries-file-filename v)]
			      [group (aries-file-group v)])
			  (if group
			      (begin
				(send group open-file fn)
				(let ([b (send (ivar group buffers)
					       find-buffer-by-name 'file fn)])
				  (if b
				      (hilite-in-buffer b position group)
				      (hilite-in-file fn position))))
			      (hilite-in-file fn position)))))))))))]
     [hilite-in-file
      (lambda (file position)
	(let* ([f (mred:edit-file file)]
	       [buffer (send f active-edit)])
	  (hilite-in-buffer buffer position #f)))]
     [hilite-in-buffer
      (lambda (buffer position group)
	(let* ([frame (send buffer get-frame)])
	  (if frame
	      (send frame show #t)
	      (if group
		  (send group new-frame-for-buffer buffer)
		  (make-object mred:frame% buffer #t group))))
	(send buffer set-position position)
	(select-s-expression buffer)
	(let ([canvas (send buffer get-canvas)])
	  (if canvas
	      (send canvas set-focus)))
	buffer)]
     [select-s-expression
      (lambda (buffer)
	(let* ([pos (send buffer get-start-position)]
	       [end (mred:scheme-forward-match buffer pos
					       (send buffer last-position))])
	  (if end
	      (send buffer set-position pos end #f #t 
		    wx:const-local-select))))]))))

(define drscheme:make-spawned-process-console-edit%
  (lambda (super%)
   (class-asi super%
    (inherit this-err-write
	     this-out
	     this-out-write
	     this-err
	     this-in
	     do-pre-eval
	     do-post-eval
	     display-result
	     insert-prompt
	     get-canvas
	     ready-non-prompt autoprompting?
	     last-position
	     position-line
	     set-position
	     get-frame
	     begin-edit-sequence
	     end-edit-sequence
	     scroll-to-position)
    (rename
     [super-initialize-console initialize-console])
    (private
     [eval-sema (make-semaphore)]
     [return-sema (make-semaphore)]
     return-value
     return-error
     to-be-evaluated
     [break-flag #f]
     [break-ok? #f]
     [exiting? #f]
     [get-input
      (lambda ()
	to-be-evaluated)]
     [set-result
      (lambda (v)
	(set! return-value v)
	(semaphore-post return-sema))]
     [set-error
      (lambda (args)
	(set! return-error args)
	(if use-aries? (semaphore-post zodiac-reply))
	(semaphore-post return-sema))]
     [on-exit
      (lambda args
	(let ([frame (get-frame)])
	  (when frame
		(set! exiting? #t)
		(send frame on-quit))))]
     [check-break
      (lambda ()
	(and break-ok?
	     (begin0
	      break-flag
	      (if break-flag
		  (set! break-ok? #f)))))]
     [reset-break 
      (lambda () 
	(set! break-ok? #t)
	(set! break-flag #f))]

     [busy? #f]
     [busy-sema (make-semaphore 1)]

     [initialized? #f]
     
     waiting-for-loaded
     [load-success? #f]

     [zodiac-reply (make-semaphore)]

     [set-exported-eval
      (lambda (eval)
	(set! exported-eval eval))]

     [exported-print-graph print-graph]
     [set-exported-print-graph
      (lambda (p-g)
	(set! exported-print-graph p-g))]
     [exported-print-struct print-struct]
     [set-exported-print-struct
      (lambda (p-s)
	(set! exported-print-struct p-s))]

     [registry (make-object drscheme:process-buffer-registry%)])

    (public
     [scheme-type 'std-nc]
     [use-aries? #t]
     [use-rice? #t]
     [constant-globals? #t]
     [set!-undefined? #f]

     process
     [break (lambda () (set! break-flag #t))]
     [send-scheme
      (lambda (expr)
	(set! to-be-evaluated expr)
	(set! return-error #f)
	(semaphore-post eval-sema)
	(semaphore-wait return-sema))]
	   
     exported-eval 

     [eval-busy? (lambda () busy?)]
     [pretty-print-out
      (lambda (v)
	(pretty-print v this-out #f 
		      (exported-print-graph)
		      (exported-print-struct)))]
     [do-eval
      (lambda (str)
	(if initialized?
	    (let ([go 
		   (begin
		     (semaphore-wait busy-sema)
		     (begin0
		      (if busy?
			  #f
			  (begin
			    (set! busy? #t)
			    #t))
		      (semaphore-post busy-sema)))]) 
	      (if go
		  (thread
		   (lambda ()
		     (mred:local-busy-cursor
		      (get-canvas)
		      (lambda ()
			(do-pre-eval)
			(send-scheme 
			 (if (not use-aries?)
			     `(begin ,@(read-string-all str))
			     `(aries:transform (#%open-input-string 
						(#%string-append
						 ,str
						 (#%string #\newline)))
					       ,(- (last-position)
						   (string-length str)
						   1)
					       #t)))
			(unless return-error
				(display-result return-value))
			(do-post-eval)
			(set! busy? #f)))))))))]
     [shut-down
      (lambda ()
	(when (and initialized?
		   (not exiting?))
	   (break)
	   (begin
	     (set! exiting? #t)
	     (send-scheme '(#%exit)))))])

    (public
     [load-non-aries
      (lambda (src)
	(if (string? src)
	    (send-scheme `(#%load ,src))
	    (send-scheme `(let ([p ,(mred:open-input-buffer src)])
			    (let loop ()
			      (let ([e (#%read p)])
				(#%unless (#%eof-object? e)
					  (#%eval e)
					  (loop))))))))]

     [load-mrslatex
       (lambda (src id)
	 (let ((src (if (string? src) src
		      (mred:open-input-buffer src))))
	   (send-scheme `(mrslatex:transform ,src ,(get-frame)))))]

     [load-aries
      (lambda (src id)
	(let ([src (if (string? src)
		       src
		       (mred:open-input-buffer src))])
	  (send-scheme `(aries:transform ,src (quote ,id) #f))))])

    (private

     [slatex-thing
       (lambda (buffer file frame-group)
	 (unless busy?
	   (set! busy? #t)
	   (let* ((read-from (if buffer buffer file))
		   (id (send registry make-aries-id buffer file frame-group)))
	     (set! waiting-for-loaded id)
	     (set! load-success? #f)
	     (load-mrslatex read-from id)
	     (semaphore-wait zodiac-reply))
	   (set! busy? #f)))]

     [load-thing-into-scheme
      (lambda (buffer file frame-group)
	(if (not busy?)
	    (let ([read-from
		   (if buffer
		       buffer
		       file)])
	      (set! busy? #t)
	      (if (not use-aries?)
		  (load-non-aries read-from)
		  (let ([id (send registry make-aries-id 
				  buffer file frame-group)])
		    (set! waiting-for-loaded id)
		    (set! load-success? #f)
		    (load-aries read-from id)
		    (semaphore-wait zodiac-reply)))
	      (when (and (not return-error)
			 (not use-aries?))
		    (ready-non-prompt)	
		    (fprintf this-out "Executed ~s." 
			     (if buffer
				 (send buffer get-filename)
				 file)))
	      (if autoprompting?
		  (insert-prompt))
	      (if use-aries?
		  load-success?
		  (not return-error))
	      (set! busy? #f))))])
     
    (public
     [slatex-file
       (lambda (file frame-group)
	 (slatex-thing #f file frame-group))]
     [slatex-buffer
       (lambda (buffer frame-group)
	 (slatex-thing buffer #f frame-group))]
     [load-file-into-scheme
      (lambda (file frame-group)
	(load-thing-into-scheme #f file frame-group))]
     [load-buffer-into-scheme
      (lambda (buffer frame-group)
	(load-thing-into-scheme buffer #f frame-group))]

     [zodiac-transmit-to-device
      (lambda (tag . args)
	(semaphore-post zodiac-reply)
	(case tag
	  [(dynamic-error static-error)
	   (apply
	    (lambda (string line column offset file . extras)
	      (fprintf this-err
		       "~a in ~s, line ~s~n"
		       string (if (number? file)
				  "console"
				  (send registry file-name-from-id file)) 
		       (+ line
			  (if (number? file) 
			      (position-line file)
			      0)))
	      (unless (or (null? extras) (null? (car extras)))
		(fprintf this-err " provided:")
		(let loop ([extras (car extras)])
                  (unless (null? extras)
		    (fprintf this-err " ~s" (car extras))
		    (loop (cdr extras))))
		(newline this-err))
	      (if (number? file)
		  (begin
		    (begin-edit-sequence)
		    (set-position (+ file offset))
		    (send registry select-s-expression this)
		    (scroll-to-position (last-position))
		    (end-edit-sequence))
		  (send registry hilite-in-id file offset)))
	    args)]
	  [result
	   (apply
	    (lambda (file value)
	      (display-result (car value)))
	    args)]
	  [loaded
	   (apply
	    (lambda (file empty-list)
	      (if (eq? file waiting-for-loaded)
		  (set! load-success? #t))
	      (fprintf this-out "Executed ~s.~n" 
		       (send registry file-name-from-id file)))
	    args)]
	  [break
	   (void)]))]
	   
     [initialize-console
      (lambda ()
	(super-initialize-console)
	(set! process
	      ((make-spawner (let ([l (list
				       (if constant-globals?
					   'constants
					   'no-constants)
				       (if (or use-rice? set!-undefined?)
					   'set!-undefined
					   'no-set!-undefined))])
			       (if (eq? scheme-type 'std)
				   l
				   (cons 'wx l)))
			     (eval-sema
			      get-input send-error send-result
			      stdout stderr stdin on-exit 
			      reset-break check-break
			      export-eval export-print-graph
			      export-print-struct)
		(export-eval eval)
		(export-print-graph print-graph)
		(export-print-struct print-struct)
 	        (let ([exit-process #f])
  	         (exit-handler
		  (lambda (arg)
		    (on-exit arg)
		    (send-result 'done-not-coming-back-done-will-be-fixed-eventually-done)
		    (exit-process #f)))
		 (user-break-poll-handler check-break)
		 (break-enabled #t)
		 (current-output-port stdout)
		 (current-error-port stderr)
		 (current-input-port stdin)
		 (catch exit-loop
		    (set! exit-process exit-loop)
		    ;; In this loop, the user can (possibly) re-bind globals
		    ;;  Use #% for builtin primitive procedures
  		    (let loop ()
		      (#%semaphore-wait eval-sema)
		      (let ([orig-handler (#%error-escape-handler)])
			(#%catch escape
			  (dynamic-wind
			   (lambda ()
			     (#%error-escape-handler 
			      (lambda ()
				(send-error #t)
				(escape #f)))
			     (if (#%defined? 'zodiac:top-level-contn)
				 (zodiac:top-level-contn
				  (lambda (x)
				    (send-error #t)
				    (escape x)))))
			   (lambda ()
			     (reset-break)
			     (send-result (#%eval (get-input))))
			   (lambda ()
			     (#%error-escape-handler orig-handler)))))
		      (loop)))))
	       eval-sema
	       get-input
	       set-error
	       set-result
	       this-out
	       this-err
	       this-in
	       on-exit
	       reset-break
	       check-break
	       set-exported-eval
	       set-exported-print-graph
	       set-exported-print-struct))
	
	(when use-aries?
	 (send-scheme `(current-library-path ,(current-library-path)))
	 (send-scheme `(require-library "core.ss"))

	 (send-scheme `(define zodiac:home-directory 
			 ,drscheme:zodiac-directory))
	 (send-scheme `(load ,(build-absolute-path drscheme:zodiac-directory
					     "load.ss")))

	 (send-scheme `(zodiac:require (build-relative-path "Source" "do-load")))
	 (send-scheme `(fetch "mzscheme"))
	 (send-scheme '(aries))
	 (send-scheme `(load ,(build-absolute-path
				drscheme:mrslatex-directory
				"mrslatex.ss")))
	 (when use-rice?
	   (send-scheme '(rice))
	   (send-scheme `(load ,(build-absolute-path drscheme:rice-directory
						     "ricedefs.ss")))
	   (send-scheme `(load ,(build-absolute-path drscheme:rice-directory
						     "x-rice.ss"))))

	 (send-scheme `(define zodiac:transmit-to-device
			 ,zodiac-transmit-to-device)))
	(unless (eq? scheme-type 'std)
	 (when use-rice?
	       (send-scheme `(load ,(build-absolute-path drscheme:rice-directory
							 "graphics.ss"))))

	 (unless (eq? scheme-type 'wx)
	    (send-scheme
	     `(begin
		(let ([old-dir (current-directory)])
		  (current-directory ,mred:system-source-directory)
		  (load "mrsystem.ss")
		  (current-directory old-dir))))))
	(when (and use-aries? #f) ; #f until zodiac is reasonably fast
	 (send-scheme '(define primitive-eval
			 (let ([orig-eval (current-eval)])
			   (lambda (expr)
			     (orig-eval expr)))))
	 (send-scheme '(define primitive-load
			 (let ([orig-eval (current-eval)])
			   (lambda (expr)
			     (let ([curr-eval (current-eval)])
			       (dynamic-wind
				(lambda () (current-eval orig-eval))
				(lambda () (load expr))
				(lambda () (current-eval curr-eval))))))))
	 (send-scheme '(current-eval
			(let ([orig-eval (current-eval)])
			  (lambda (expr)
			    (orig-eval (zodiac:expr-stripper 
					(zodiac:expand expr))))))))
	(set! initialized? #t))]))))
(define drscheme:spawned-process-console-edit%
  (drscheme:make-spawned-process-console-edit% mred:console-edit%))
    
(define drscheme:make-spawned-process-console-frame%
 (lambda (super%)
  (class super% ([frame-scheme-type 'std] [frame-use-aries? #t]
					  [frame-use-rice? #t]
					  [frame-constant-globals? #t]
					  [frame-set!-undefined? #f])
     (inherit active-edit show on-close make-menu active-canvas
	      title-prefix set-title)
     (rename [super-get-edit% get-edit%]
	     [super-make-menu-bar make-menu-bar]
	     [super-set-title-prefix set-title-prefix])
     (private
      [quitting? #f])
     (public 
      [running? #t]
      [on-quit (lambda () (shut-down))]
      [quit-menu-label "Close Console"]
      [frame-title "Console"]
      [set-title-prefix
       (lambda (s)
	 (super-set-title-prefix s)
	 (set-title
	  (string-append title-prefix "Console")))]
      [edit% drscheme:spawned-process-console-edit%]
      [get-edit%
       (lambda ()
         (class-asi (super-get-edit%)
	    (public
	     [scheme-type frame-scheme-type]
	     [use-aries? frame-use-aries?]
	     [use-rice? frame-use-rice?]
	     [constant-globals? frame-constant-globals?]
	     [set!-undefined? frame-set!-undefined?])))]

      [make-menu-bar
       (lambda ()
	 (let ([mb (super-make-menu-bar)]
	       [break-menu (make-menu)])
	   (send break-menu append-item "Break"
		 (lambda () (break)))
	   (send mb append break-menu "Break")
	   mb))]

      [shut-down
       (lambda ()
	 (unless quitting?
	     (set! quitting? #t)
	     (on-close)
	     (send (active-edit) shut-down)
	     (set! running? #f)
	     (show #f)))]

      [slatex-file
       (opt-lambda (filename [frame-group #f])
	 (if running?
	     (mred:local-busy-cursor
	      (active-canvas)
	      (lambda () 
		(send (active-edit) slatex-file filename frame-group)))))]
      [load-file-into-scheme
       (opt-lambda (filename [frame-group #f])
	 (if running?
	     (mred:local-busy-cursor
	      (active-canvas)
	      (lambda () 
		(send (active-edit) load-file-into-scheme filename frame-group)))))]
      [load-file 
       ;; Menu item callback
       (lambda (filename)
	 (thread (lambda ()
		   (load-file-into-scheme filename))))]
      [slatex-buffer
       (opt-lambda (buffer [frame-group #f])
	 (if running?
	     (mred:local-busy-cursor
	      (active-canvas)
	      (lambda () 
		(send (active-edit) slatex-buffer buffer frame-group)))))]
      [load-buffer-into-scheme
       (opt-lambda (buffer [frame-group #f])
	 (if running?
	     (mred:local-busy-cursor
	      (active-canvas)
	      (lambda () 
		(send (active-edit) load-buffer-into-scheme buffer frame-group)))))]
      [break
       (lambda () (send (active-edit) break))])
     (sequence
       (super-init #f (string-append
		       "Welcome to "
		       (case scheme-type
			 (std "MzScheme")
			 (wx "wxScheme")
			 (mred "MrEd")
			 (else (error 'spawned-console 
				      "invalid Scheme type: ~s"
				      scheme-type)))
		       " again."))
       (let ([edit (send this active-edit)])
	 (send edit release-output))))))
(define drscheme:spawned-process-console-frame%
  (drscheme:make-spawned-process-console-frame% mred:console-frame%))

(define drscheme:spawn-console
  (lambda args
    (thread 
     (lambda ()
       (apply make-object drscheme:spawned-process-console-frame% args)))))
