
;;; Define the standard editing window class

(define-sigfunctor (mred:frame@ mred:frame^)
  (import mred:edit^ mred:canvas^ mred:icon^ mred:menu^ 
	  mred:group^ mred:finder^ mred:handler^
	  mred:exit^ mred:autosave^ mred:gui-utils^
	  mzlib:function^ mzlib:file^)

  (define frame-name "MrEd")

  (define empty-frame%
    (class-asi wx:frame%
      (public
       [on-close (lambda () #t)])))

  (define frame-width 600)
  (define frame-height 600)
  (let ([w (box 0)]
	[h (box 0)])
    (wx:display-size w h)
    (if (< (unbox w) frame-width)
	(set! frame-width (- (unbox w) 50)))
    (if (< (unbox h) frame-height)
	(set! frame-height (- (unbox h) 50))))

  ; This defines a simple editing frame without a menu bar.
  ; On initialization, it doesn't automatically show itself
  (define make-simple-frame%
    (lambda (super%)
      (class super% ([name frame-name])
        (inherit get-client-size get-title set-title set-icon)
	(rename [super-on-size on-size])
	; Overrideable constants
	(public
	 [WIDTH frame-width]
	 [HEIGHT frame-height]
	 [MARGIN 2]
      
	 [TOP-MARGIN 0]
	 [BOTTOM-MARGIN 0]
	 [LEFT-MARGIN 0]
	 [RIGHT-MARGIN 0])
	(public
	 [edit% mred:edit^:edit%]
	 [canvas% mred:canvas^:simple-frame-canvas%])
	(public
	 [set-last-focus-canvas
	  (lambda (c)
	    (set! last-focus-canvas c))]
	 [title-prefix ""])
	
	; we need to call this one directly
	(public
	 [on-simple-size 
	  ; Window resize -> resize canvases
	  (lambda ()
	    (if (not (void? canvas))
		(let ((client-w (box 0))
		      (client-h (box 0)))
		  (get-client-size client-w client-h)
		  (let* ([w (- (unbox client-w) (* 2 MARGIN)
			       LEFT-MARGIN RIGHT-MARGIN)]
			 [h (- (unbox client-h) (* 2 MARGIN)
			       TOP-MARGIN BOTTOM-MARGIN)]
			 [count (length canvases)]
			 [h (if (zero? count) 10 (floor (/ h count)))])
		    (let loop ([canvases canvases]
			       [v (+ MARGIN TOP-MARGIN)])
		      (unless (null? canvases)
			      (let* ([canvas (car canvases)]
				     [panel (if (ivar canvas use-panel?)
						(ivar canvas panel)
						#f)]
				     [panel-w-box (box 0)]
				     [panel-h-box (box 0)])
				(when panel
				      (send panel set-size 
					    (+ MARGIN LEFT-MARGIN) v
					    w -1)
				      (send panel get-size
					    panel-w-box panel-h-box))
				(send canvas set-size 
				      (+ MARGIN LEFT-MARGIN) 
				      (+ v (unbox panel-h-box))
				      w (- h (unbox panel-h-box)))
				(let ([edit (send canvas get-media)])
				  (if (and (not (null? edit))
					   (ivar edit auto-set-wrap?))
				      (let ([admin (send edit get-admin)]
					    [w-box (box 0)]
					    [h-box (box 0)])
					(unless (null? admin)
						(send admin get-view 
						      () () w-box h-box #f)
						(send edit set-max-width 
						      (unbox w-box)))))))
			      (loop (cdr canvases) (+ v h))))))))])

	; methods
	(public
	 [on-frame-active (lambda () (void))]
	 [set-title-prefix
	  (lambda (s)
	    (if (string? s)
		(set! title-prefix s)))]
	 [get-canvas% (lambda () canvas%)]
	 [make-canvas
	  (lambda ()
	    (let ([canvas (make-object (get-canvas%) this)])
	      (when (ivar canvas use-panel?)
		    (send canvas create-panel this))
	      canvas))]
	 [get-edit% (lambda () edit%)]
	 [make-edit
	  (lambda ()
	    (make-object (get-edit%)))]
	 [on-size
	  (lambda (w h)
	    (super-on-size w h)
	    (on-simple-size))]
	 
	 [add-canvas
	  (opt-lambda ([canvas (make-canvas)][prefix? #f])
		      (unless (member canvas canvases)
			      (set! canvases 
				    (if prefix?
					(cons canvas canvases)
					(append canvases (list canvas))))
			      (send canvas set-frame this)
			      (on-simple-size))
		      canvas)]
	 [remove-canvas
	  (lambda (canvas)
	    (let ([pre-len (length canvases)])
	      (set! canvases
		    (mzlib:function^:remove 
		     canvas canvases
		     (lambda (canvas c)
		       (if (eq? c canvas)
			   (begin
			     (send canvas set-frame #f)
			     (send canvas set-size 50000 50000 10 10)
			     (when (ivar canvas use-panel?)
				   (send (ivar canvas panel)
					 set-size 50000 50000 -1 -1))
			     (if (eq? canvas last-focus-canvas)
				 (set! last-focus-canvas #f))
			     #t)
			   #f))))
	      (if (< (length canvases) pre-len)
		  (on-simple-size))))]
	 [active-canvas
	  (lambda ()
	    (let loop ([canvases canvases])
	      (cond
	       [(null? canvases) last-focus-canvas]
	       [(send (car canvases) is-focus-on?) (car canvases)]
	       [else (loop (cdr canvases))])))]
	 [active-edit
	  (lambda ()
	    (send (active-canvas) get-media))])
	
	; Initialization:
	(sequence
	  (super-init () name -1 -1 WIDTH HEIGHT
		      (+ wx:const-default-frame wx:const-sdi)
		      name))
	(public
	 [last-focus-canvas #f] ; Does this need to be inited during make-canvas?
	 [canvas (make-canvas)])
	(sequence
	  (if (null? (send canvas get-media))
	      (send canvas set-media (make-edit)))
	  
	  (send canvas set-frame this)
	  (set! last-focus-canvas canvas))
	(public
	 [edit (send canvas get-media)]
	 [canvases (list canvas)])
	(sequence
	  (if (send mred:icon^:icon ok?)
	      (set-icon mred:icon^:icon))
	  (on-simple-size)))))

  (define simple-frame% (make-simple-frame% empty-frame%))

  (define make-menu-frame%
    (lambda (super%)
      (class super% args
        (inherit set-menu-bar get-menu-bar)
	(public
	 [menu% mred:menu^:menu%]
	 [menu-bar% mred:menu^:menu-bar%])
	(public
	 [get-menu-bar% (lambda () menu-bar%)]
	 [get-menu% (lambda () menu%)]
	 [next-menu-id (lambda () 0)]
	 [make-menu-bar
	  (lambda ()
	    (make-object (get-menu-bar%)))]
	 [make-menu
	  (lambda ()
	    (make-object (get-menu%)))]
	 [on-menu-command
	  (lambda (op)
	    (send (get-menu-bar) dispatch op))])
	(sequence
	  (apply super-init args)

	  ; Build and install the menu bar:
	  (let ([menu-bar (make-menu-bar)])
	    (set-menu-bar menu-bar))))))

  (define menu-frame% (make-menu-frame% empty-frame%))

  (define simple-menu-frame%
    (make-menu-frame% simple-frame%))

  ; This defines the standard editing window.
  (define make-frame%
    (lambda (super%)
      (class super% ([filename #f][show? #t][frameset mred:group^:frames])
	(inherit active-edit active-canvas make-menu make-edit
		 show add-canvas remove-canvas canvases on-simple-size)
	(rename [super-on-close on-close]
		[super-make-menu-bar make-menu-bar]
		[super-on-menu-command on-menu-command]
		[super-next-menu-id next-menu-id])
	(public
	 [allow-split? #t])
	(private 
	 [edit-offset 0]
	 [font-offset 0]
	 [other-offset 0]
	 [keep-buffers? (is-a? frameset mred:group^:frame-group%)]
	 
	 [frames frameset]
	 [buffers (if keep-buffers? (ivar frames buffers))]
	 
	 [save-as
	  (lambda (format)
	    (let ([file (mred:finder^:put-file)])
	      (if file
		  (send edit save-file file format))))])
      
	(public
	 [auto-save? #t]
	 [on-frame-active
	  (lambda ()
	    (if keep-buffers?
		(send frames set-active-frame this)))]
	 [check-saved
	  (opt-lambda (canvas [reason "Close"])
		      (let* ([edit (send canvas get-media)]
			     [name (send edit get-filename)]
			     [name (if (string? name)
				       name
				       "Untitled")])
			(or (not (send edit modified?))
			    (let ([action
				   (mred:gui-utils^:unsaved-warning name reason #t)])
			      (case action
				[save (send edit save-file)]
				[continue (send edit remove-autosave) #t]
				[else #f])))))]
	 [get-frame-group
	  (lambda ()
	    (if keep-buffers?
		frames
		#f))]
	 [next-menu-id
	  (lambda ()
	    other-offset)]
	 [open-file
	  (opt-lambda (orig-filename [canvas (active-canvas)]
				     [check-save? (not keep-buffers?)])
	    ; filename = () => ask user
	    ; filename = #f => no file, make untitled
	    ; filename = <buffer> => use buffer
	    ; otherwise, filename = name string
	    (if (or (not check-save?)
		    (check-saved canvas))
		(let* ([filename
			(if (null? orig-filename)
			    (mred:finder^:get-file)
			    orig-filename)]
		       ; at this point, name is either a string, buffer, or #f
		       ; Is the buffer going to be untitled?
		       [untitled? (not filename)]
		       ; If filename is still #f, make up a name
		       [filename 
			(cond
			 ((object? filename)
			  (let ([name (send filename get-filename)])
			    (if (null? name)
				"Untitled"
				name)))
			 (filename filename)
			 (else (if keep-buffers?
				   (send buffers 
					 make-untitled-name)
				   "Untitled")))]
		       ; Look for an existing buffer for this file
		       [edit
			(cond
			 ((object? orig-filename) orig-filename)
			 ((or untitled? (not keep-buffers?))
			  #f)
			 (else
			  (send buffers find-buffer-by-name 'file filename)))])
		  ; If we didn't find a buffer, create one
		  (if (not edit)
		      (let ([edit (make-edit)])
			; Load in the file, if it exists
			(if (and (not untitled?)
				 (file-exists? filename))
			    (send edit load-file filename)
			    (begin
			      (send edit erase)
			      (send edit set-filename filename 
				    untitled?)))
			; Pick the editing mode
			(let* ([filename (send edit get-filename)]
			       [mode-handler (if (null? filename)
						 #f
						 (mred:handler^:find-mode-handler
						  filename))])
			  (if mode-handler
			      (let ([mode (mode-handler edit)])
				(send edit set-mode mode))))
			; We created this buffer; add to the global list
			(if keep-buffers?
			    (send buffers add-buffer 'file #f edit))
			(send canvas set-media edit))
		      ; Found an existing edit
		      (begin
			(if keep-buffers?
			    (send buffers reset-buffer-state edit))
			(send canvas set-media edit)))
		  #t)
		#f))]
	 [on-menu-command
	  ; Handle menu selections
	  (lambda (op)
	    (cond
	     [(< op font-offset)
	      (send (active-edit) do-edit (- op edit-offset))
	      #t]
	     [(< op other-offset)
	      (send (active-edit) do-font (- op font-offset))
	      #t]
	     [else (super-on-menu-command op)]))]
	 
	 [check-nonunique-or-saved
	  (lambda (canvas)
	    (let ([e (send canvas get-media)])
	      (if (ormap
		   (lambda (c)
		     (and (not (eq? c canvas))
			  (eq? e (send c get-media))))
		   canvases)
		  #t
		  (check-saved canvas))))]

	 [check-all-saved-for-quit
	  (lambda () 
	    (check-all-saved "Quit"))]
	 [check-all-saved
	  (opt-lambda ([reason "Close"])
		      (andmap (lambda (c) (check-saved c reason)) canvases))]
	 
	 [on-close
	  (lambda ()
	    (if (and (super-on-close)
		     (or (not keep-buffers?)
			 (send frames remove-frame this)))
		(if keep-buffers?
		    (begin
		      (for-each (lambda (canvas)
				  (send canvas set-media '()))
				canvases)
		      #t)
		    (if (check-all-saved)
			(begin
			  (mred:exit^:remove-exit-callback
			   exit-callback-tag)
			  (set! auto-save? #f)
			  #t)
			#f))
		#f))]
	 
	 [do-autosave
	  (lambda ()
	    (if auto-save?
		(for-each (lambda (canvas)
			    (let ([m (send canvas get-media)])
			      (unless (null? m)
				      (send m do-autosave))))
			  canvases)))]
	 [make-menu-bar
	  (lambda ()
	    (let ([file-menu (make-menu)]
		  [edit-menu (make-menu)])
	      (send file-menu append-item "Open..."
		    (lambda () 		
		      (let ((file (mred:finder^:get-file)))
			(if file
			    (open-file file)))))
	      (send file-menu append-item "New File..."
		    (lambda ()
		      (let ((file (mred:finder^:put-file)))
			(if file
			    (open-file file)))))

	      (if keep-buffers?
		  (send file-menu append-item "Switch to..."
			(lambda () (send buffers pick (active-canvas)))))
	      (send file-menu append-item "Revert"
		    (lambda () (send (active-edit) load-file)))
	      (send file-menu append-separator)
	      (send file-menu append-item "New Frame"
		    (lambda () 
		      (if keep-buffers?
			  (send frames new-frame #f)
			  (mred:handler^:edit-file #f))))
	      (send file-menu append-separator)
	      (send file-menu append-item "Save"
		    (lambda () (send (active-edit) save-file)))
	      (send file-menu append-item "Save As..."
		    (lambda () (save-as wx:const-media-ff-same)))
	      (send file-menu append-item "Save As Text..."
		    (lambda () (save-as wx:const-media-ff-text)))
	      (send file-menu append-item "Save As Text and Styles..."
		    (lambda () (save-as wx:const-media-ff-std)))
	      (send file-menu append-item "Print..."
		    (lambda () (send (active-edit) print '())))
	      (when allow-split?
		    (send file-menu append-separator)
		    (send file-menu append-item "Split"
			  (lambda () 	       
			    (let ([new-canvas (add-canvas)])
			      (send new-canvas set-media (active-edit)))))
		    (send file-menu append-item "Collapse"
			  (lambda ()
			    (when (> (length canvases) 1)
				  (let ([canvas (active-canvas)])
				    (if (and canvas
					     (or keep-buffers?
						 (check-nonunique-or-saved canvas)))
					(remove-canvas canvas)))))))
	      (send file-menu append-item "Set Mode..."
		    (lambda ()
		      (let* ([modes (map mred:handler^:handler-name 
					 mred:handler^:mode-handlers)]
			     [name (wx:get-single-choice
				    "Select a Mode" "Mode"
				    modes)])
			(unless (null? name)
				(let* ([handler (mred:handler^:find-named-mode-handler name)]
				       [mode (handler (active-edit))])
				  (send (active-edit) set-mode mode))))))
	      (send file-menu append-separator)
	      (send file-menu append-item "Close"
		    (lambda () (if (on-close) (show #f))))
	      
	      ; Use the built-in menu item lists for the edit menu:
	      (set! edit-offset (super-next-menu-id))
	      (set! font-offset
		    (send (active-edit) append-edit-items 
			  edit-menu edit-offset))
	      (send edit-menu append-separator)
	      (send edit-menu append-item "Toggle Wrap Text"
		    (lambda ()
		      (let ([edit (active-edit)])
			(send edit set-auto-set-wrap
			      (not (ivar edit auto-set-wrap?)))
			(on-simple-size))))
	      (send edit-menu append-separator)
	      (set! other-offset
		    (send (active-edit) append-font-items 
			  edit-menu font-offset))
	      
	      (let ([mb (super-make-menu-bar)])
		(send mb append file-menu "File")
		(send mb append edit-menu "Edit")
		mb)))])

	(sequence
	  (super-init))
	
	(public
	 [exit-callback-tag
	  (if (not keep-buffers?)
	      (mred:exit^:insert-exit-callback check-all-saved-for-quit))])
	
	(sequence
	  (when show? (show #t))
	
	  (if keep-buffers?
	      (send frames insert-frame this)
	      (mred:autosave^:register-autosave this))
	  
	  (let ([filename (if (string? filename)
			      (mzlib:file^:normalize-path filename)
			      filename)])
	    (open-file filename))
	  
	  (when show?
		(wx:yield)
		(send (active-canvas) set-focus)))
	(public
	 [edit (active-edit)]))))

  (define frame% (make-frame% simple-menu-frame%))

  (define make-pasteboard-frame%
    (lambda (super%)
      (class-asi super%
        (public
	 [edit% mred:edit^:pasteboard%]))))
  (define pasteboard-frame% (make-pasteboard-frame% frame%))

  (define make-status-frame%
    (lambda (super%)
      (class-asi super%
       (rename [super-get-canvas% get-canvas%])
       (public
	[get-canvas%
	 (lambda ()
	   (class-asi super-get-canvas%
	     (public [use-panel? #t])))]))))
  )

