; File Formats and Modes

(define-sigfunctor (mred:handler@ mred:handler^)
  (import mred:group^ mred:frame^ mred:finder^ mzlib:file^)

  (define-struct handler (name extension handler))

  (define format-handlers '())
  (define mode-handlers '())

  ; Inserting format & mode handlers
  (define make-insert-handler
    (letrec ([string-list?
              (lambda (l)
                (cond
		 [(null? l) #t]
		 [(not (pair? l)) #f]
		 [else
		  (and (string? (car l))
		       (string-list? (cdr l)))]))])
      (lambda (who name extension handler)
        (cond
	 [(not (string? name))
	  (error who "name was not a string")]
	 [(and (not (procedure? extension))
	       (not (string? extension))
	       (not (string-list? extension)))
	  (error who
		 "extension was not a string, list of strings, or a predicate")]
	 [(not (procedure? handler))
	  (error who "handler was not a function")]
	 [else (make-handler name
			     extension
			     handler)]))))
  
  (define insert-format-handler
    (lambda args
      (set! format-handlers 
            (cons (apply make-insert-handler 'insert-format-handler args)
                  format-handlers))))
  
  (define insert-mode-handler
    (lambda args
      (set! mode-handlers 
            (cons (apply make-insert-handler 'insert-mode-handler args)
                  mode-handlers))))

  ; The default mode: 
  (insert-mode-handler "None" () (lambda args #f))

  ; Finding format & mode handlers
  (define find-handler
    (lambda (name handlers)
      (catch exit
        (let ([extension (if (string? name)
			     (mzlib:file^:filename-extension name)
			     #f)])
	  (if extension
	      (for-each
	       (lambda (handler)
		 (let ([ext (handler-extension handler)])
		   (if (or (and (procedure? ext)
				(ext name ext))
			   (and (string? ext)
				(string=? ext extension))
			   (and (pair? ext)
				(ormap (lambda (ext) 
					 (string=? ext extension))
				       ext)))
		       (exit (handler-handler handler)))))
	       handlers))
	  #f))))
  
  (define find-format-handler
    (lambda (name)
      (find-handler name format-handlers)))
  
  (define find-mode-handler
    (lambda (name)
      (find-handler name mode-handlers)))

  ; Finding format & mode handlers by name
  (define find-named-handler
    (lambda (name handlers)
      (let loop ([l handlers])
        (cond
          [(null? l) #f]
          [(string-ci=? (handler-name (car l)) name)
           (handler-handler (car l))]
          [else (loop (cdr l))]))))
  
  (define find-named-format-handler
    (lambda (name)
      (find-named-handler name format-handlers)))
  
  (define find-named-mode-handler
    (lambda (name)
      (find-named-handler name mode-handlers)))

  ; Open a file for editing
  (define edit-file
    (opt-lambda (filename [group mred:group^:frames] [group-handles? #t] 
			  [make-default (lambda (filename group)
					  (make-object mred:frame^:frame% filename #t group))])
       (if (and group group-handles?)
	   (send group open-file filename)
	   (let ([handler
		  (if (string? filename)
		      (find-format-handler filename)
		      #f)])
	     (if handler
		 (handler filename group)
		 (make-default filename group))))))
  
  ; Query the user for a file and then edit it
  (define open-file
    (opt-lambda ([frames mred:group^:frames])
       (let ((file (mred:finder^:get-file)))
	 (if file
	     (edit-file file frames)))))
  )


