; Always load the standard system with load/cd

; Print a little more than MzScheme automatically does:
(error-print-width 100)

; Remember this directory
(define mred:system-source-directory
  (current-directory))
(constant-name 'mred:system-source-directory)

(unless (defined? 'mred:startup-print-status)
   (define mred:startup-print-status 
     (if (and (defined? 'mred:verbose-load?)
	      mred:verbose-load?)
	 (lambda (x) (display x) (newline))
	 (lambda (x) #f))))

(mred:startup-print-status "Loading mzlib...")
(let ([libdir
       (let ([try-dir (build-path (current-directory) 'up "mzlib")])
	 (if (directory-exists? try-dir)
	     try-dir
	     (let ([try-dir (build-path (current-directory) 'up 
						 "mzscheme" "mzlib")])
	       (if (directory-exists? try-dir)
		   try-dir
		   (let ([try-dir (build-path (current-directory) 'up 'up 
						       "mzscheme" "mzlib")])
		     (if (directory-exists? try-dir)
			 try-dir
			 (let* ([v (getenv "PLTHOME")]
				[dir (if v 
					 v 
					 (cond
					  [(eq? wx:platform 'unix)
					   "/usr/local/lib/plt"]
					  [(eq? mwx:platform 'windows)
					   "c:\\plt"]
					  [else ; macintosh
					   (current-directory)]))])
			   (build-path dir "mzscheme" "mzlib"))))))))])
  (current-library-path libdir))
(when (not (directory-exists? (current-library-path)))
    (wx:message-box "Cannot find the MzScheme libraries." "Error")
    (error 'mrsystem "cannot find the MzScheme libraries"))
(define mzlib:constant-lib? #t)
(require-library "core.ss")
(require-library "trigger.ss")

(define mred:plt-home-directory
  (if (defined? 'mred:plt-home-directory)
      (normalize-path mred:plt-home-directory)
      (let ([plt (getenv "PLTHOME")])
	(if plt
	    (normalize-path plt)
	(let-values ([(base name dir?) 
		      (split-path mred:system-source-directory)])
	  (if (string? base)
	      (let-values ([(base name dir?) (split-path base)])
		(if (string? base)
		    base
		    mred:system-source-directory))
	      mred:system-source-directory))))))
(constant mred:plt-home-directory)

(mred:startup-print-status "Loading signatures...")
(load "sig.ss")

(mred:startup-print-status "Loading autoload...")
(load "autoload.ss")
(mred:startup-print-status "Loading autosave...")
(load "autosave.ss")

(mred:startup-print-status "Loading exit...")
(load "exit.ss")
(mred:startup-print-status "Loading handler...")
(load "handler.ss")
(mred:startup-print-status "Loading mode...")
(load "mode.ss")
(mred:startup-print-status "Loading keys...")
(load "keys.ss")

(mred:startup-print-status "Loading match cache...")
(load "mcache.ss")
(mred:startup-print-status "Loading parens...")
(load "paren.ss")
(mred:startup-print-status "Loading Scheme parens...")
(load "sparen.ss")
; (mred:startup-print-status "Loading C++ parens...")
; (load "cparens.ss")

(mred:startup-print-status "Loading path utils...")
(load "fileutil.ss")

(mred:startup-print-status "Loading GUI utils...")
(load "guiutils.ss")

(mred:startup-print-status "Loading finder...")
(load "finder.ss")

(mred:startup-print-status "Loading icon...")
(load "icon.ss")

(mred:startup-print-status "Loading menu...")
(load "menu.ss")
(mred:startup-print-status "Loading edit...")
(load "edit.ss")
(mred:startup-print-status "Loading canvas...")
(load "canvas.ss")
(mred:startup-print-status "Loading frame...")
(load "frame.ss")
(mred:startup-print-status "Loading group...")
(load "group.ss")
(mred:startup-print-status "Loading project...")
(load "project.ss")

(mred:startup-print-status "Loading console...")
(load "console.ss")

(mred:startup-print-status "Loading find string...")
(load "findstr.ss")

(mred:startup-print-status "Loading Scheme mode...")
(load "ssmode.ss")
; (mred:startup-print-status "Loading C++ mode...")
; (load "cppmode.ss")

(mred:startup-print-status "Loading hyper...")
(load/cd (build-relative-path "hyper" "hyper.ss"))

(define-sigfunctor-set mred@@
  [mzlib:string@]
  [mzlib:file@ mzlib:string@]
  [mzlib:function@]
  [mzlib:pretty-print@]
  [mzlib:trigger@]
  [mred:autoload@ mzlib:file@]
  [mred:autosave@]
  [mred:exit@]
  [mred:mode@ mred:keymap@]
  [mred:handler@ mred:group@ mred:frame@ mred:finder@ mzlib:file@]
  [mred:keymap@ mred:finder@ mred:handler@ mred:find-string@ mred:scheme-paren@]
  [mred:match-cache@]
  [mred:paren@]
  [mred:scheme-paren@ mred:paren@]
  [mred:path-utils@]
  [mred:gui-utils@ mzlib:function@ mzlib:trigger@]
  [mred:finder@ mzlib:string@ mzlib:function@ mzlib:file@]
  [mred:icon@]
  [mred:menu@ mzlib:function@]
  [mred:edit@ mred:finder@ mred:path-utils@ mred:mode@ mred:scheme-paren@ 
	      mred:keymap@ mzlib:function@]
  [mred:canvas@ mred:edit@ mzlib:file@]
  [mred:frame@ mred:edit@ mred:canvas@ mred:icon@ mred:menu@ 
	       mred:group@ mred:finder@ mred:handler@
	       mred:exit@ mred:autosave@ mred:gui-utils@
	       mzlib:function@ mzlib:file@]
  [mred:group@ mred:gui-utils@ mred:exit@ mred:autosave@ mred:handler@
	       mzlib:function@]
  [mred:project@ mred:group@ mred:gui-utils@ mred:exit@ mred:finder@
		 mred:frame@ mred:handler@
		 mzlib:file@ mzlib:function@]
  [mred:console@ mred:edit@ mred:frame@ mred:exit@ mred:finder@ mred:handler@
		 mred:gui-utils@ mred:scheme-mode@ mred:scheme-paren@
		 mzlib:function@ mzlib:string@ mzlib:pretty-print@
		 mzlib:trigger@]
  [mred:find-string@ mred:canvas@ mred:edit@ mred:frame@]
  [mred:scheme-mode@ mred:mode@ mred:match-cache@ mred:paren@ mred:scheme-paren@ 
		     mred:icon@ mred:handler@ mred:keymap@
		     mzlib:string@]
  [mred:hyper-edit@ mred:edit@ mred:hyper-dialog@ mzlib:file@ mzlib:string@]
  [mred:hyper-dialog@ mred:hyper-edit@ mzlib:file@]
  [mred:hyper-frame@ mred:hyper-edit@ mred:hyper-dialog@
		     mred:frame@ mred:canvas@ mred:group@ mred:handler@])

; Used for running programs that don't use functors:
(define-signature mred:main^)
(define-compound-sigfunctor (mred:main@ mred:main^)
  (import)
  (with mred@@)
  (export mred@@))

(define mred:console-frame (void))
(define mred:console-canvas (void))
(define mred:console-edit (void))

(define mred:initialize
  (lambda args
    (if (and (pair? args)
	     (equal? (car args) "-i"))
	(begin
	  (eval-string (cadr args))
	  (apply mred:initialize (cddr args)))
	(invoke-open-functor (sigfunctor->functor mred:main@) mred))))

(define mred:startup
  (lambda args
    (if (null? args)
	(let ([frame (make-object mred:console-frame% #f)])
	  (set! mred:console-frame frame)
	  (set! mred:console-canvas (ivar frame canvas))
	  (set! mred:console-edit (ivar frame edit))
	  frame)
  	(let loop ([args args])
	  (if (null? args)
	      (mred:startup)
	      (let ([arg (car args)])
		(cond
		 [(string-ci=? "-f" arg)
		  (let loop ([args (cdr args)])
		    (cond
		     [(null? args) (mred:startup)]
		     [(string=? (car args) "--") 
		      (apply mred:startup (cdr args))]
		     [else
		      (load-with-cd (car args))
		      (loop (cdr args))]))]
		 [(string-ci=? "-e" arg)
		  (if (null? (cdr args))
		      (error "expected a string to evaluate after -e flag")
		      (eval-string (cadr args)))
		  (loop (cddr args))]
		 [(string-ci=? "--" arg)
		  (apply mred:startup (cdr args))]
		 [else (loop (cdr args))])))))))

(define mred:user-setup
  (lambda ()
    (let* ([home (if (eq? wx:platform 'unix)
		     (expand-path "~")
		     #f)]
	   [home (if (string? home)
		     (string-append home "/")
		     "")]
	   [file (string-append home 
				(if (eq? wx:platform 'windows)
				    "mredrc.ss"
				    ".mredrc"))])
      (if (file-exists? file)
	  (let ([orig-escape (error-escape-handler)])
	  (catch-errors
	   (lambda (s)
	     (wx:message-box s "Error"))
	   (lambda ()
	     (orig-escape))
	   (load-with-cd file)))))))

(if (eq? wx:platform 'unix)
    (let* ([default-path "/usr/local/transcript-4.0/lib/"]
	   [path-box (box default-path)])
      (wx:get-resource "MrEd" "afmPath" path-box)
      (wx:set-afm-path 
       (if (or (directory-exists? (unbox path-box))
	       (not (directory-exists? default-path)))
	   (unbox path-box)
	   default-path))))



