
(require-library "constant.ss")
(require-library "sfunctor.ss")
(require-library "strings.ss")

(define-signature mzlib:file^
   find-relative-path
   explode-path
   normalize-path
   build-absolute-path
   build-relative-path
   filename-extension
   file-name-from-path
   path-only)

(define-sigfunctor (mzlib:file@ mzlib:file^)
  (import mzlib:string^)

  (define build-relative-path
    (lambda (p . args)
      (if (relative-path? p)
	  (apply build-path p args)
	  (error 'build-relative-path "base path ~s is absolute" p))))

  (define build-absolute-path
    (lambda (p . args)
      (if (relative-path? p)
	  (error 'build-absolute-path "base path ~s is relative" p)
	  (apply build-path p args))))

  ; Note that normalize-path does not normalize the case
  (define normalize-path
    (letrec ([normalize-path
	      (case-lambda 
	       [(orig-path) (normalize-path orig-path (current-directory))]
	       [(orig-path wrt)
		(let normalize ([path (expand-path orig-path)])
		  (let-values ([(base name dir?) (split-path path)])
			      (cond
			       [(eq? name 'up)
				(let up ([base (if (eq? base 'relative)
						   wrt
						   (resolve-path base))])
				  (let-values ([(prev name dir?) (split-path base)])
					      (cond
					       [(not prev) (error 'normalize-path
								  "root has no parent directory: ~s"
								  orig-path)]
					       [else
						(let ([prev
						       (if (eq? prev 'relative)
							   wrt
							   (normalize prev))])
						  (cond
						   [(eq? name 'same) (up prev)]
						   [(eq? name 'up) (up (up prev))]
						   [else prev]))])))]
			       [(eq? name 'same)
				(cond
				 [(eq? base 'relative) wrt]
				 [else (normalize base)])]
			       [else
				(cond
				 [(not base) path]
				 [else (let* ([base (if (eq? base 'relative)
							(normalize wrt)
							(normalize base))]
					      [path (build-path base name)]
					      [resolved (expand-path (resolve-path path))])
					 (if (relative-path? resolved)
					     (normalize (build-path base resolved))
					     resolved))])])))])])
      normalize-path))

  ; Argument must be in normal form
  (define explode-path
    (lambda (orig-path)
      (let loop ([path orig-path][rest '()])
	(let-values ([(base name dir?) (split-path path)])
		    (if (or (and base
				 (not (string? base)))
			    (not (string? name)))
			(error 'explode-path "input was not in normal form: ~s" orig-path))
		    (if base
			(loop base (cons name rest))
			(cons name rest))))))

  ; Arguments must be in normal form
  (define find-relative-path
    (lambda (directory filename)
      (let ([dir (explode-path directory)]
	    [file (explode-path filename)])
	(if (not (string=? (car dir) (car file)))
	    filename
	    (let loop ([dir (cdr dir)][file (cdr file)])
	      (cond
	       [(null? dir) (apply build-path file)]
	       [(null? file) (apply build-path (map (lambda (x) 'up) dir))]
	       [(string=? (car dir) (car file))
		(loop (cdr dir) (cdr file))]
	       [else
		(apply build-path 
		       (append (map (lambda (x) 'up) dir)
			       file))]))))))

  (define file-name-from-path
    (lambda (name)
      (let-values ([(base file dir?) (split-path name)])
	(if (and (not dir?) (string? file))
	    file
	    #f))))

  (define path-only
    (lambda (name)
      (let-values ([(base file dir?) (split-path name)])
	 (cond
	  [dir? name]
	  [(string? base) base]
	  [else #f]))))

  ; name can be any string; we just look for a dot
  (define filename-extension
    (lambda (name)
      (let* ([len (string-length name)]
	     [extension
	      (let loop ([p (sub1 len)])
		(cond
		 [(negative? p) #f]
		 [(char=? (string-ref name p) #\.)
		  (substring name (add1 p) len)]
		 [else (loop (sub1 p))]))])
	(if extension
	    (mzlib:string^:string-lowercase! extension))
	extension))))

(let ()
  (define-compound-sigfunctor (linked@ mzlib:file^)
    (import)
    (with (mzlib:file@ mzlib:string@)
	  (mzlib:string@))
    (export mzlib:file@))

  (open-mzlib-functor linked@ (mzlib:file@)))

(mzlib-constant find-relative-path)
(mzlib-constant explode-path)
(mzlib-constant normalize-path)
(mzlib-constant build-absolute-path)
(mzlib-constant build-relative-path)
(mzlib-constant filename-extension)
(mzlib-constant file-name-from-path)
(mzlib-constant path-only)
