
(require-library "constant.ss")
(require-library "macros.ss")
(require-library "sfunctor.ss")

(define-signature mzlib:function^
  make-parameter
  dynamic-disable-break
  remv
  remq
  remove
  ignore-errors
  quicksort
  identity
  compose
  foldl
  foldr)

(define-sigfunctor (mzlib:function@ mzlib:function^)
  (import)

  (define identity (lambda (x) x))

  (define compose
    (case-lambda 
     [(f) f]
     [(f g)
      (if (eqv? 1 (arity f)) ; optimize: don't use call-w-values
	  (if (eqv? 1 (arity g)) ; optimize: single arity everywhere
	      (lambda (x) (f (g x)))
	      (lambda args (f (apply g args))))
	  (lambda args
	    (call-with-values
	     (lambda () (apply g args))
	     f)))]
     [(f . more)
      (let ([m (apply compose more)])
	(compose f m))]))

  (define quicksort
    (lambda (l less-than)
      (let* ([v (list->vector l)]
	     [count (vector-length v)])
	(let loop ([min 0][max count])
	  (if (< min (sub1 max))
	      (let ([pval (vector-ref v min)])
		(let pivot-loop ([pivot min]
				 [pos (add1 min)])
		  (if (< pos max)
		      (let ([cval (vector-ref v pos)])
			(if (less-than cval pval)
			    (begin
			      (vector-set! v pos (vector-ref v pivot))
			      (vector-set! v pivot cval)
			      (pivot-loop (add1 pivot) (add1 pos)))
			    (pivot-loop pivot (add1 pos))))
		      (if (= min pivot)
			  (loop (add1 pivot) max)
			  (begin
			    (loop min pivot)
			    (loop pivot max))))))))
	(vector->list v))))

  (define ignore-errors
    (lambda (thunk)
      (catch escape
	     (catch-errors
	      (lambda (s) (void))
	      (lambda () (void))
	      (thunk)))))

  (define remove
    (letrec ([rm (case-lambda 
		  [(item list) (rm item list equal?)]
		  [(item list equal?)
		   (let loop ([list list])
		     (cond
		      [(null? list) ()]
		      [(equal? item (car list)) (cdr list)]
		      [else (cons (car list)
				  (loop (cdr list)))]))])])
      rm))

  (define remq
    (lambda (item list)
      (remove item list eq?)))

  (define remv
    (lambda (item list)
      (remove item list eqv?)))

  (define dynamic-disable-break
    (lambda (thunk)
      (let ([v (break-enabled)])
	(if v
	    (dynamic-wind
	     (lambda ()
	       (break-enabled #f))
	     thunk
	     (lambda ()
	       (break-enabled #t)))
	    (thunk)))))

  ; make-parameter helps the programmer define state managers, like
  ; the built-in current-output-port, etc.
  (define make-parameter 
    (letrec ([mp
	      (case-lambda
	       [(initval) (mp initval (lambda (x) x))]
	       [(initval guard)
		(let ([val initval])
		  (case-lambda
		   [() val]
		   [(new-val) (set! val (guard new-val))]))])])
      mp))

  ;; fold : ((A -> B) B (listof A) -> B)
  ;; fold : ((A1 ... An -> B) B (listof A1) ... (listof An) -> B)

  ;; foldl builds "B" from the beginning of the list to the end of the
  ;; list and foldr builds the "B" from the end of the list to the
  ;; beginning of the list.

  (define mapadd
    (lambda (f l last)
      (letrec ((helper
		(lambda (l)
		  (cond
		   [(null? l) (list last)]
		   [else (cons (f (car l)) (helper (cdr l)))]))))
	(helper l))))

  (define foldl
    (letrec ((fold-one
	      (lambda (f init l)
		(letrec ((helper
			  (lambda (init l)
			    (cond
			     [(null? l) init]
			     [else (helper (f (car l) init) (cdr l))]))))
		  (helper init l))))
	     (fold-n
	      (lambda (f init . l)
		(cond
		 [(ormap null? l)
		  (if (andmap null? l) 
		      init
		      (error 'foldl "received non-equal length input lists"))]
		 [else (apply fold-n
			      f
			      (apply f (mapadd car l init))
			      (map cdr l))]))))
      (case-lambda
       [(f init l) (fold-one f init l)]
       [(f init . ls) (apply fold-n f init ls)])))
  
  (define foldr
    (letrec ((fold-one
	      (lambda (f init l)
		(letrec ((helper
			  (lambda (init l)
			    (cond
			     [(null? l) init]
			     [else (f (car l) (helper init (cdr l)))]))))
		  (helper init l))))
	     (fold-n
	      (lambda (f init . l)
		(cond
		 [(ormap null? l)
		  (if (andmap null? l)
		      init
		      (error 'foldr "received non-equal length input lists"))]
		 [else (apply f
			      (mapadd car l
				      (apply fold-n f init (map cdr l))))]))))
      (case-lambda
       [(f init l) (fold-one f init l)]
       [(f init . ls) (apply fold-n f init ls)])))
  )

(open-mzlib-functor mzlib:function@)

(mzlib-constant make-parameter)
(mzlib-constant dynamic-disable-break)
(mzlib-constant remv)
(mzlib-constant remq)
(mzlib-constant remove)
(mzlib-constant ignore-errors)
(mzlib-constant quicksort)
(mzlib-constant identity)
(mzlib-constant compose)
(mzlib-constant foldl)
(mzlib-constant foldr)
