
(require-library "constant.ss")
(require-library "macros.ss")
(require-library "sfunctor.ss")

(define-signature mzlib:string^
  string-lowercase!
  string-uppercase!
  eval-string
  read-string
  read-string-all
  expr->string
  newline-string
  regexp-match-exact?)

(define-sigfunctor (mzlib:string@ mzlib:string^)
  (import)

  (define make-string-do!
    (lambda (translate)
      (lambda (s)
	(let loop ([n (sub1 (string-length s))])
	  (unless (negative? n)
		  (string-set! s n
			       (translate (string-ref s n)))
		  (loop (sub1 n)))))))
  (define string-lowercase! (make-string-do! char-downcase))
  (define string-uppercase! (make-string-do! char-upcase))

  (define eval-string
    (let ([do-eval
	   (lambda (str)
	     (let ([p (open-input-string str)])
	       (apply
		values
		(let loop ()
		  (let ([e (read p)])
		    (if (eof-object? e)
			'()
			(call-with-values
			 (lambda () (eval e))
			 (case-lambda
			  [() (loop)]
			  [(only) (cons only (loop))]
			  [multi 
			   (append multi (loop))]))))))))])
      (opt-lambda (str [error-display #f] [error-result #f])
		  (if (or error-display error-result)
		      (catch-errors 
		       error-display
		       (if error-result error-result (lambda () #f))
		       (do-eval str))
		      (do-eval str)))))

  (define read-string-one-or-all
    (opt-lambda (all? str [error-display #f] [error-result (lambda () #f)])
		(let* ([p (open-input-string str)]
		       [go (lambda ()
			     (let loop ()
			       (let ([v (read p)])
				 (if (eof-object? v)
				     '()
				     (cons v
					   (if all?
					       (loop)
					       '()))))))])
		  (if error-display
		      (go)
		      (catch-errors 
		       error-display
		       error-result
		       (go))))))

  (define read-string
    (lambda args
      (let ([l (apply read-string-one-or-all #f args)])
	(if (null? l)
	    eof
	    (car l)))))

  (define read-string-all
    (lambda args
      (apply read-string-one-or-all #t args)))
  
  (define expr->string
    (lambda (v)
      (let* ([s ""]
	     [write-to-s
	      (lambda (str)
		(set! s (string-append s str)))]
	     [port (make-output-port write-to-s (lambda () #f))])
	(write v port)
	s)))
  
  (define newline-string (string #\newline))
  
  (define regexp-match-exact?
    (lambda (p s)
      (let ([m (regexp-match p s)])
	(and m
	     (string=? (car m) s)))))
  )

(open-mzlib-functor mzlib:string@)
  
(mzlib-constant string-lowercase!)
(mzlib-constant string-uppercase!)
(mzlib-constant eval-string)
(mzlib-constant read-string)
(mzlib-constant read-string-all)
(mzlib-constant expr->string)
(mzlib-constant newline-string)
(mzlib-constant regexp-match-exact?)
