; -*- Emacs-Lisp -*-

(defvar zodiac-base-path "/home/shriram/Works/Zodiac/Export/"
  "Where the Zodiac things are; should end in a slash.")

(require 'cmuscheme)

; ----------------------------------------------------------------------

(defvar zodiac-require-loaded '())

(defun zodiac-require (filename)
  (interactive "f")
  (if (member filename zodiac-require-loaded)
      t
    (load (concat zodiac-base-path filename))
    (setq zodiac-require-loaded (cons filename zodiac-require-loaded))
    'loaded))

; ----------------------------------------------------------------------

(defvar zodiac-max-display-lines 10
  "*Maximum number of lines of Scheme output that are displayed.")

(defun zodiac-display-output (new-buffer-name-string)
  (delete-other-windows)
  (let ((this-buffer (current-buffer)))
    (set-buffer zodiac-s2d-buffer)
    (let* ((lines-present (count-lines (point-min) (point-max)))
	   (lines-to-display (max (min lines-present
				       (1+ zodiac-max-display-lines))
				  window-min-height)))
      (set-buffer this-buffer)
      (split-window-vertically (- 0 lines-to-display))
      (other-window 1)
      (switch-to-buffer zodiac-s2d-buffer t)
      (rename-buffer new-buffer-name-string)
      (other-window -1))))

; ----------------------------------------------------------------------

(defvar zodiac-file/buffer-number 0
  "Number of next transmitted file.  Incremented after each `zodiac-send'.")
(defvar zodiac-file/buffer-alist nil
  "Association list maintaining correspondence between files and buffers.")

(defvar zodiac-transformer-name nil
  "Name of Scheme procedure to be invoked.")

(defun zodiac-tell-scheme (string)
  (comint-send-string (scheme-proc) string))

(defun zodiac-send ()
  "Sends contents of current buffer to Scheme."
  (interactive)
  (copy-to-buffer zodiac-d2s-buffer 1 (point-max))
  (let ((edit-buffer (current-buffer)))
    (set-buffer zodiac-d2s-buffer)
    (write-file zodiac-d2s-filename)
    (set-buffer edit-buffer))
  (zodiac-primitive-send (buffer-file-name) zodiac-d2s-filename t))

(defun zodiac-send-file (filename)
  "Sends contents of named file to Scheme."
  (interactive "fScheme file to load: ")
  (zodiac-primitive-send filename filename nil))

(defun zodiac-primitive-send (actual-filename buffer-filename interactivep)
  (setq zodiac-file/buffer-number (1+ zodiac-file/buffer-number))
  (setq zodiac-file/buffer-alist (cons (cons zodiac-file/buffer-number
					    actual-filename)
				      zodiac-file/buffer-alist))
  (message "Loading %s" actual-filename)
  (zodiac-tell-scheme (concat "("
			     zodiac-transformer-name
			     " \""
			     buffer-filename
			     "\" "
			     zodiac-file/buffer-number " "
			     (if interactivep "#t" "#f") ")\n")))

; ----------------------------------------------------------------------

(defun zodiac-filter (process message)
  "Filter attached to the FIFO."
  (let ((tag+handler (assoc message zodiac-filter-handlers)))
    (if (not tag+handler)
	(message "Don't know how to handle %s" message)
      (funcall (cdr tag+handler)))))

; ----------------------------------------------------------------------

(defvar zodiac-filter-handlers '())

(defun zodiac-add-handler (tag handler)
  (setq zodiac-filter-handlers
	(cons (cons tag handler) zodiac-filter-handlers)))

(defun zodiac-get+cut-1st-line ()
  (goto-line 1)
  (end-of-line)
  (let ((text (buffer-substring (point-min) (point))))
    (delete-region (point-min) (point))
    (delete-char 1)
    text))

(defun zodiac-common-handler (processing-extent)
  (let ((this-buffer (current-buffer))
	(message 'dummy)
	(row     'dummy)
	(column  'dummy))
    (set-buffer zodiac-s2d-buffer)
    (erase-buffer)
    (insert-file-contents zodiac-s2d-filename)
    (if (eq 'w/-message&location processing-extent)
	(progn
	  (setq message (zodiac-get+cut-1st-line))
	  (setq row     (string-to-number (zodiac-get+cut-1st-line)))
	  (setq column  (string-to-number (zodiac-get+cut-1st-line)))
	  (setq offset  (string-to-number (zodiac-get+cut-1st-line)))))
    (find-file (cdr (assoc (string-to-number (zodiac-get+cut-1st-line))
			   zodiac-file/buffer-alist)))
    (if (eq 'w/-message&location processing-extent)
	(progn
	  (goto-line (max row 1))
	  (move-to-column (max 0 (1- column)))
	  message))))

(defun zodiac-handle-dynamic-error-return ()
  (message "Error in %s" (zodiac-common-handler 'w/-message&location))
  (zodiac-display-output "Error Arguments"))

(defun zodiac-handle-static-error-return ()
  (let ((error-message (zodiac-common-handler 'w/-message&location)))
    (message "Error: %s"
	     (substring error-message  ;;; need to strip off quotes
			1 (1- (length error-message))))))

(defun zodiac-handle-result-return ()
  (zodiac-common-handler 'partial)
  (message "Done")
  (zodiac-display-output "Result"))

(defun zodiac-handle-loaded-return ()
  (message "Loaded"))

(zodiac-add-handler "dynamic-error\n" 'zodiac-handle-dynamic-error-return)
(zodiac-add-handler "static-error\n" 'zodiac-handle-static-error-return)
(zodiac-add-handler "result\n" 'zodiac-handle-result-return)
(zodiac-add-handler "loaded\n" 'zodiac-handle-loaded-return)

; ----------------------------------------------------------------------

(defvar zodiac-fifo-filename nil
  "The filename of the FIFO used to communicate from Scheme to Emacs.")
(defvar zodiac-reader-process "fifo-reader"
  "The name we assign to the process reading from the FIFO.")

(defvar zodiac-d2s-filename nil
  "The filename used to transmit code from Emacs to Scheme.")
(defvar zodiac-d2s-buffer nil
  "The buffer holding the contents associated with `zodiac-d2s-filename'.")

(defvar zodiac-s2d-filename nil
  "The filename used to transmit information from Scheme to Emacs.")
(defvar zodiac-s2d-buffer nil
  "Buffer holding the contents associated with `zodiac-s2d-filename'.")

(defvar zodiac-piper-path (concat zodiac-base-path "Library/piper")
  "Full path and filename of the executable that sets up the FIFO.")

(defvar zodiac-scheme-path (concat zodiac-base-path "Devices/Emacs/emacs.ss"))

(defun zodiac-start ()
  "Set up things for our system to work."
  (interactive)

  (setq zodiac-fifo-filename (make-temp-name "/tmp/fifo"))
  (start-process zodiac-reader-process nil
		 zodiac-piper-path zodiac-fifo-filename)
  (set-process-filter (get-process zodiac-reader-process) 'zodiac-filter)

  (setq zodiac-d2s-filename (make-temp-name "/tmp/d2s"))
  (setq zodiac-d2s-buffer (create-file-buffer zodiac-d2s-filename))

  (setq zodiac-s2d-filename (make-temp-name "/tmp/s2d"))
  (setq zodiac-s2d-buffer (create-file-buffer zodiac-s2d-filename))

  (zodiac-tell-scheme (concat "(load \"" zodiac-scheme-path "\")\n"))
  (zodiac-tell-scheme (concat "(zodiac:open-fifo \""
			     zodiac-fifo-filename "\")\n"))
  (zodiac-tell-scheme (concat "(zodiac:s2d-file \""
			      zodiac-s2d-filename "\")\n"))
  (zodiac-tell-scheme "(call-with-current-continuation
                        (lambda (k) (zodiac:top-level-contn k)))\n"))

(defun zodiac-stop ()
  "Close down things that were started up."
  (interactive)
  (kill-buffer zodiac-d2s-buffer)
  (delete-process zodiac-reader-process))
