; -*- Emacs-Lisp -*-

(defvar aries-scheme-path
  (concat zodiac-base-path "Packages/Aries/Source/aries.ss")
  "Relative path and filename of the Scheme code.")

(setq zodiac-transformer-name "aries:transform")

(defvar aries-tracing-delay 1
  "*Pause time during tracing.  Set with `set-aries-tracing-delay'.")

(defvar aries-tracingp nil
  "*Tells us whether tracing has been enabled.  Three meaningful values:
    nil     --  no tracing
    t       --  stepping
    'trace  --  tracing
Access via `aries-stepping', `aries-continuous' and `aries-tracing'.")

(defvar aries-restartablep nil
  "*Records whether an `aries-resume' at this point is meaningful.")

(defun set-aries-tracing-delay (new-delay)
  "Set the value of the delay period."
  (interactive "NNew delay period (in (fractional) multiple of one second): ")
  (setq aries-tracing-delay new-delay))

(defun aries-send-file (filename)
  (setq aries-restartablep nil)
  (zodiac-send-file filename))

(defun aries-handle-dynamic-error-return ()
  (setq aries-restartablep nil)
  (zodiac-handle-dynamic-error-return))

(defun aries-handle-static-error-return ()
  (setq aries-restartablep nil)
  (zodiac-handle-static-error-return))

(defun aries-handle-result-return ()
  (setq aries-restartablep nil)
  (zodiac-handle-result-return))

(defun aries-handle-loaded-return ()
  (setq aries-restartablep nil)
  (zodiac-handle-loaded-return))

(defun aries-handle-primitive-return ()
  (setq aries-restartablep t)
  (message "Evaluating %s" (zodiac-common-handler 'w/-message&location))
  (zodiac-display-output "Trace Arguments")
  (if (eq aries-tracingp 'trace)
      (progn
	(sleep-for aries-tracing-delay)
	(aries-resume))))

(zodiac-add-handler "dynamic-error\n" 'aries-handle-dynamic-error-return)
(zodiac-add-handler "static-error\n" 'aries-handle-static-error-return)
(zodiac-add-handler "result\n" 'aries-handle-result-return)
(zodiac-add-handler "loaded\n" 'aries-handle-loaded-return)
(zodiac-add-handler "primitive\n" 'aries-handle-primitive-return)

(defun aries-resume ()
  "If in the midst of manual tracing, runs one step."
  (interactive)
  (if aries-restartablep
      (zodiac-tell-scheme "(aries:resume)\n")
    (message "Scheme not paused")))

(defun aries-stepping ()
  "Switch to manual stepping."
  (interactive)
  (setq aries-tracingp t)
  (zodiac-tell-scheme "(set! aries:trace-primitives #t)\n"))

(defun aries-continuous ()
  "Turn off stepping."
  (interactive)
  (setq aries-tracingp nil)
  (zodiac-tell-scheme "(set! aries:trace-primitives #f)\n"))

(defun aries-tracing ()
  "Switch to automatic stepping."
  (interactive)
  (setq aries-tracingp 'trace)
  (zodiac-tell-scheme "(set! aries:trace-primitives #t)\n"))

(define-key global-map "\M-gs" 'zodiac-send)
(define-key global-map "\M-gf" 'aries-send-file)
(define-key global-map "\M-gr" 'aries-resume)
(define-key global-map "\M-gt" 'aries-tracing)
(define-key global-map "\M-gd" 'aries-stepping)
(define-key global-map "\M-gc" 'aries-continuous)
(define-key global-map "\M-g+" 'aries-start)
(define-key global-map "\M-g-" 'aries-stop)

(define-key scheme-mode-map [menu-bar aries-entry]
  (cons "Aries" (make-sparse-keymap "Aries")))

(define-key scheme-mode-map [menu-bar aries-entry stop]
  '("Stop"         . aries-stop))		 
(define-key scheme-mode-map [menu-bar aries-entry start]
  '("Start"        . aries-start))		 
(define-key scheme-mode-map [menu-bar aries-entry continuous]
  '("Continue"     . aries-continuous))		 
(define-key scheme-mode-map [menu-bar aries-entry stepping]
  '("Stepping"     . aries-stepping))		 
(define-key scheme-mode-map [menu-bar aries-entry tracing]
  '("Tracing"      . aries-tracing))		 
(define-key scheme-mode-map [menu-bar aries-entry resume]
  '("Resume"       . aries-resume))		 
(define-key scheme-mode-map [menu-bar aries-entry send-file]
  '("Send File"    . aries-send-file))		 
(define-key scheme-mode-map [menu-bar aries-entry send-current]
  '("Send Current" . aries-send))
