; -*- Scheme -*-

; This code is _the same_ as that in Source/Scheme/mzrice.ss,
; re-expressed.  All changes MUST BE MIRRORED.

(define aries:atom?
  (let ((not #%not) (pair? #%pair?) (null? #%null?))
    (aries:make-unary-untyped-primitive
      (lambda (a) (not (or (pair? a) (null? a)))) 'atom?)))

(define aries:cons?
  (aries:make-unary-untyped-primitive #%pair? 'cons?))

(define aries:cons
  (let ((pair? #%pair?) (null? #%null?) (cons #%cons))
    (aries:make-binary-typed-primitive
      (lambda (arg1) #t)
      (lambda (arg2) (or (pair? arg2) (null? arg2)))
      cons
      'cons)))

(define aries:list?
  (let ((pair? #%pair?) (null? #%null?))
    (aries:make-unary-untyped-primitive
      (lambda (l) (or (pair? l) (null? l))) 'list?)))

(define aries:set-cdr!
  (let ((pair? #%pair?) (null? #%null?) (set-cdr! #%set-cdr!))
    (aries:make-binary-typed-primitive
      pair?
      (lambda (l) (or (pair? l) (null? l)))
      set-cdr! 'set-cdr!)))

; Distinguish vectors from structures by checking first field:
; is it zodiac:dt-structure-tagger?

(define aries:structure?
  (let ((vector? #%vector?) (vector-length #%vector-length)
	 (vector-ref #%vector-ref) (eq? #%eq?) (>= #%>=))
    (aries:make-unary-untyped-primitive
      (lambda (v)
	(and (vector? v)
	  (and (>= (vector-length v) 1)
	    (eq? zodiac:dt-structure-tagger (vector-ref v 0)))))
      'structure?)))

(define aries:real-vector?
  (let ((vector? #%vector?) (vector-ref #%vector-ref) (eq? #%eq?)
	 (not #%not) (vector-length #%vector-length) (= #%=))
    (lambda (v)
      (and (vector? v)
	(or (= (vector-length v) 0)
	  (not (eq? zodiac:dt-structure-tagger (vector-ref v 0))))))))

(define aries:vector?
  (let ((vector? #%vector?) (vector-ref #%vector-ref) (eq? #%eq?)
	 (not #%not) (vector-length #%vector-length) (= #%=))
    (aries:make-unary-untyped-primitive aries:real-vector? 'vector?)))

(define aries:vector-length
  (let ((vector-length #%vector-length))
    (aries:make-unary-typed-primitive
      aries:real-vector? vector-length 'vector-length)))

(define aries:vector-set!
  (let ((number? #%number?) (>= #%>=) (< #%<)
	 (vector-length #%vector-length) (vector-set! #%vector-set!))
    (aries:make-ternary-checked-primitive
      (lambda (v n x)
	(and (aries:real-vector? v)
	  (number? n)
	  (>= n 0)
	  (< n (vector-length v))))
      vector-set! 'vector-set!)))

(define aries:vector->list
  (let ((vector->list #%vector->list))
    (aries:make-unary-typed-primitive
      aries:real-vector? vector->list 'vector->list)))

(define aries:vector-fill!
  (let ((vector-fill! #%vector-fill!))
    (aries:make-binary-typed-primitive
      aries:real-vector? (lambda (x) #t) vector-fill! 'vector-fill!)))

(define aries:vector-ref
  (let ((vector-length #%vector-length) (vector-ref #%vector-ref)
	 (integer? #%integer?) (>= #%>=) (< #%<))
    (aries:make-binary-checked-primitive
      (lambda (v n)
	(and (aries:real-vector? v)
	  (integer? n)
	  (>= n 0)
	  (< n (vector-length v))))
      vector-ref 'vector-ref)))
