; Yo, Emacs & MrEd: this here be -*- Scheme -*-

(define-struct zodiac:begin-form (first rest open close back))

(define-term begin () Env O C
  ((_ b)
    #t
    (zodiac:dt-expand (syntax b) Env O C))
  ((_ b0 b1 ...)
    #t
    (make-zodiac:begin-form (zodiac:dt-expand (syntax b0) Env O C)
      (zodiac:dt-expand (syntax (begin b1 ...)) Env O C)
      O C (zodiac:make-back))))

(define-struct zodiac:lambda-form (args body level open close back))

(define-term lambda () Env O C
  ((_ (v ...) b ...)
    (zodiac:syntax-andmap identifier? (syntax (v ...)))
    (let ((new-names (zodiac:dt-bind (zodiac:dt-generate-names (syntax (v ...))) O C)))
      (make-zodiac:lambda-form new-names
	(zodiac:dt-expand (syntax (begin b ...))
	  (zodiac:dt-extend Env new-names (syntax (v ...))) O C)
	-1
	O C (zodiac:make-back))))
  ((_ vs b ...)
    (let ((listized-names
	    (let loop ((ids (syntax vs)))
	      (cond
		((zodiac:syntax-pair? ids)
		  (let ((first (zodiac:syntax-car ids)))
		    (if (identifier? first)
		      (#%cons first (loop (zodiac:syntax-cdr ids)))
		      (zodiac:syntax-error "not identifiers" ids))))
		(else (if (identifier? ids) (list ids)
			(zodiac:syntax-error "not an identifier" ids)))))))
      (let ((new-names (zodiac:dt-bind (zodiac:dt-generate-names listized-names) O C)))
	(make-zodiac:lambda-form (let loop ((names new-names))
				   (cond
				     ((null? (cdr names)) (car names))
				     (else (#%cons (car names)
					     (loop (cdr names))))))
	  (zodiac:dt-expand (syntax (begin b ...))
	    (zodiac:dt-extend Env new-names listized-names) O C)
	  -1
	  O C (zodiac:make-back))))))

(define-struct zodiac:letrec-form (vars vals body level open close back))

(define-term letrec () Env O C
  ((_ ((v e) ...) b0 b1 ...)
    (andmap identifier? (syntax (v ...)))
    (let* ((new-names (zodiac:dt-bind (zodiac:dt-generate-names (syntax (v ...))) O C))
	    (new-env (zodiac:dt-extend Env new-names (syntax (v ...)))))
      (make-zodiac:letrec-form new-names
	(map (lambda (expr) (zodiac:dt-expand expr new-env O C))
	  (syntax (e ...)))
	(zodiac:dt-expand (syntax (begin b0 b1 ...)) new-env O C)
	-1
	O C (zodiac:make-back)))))

(define-struct zodiac:let-form (vars vals body level open close back))

(define-term let () Env O C
  ((_ () b ...)
    #t
    (zodiac:dt-expand (syntax (begin b ...)) Env O C))
  ((_ ((v e) ...) b ...)
    (andmap identifier? (syntax (v ...)))
    (let ((new-names (zodiac:dt-bind (zodiac:dt-generate-names (syntax (v ...))) O C)))
      (make-zodiac:let-form new-names
	(map (lambda (expr) (zodiac:dt-expand expr Env O C)) (syntax (e ...)))
	(zodiac:dt-expand (syntax (begin b ...))
	  (zodiac:dt-extend Env new-names (syntax (v ...))) O C)
	-1
	O C (zodiac:make-back))))
  ((_ f ((v e) ...) b ...)
    (and (identifier? (syntax f)) (andmap identifier? (syntax (v ...))))
    (zodiac:dt-expand (syntax ((letrec ((f (lambda (v ...) b ...))) f) e ...)) Env O C)))

(define-term let* () Env O C
  ((_ () b ...)
    #t
    (zodiac:dt-expand (syntax (let () b ...)) Env O C))
  ((_ ((v0 e0) (v1 e1) ...) b ...)
    #t
    (zodiac:dt-expand (syntax (let ((v0 e0)) (let* ((v1 e1) ...) b ...))) Env O C)))

(define-struct zodiac:if-form (test then else open close back))

(define-term if () Env O C
  ((_ test then)
    #t
    (zodiac:dt-expand (syntax (if test then (void))) Env O C))
  ((_ test then else)
    #t
    (make-zodiac:if-form (zodiac:dt-expand (syntax test) Env O C)
      (zodiac:dt-expand (syntax then) Env O C) (zodiac:dt-expand (syntax else) Env O C)
      O C (zodiac:make-back))))

(define-term when () Env O C
  ((_ cond b ...)
    #t
    (zodiac:dt-expand (syntax (if cond (begin b ...))) Env O C)))

(define-term unless () Env O C
  ((_ cond b ...)
    #t
    (zodiac:dt-expand (syntax (when (not cond) b ...)) Env O C)))

(define-term cond (else) Env O C
  ((_ (else b ...))
    #t
    (zodiac:dt-expand (syntax (begin b ...)) Env O C))
  ((_ (test b ...))
    #t
    (zodiac:dt-expand (syntax (if test
				(begin b ...)
				(raise (make-exn:else
					 "cond: no matching clause"
					 #f))))
      Env O C))
  ((_ (test b ...) rest ...)
    #t
    (zodiac:dt-expand (syntax (if test (begin b ...) (cond rest ...))) Env O C)))

(define-term case (else) Env O C
  ((_ test (else b ...))
    #t
    (zodiac:dt-expand (syntax (begin b ...)) Env O C))
  ((_ test ((keys ...) b ...))
    #t
    (zodiac:dt-expand (syntax
			(let ((%%tmp test))
			  (if (memv %%tmp (quote (keys ...)))
			    (begin b ...)
			    (raise (make-exn:else
				     "case: no matching clause" #f)))))
      Env O C))
  ((_ test ((keys ...) b ...) cases ...)
    #t
    (zodiac:dt-expand (syntax (let ((%%tmp test))
		   (if (memv %%tmp (quote (keys ...)))
		     (begin b ...)
		     (case test cases ...))))
      Env O C)))

(define-term recur () Env O C
  ((recur fun ((v e) ...) b ...)
    (and (identifier? (syntax fun)) (andmap identifier? (syntax (v ...))))
    (zodiac:dt-expand (syntax (let fun ((v e) ...) b ...)) Env O C)))

(define-term or () Env O C
  ((_)
    #t
    (zodiac:dt-expand #f Env O C))
  ((_ e)
    #t
    (zodiac:dt-expand (syntax e) Env O C))
  ((_ e0 e1 e2 ...)
    #t
    (zodiac:dt-expand (syntax (let ((%%tmp e0)) (if %%tmp %%tmp (or e1 e2 ...)))) Env O C)))

(define-term and () Env O C
  ((_)
    #t
    (zodiac:dt-expand #t Env O C))
  ((_ e)
    #t
    (zodiac:dt-expand (syntax e) Env O C))
  ((_ e0 e1 e2 ...)
    #t
    (zodiac:dt-expand (syntax (if e0 (and e1 e2 ...) #f)) Env O C)))

(define-struct zodiac:delay-form (expr open close back))

(define-term delay () Env O C
  ((_ body)
    #t
    (make-zodiac:delay-form (zodiac:dt-expand (syntax body) Env O C)
      O C (zodiac:make-back))))

(define-struct zodiac:set!-form (var val open close back))

(define-term set! () Env O C
  ((_ v e)
    (identifier? (syntax v))
    (make-zodiac:set!-form (zodiac:dt-expand (syntax v) Env O C)
      (zodiac:dt-expand (syntax e) Env O C)
      O C (zodiac:make-back))))

(define-struct zodiac:define-form (name value open close back))

(define-term define () Env O C
  ((_ v)
    (identifier? (syntax v))
    (zodiac:dt-expand (syntax (define v (void))) Env O C))
  ((_ v e)
    (identifier? (syntax v))
    (zodiac:dt-expand (syntax (set! v e)) Env O C))
  ((_ (f . args) e ...)
    (identifier? (syntax f))
    (zodiac:dt-expand (syntax (define f (lambda args e ...))) Env O C)))

(define-struct zodiac:time-form (expr open close back))

(define-term time () Env O C
  ((_ e)
    #t
    (make-zodiac:time-form
      (zodiac:dt-expand (syntax e) Env O C)
      O C (zodiac:make-back))))

(define-struct zodiac:quote-form (expr open close back))

(define-term quote () Env O C
  ((_ e)
    #t
    (make-zodiac:quote-form (syntax e) O C (zodiac:make-back))))

(define-term unquote () Env O C
  ((_ e)
    #t
    (zodiac:syntax-error "unquote outside quasiquote" (syntax e))))

(define-term unquote-splicing () Env O C
  ((_ e)
    #t
    (zodiac:syntax-error "unquote-splicing outside quasiquote" (syntax e))))

(define-term quasiquote () Env O C
  ((_ e)
    #t
    (zodiac:dt-expand 
      (zodiac:sc-expand (syntax (quasiquote e)))
      Env O C)))

(define-term do () Env O C
  ((_ ((var init step) ...)
      (test sequence ...)
     command ...)
    (andmap identifier? (syntax (var ...)))
    (zodiac:dt-expand
      (syntax
	(letrec ((%%tmp-do$$ (lambda (var ...)
			       (if test
				 (begin sequence ...)
				 (begin command ...
				   (%%tmp-do$$ step ...))))))
	  (%%tmp-do$$ init ...)))
      Env O C)))
