; -*- Scheme -*-

; set-once
; handle

(define-term let/cc () Env O C
  ((_ k e ...)
    (identifier? (syntax k))
    (zodiac:dt-expand (syntax (#%call/cc (lambda (k) e ...))) Env O C)))

(define-term letcc () Env O C
  ((_ k e ...)
    (identifier? (syntax k))
    (zodiac:dt-expand (syntax (#%call/cc (lambda (k) e ...))) Env O C)))

(define-term begin0 () Env O C
  ((_ b0 b1 ...)
    (zodiac:dt-expand
      (syntax
	(let ((%%begin0-tmp$$ b0))
	  b1 ...
	  %%begin0-tmp$$))
      Env O C)))

(define-term define-schema () Env O C
  ((_ x e)
    #t
    (zodiac:dt-expand (syntax (#%void)) Env O C)))

; If you change the representation of structures, remember to
; change the corresponding definitions for primitives in Aries!

(define zodiac:dt-structure-tagger (string->uninterned-symbol "structure"))
(define zodiac:dt-structure-counter (box 0))

(define-term define-structure () Env O C
  ((_ (type fields ...))
    #t
    (let ((type-name (syntax-object->datum (syntax type)))
	   (field-names (#%map syntax-object->datum (syntax (fields ...)))))
      (let ((*-append (lambda names
			(#%string->symbol
			  (#%apply #%string-append
			    (#%map (lambda (s)
				   (cond
				     ((zodiac:symbol? s)
				       (#%symbol->string
					 (zodiac:symbol-text s)))
				     ((#%symbol? s)
				       (#%symbol->string s))
				     ((zodiac:string? s)
				       (zodiac:string-text s))
				     ((#%string? s)
				       s)
				     ((zodiac:number? s)
				       (#%number->string
					 (zodiac:number-text s)))
				     ((#%number? s)
				       (#%number->string s))))
			      names))))))
	(let ((tag-count (let ((c (unbox zodiac:dt-structure-counter)))
			   (#%set-box! zodiac:dt-structure-counter
			     (#%+ c 1))
			   c))
	       (real-predicate (#%gensym))
	       (predicate (*-append type-name "?"))
	       (constructor (*-append "make-" type-name))
	       (accessors (#%map (lambda (s) (*-append type-name "-" s))
			    field-names))
	       (setters (#%map (lambda (s)
				 (*-append "set-" type-name "-" s "!"))
			  field-names))
	       (counters (let loop ((n 4) (size (length field-names)))
			   (if (#%zero? size)
			     '()
			     (#%cons n (loop (#%+ n 1) (#%- size 1))))))
	       (vector-size (#%+ 4 (#%length field-names))))
	  (let ((expr
		  `(begin
		     (define ,real-predicate
		       (lambda (v)
			 (and (#%vector? v)
			   (#%= ,vector-size (#%vector-length v))
			   (#%eq? (#%vector-ref v 0)
			     ',zodiac:dt-structure-tagger)
			   (#%eq? (#%vector-ref v 1) ',type-name)
			   (#%= (#%vector-ref v 3) ,tag-count))))
		     (define ,predicate
		       (lambda (v)
			 (,real-predicate v)))
		     (define ,constructor
		       (lambda ,field-names
			 (#%vector ',zodiac:dt-structure-tagger
			   ',type-name 'version ,tag-count
			   ,@field-names)))
		     ,@(map
			 (lambda (f-name posn)
			   `(define ,f-name
			      (lambda (v)
				(if (,real-predicate v)
				  (#%vector-ref v ,posn)
				  (#%error ',f-name
				    "~s not of right type" v)))))
			 accessors counters)
		     ,@(map
			 (lambda (f-name posn)
			   `(define ,f-name
			      (lambda (v val)
				(if (,real-predicate v)
				  (#%vector-set! v ,posn val)
				  (#%error ',f-name
				    "~s not of right type" v)))))
			 setters counters))))
	    (zodiac:dt-expand expr Env O C)))))))

(define zodiac:dt-local-extract-declared-names
  (lambda (ast)
    (cond
      ((zodiac:set!-form? ast)
	(list (zodiac:set!-form-var ast)))
      ((zodiac:define-form? ast)
	(list (zodiac:define-form-name ast)))
      ((zodiac:begin-form? ast)
	(append (zodiac:dt-local-extract-declared-names
		  (zodiac:begin-form-first ast))
	  (zodiac:dt-local-extract-declared-names
	    (zodiac:begin-form-rest ast))))
      (else '()))))

(define zodiac:dt-local-get-name
  (lambda (n)
    (cond
      ((zodiac:unbound-varref? n)
	(zodiac:unbound-varref-var n))
      ((zodiac:top-level-varref? n)
	(zodiac:top-level-varref-var n))
      ((zodiac:lexical-varref? n)
	(zodiac:lexical-varref-var n)))))

(define-term local () Env O C
  ((_ (defs ...) b ...)
    #t
    (let loop
       ((definitions (syntax (defs ...)))
	 (symbols-defined '())
	 (sym/struct-defs '()))
      (if (zodiac:syntax-null? definitions)
	(let ((exp `(let ,(map (lambda (s) `(,(zodiac:dt-local-get-name s)
					      (void)))
			    symbols-defined)
		      ,@(reverse sym/struct-defs)
		      ,@(syntax (b ...)))))
	  (zodiac:dt-expand exp Env O C))
	(let ((this-definition (zodiac:syntax-car definitions))
	       (rest-definitions (zodiac:syntax-cdr definitions)))
	  (if (zodiac:syntax-null? this-definition)
	    (zodiac:report-static-error "bad local declaration"
	      this-definition))
	  (let ((decl-kind (zodiac:syntax-car this-definition)))
	    (cond
	      ((or (zodiac:same-name? decl-kind 'define)
		 (zodiac:same-name? decl-kind 'define-structure))
		(let* ((xp (expand this-definition))
			(names (zodiac:dt-local-extract-declared-names xp)))
		  (loop
		    rest-definitions
		    (append names symbols-defined)
		    (cons this-definition sym/struct-defs))))
	      (else
		(zodiac:report-static-error "not define or define-structure"
		  this-definition)))))))))

(define-struct rs:module (exporter))

(define module? (lambda (x) (rs:module? x)))

(define-term module () Env O C
  ((_ (var ...) defs ...)
    (andmap identifier? (syntax (var ...)))
    (zodiac:dt-expand
      (syntax
	(local (defs ...)
	  (make-rs:module
	    (lambda (msg)
	      (case msg
		((var) var) ...
		(else
		  (#%error 'import
		    "attempting to import ~s~n        from module with signature ~s"
		    msg (quote (var ...)))))))))
      Env O C)))

(define zodiac:dt-import-normalize
  (lambda (extractors new-mod-names)
    (let loop
      ((extractors (zodiac:strip extractors))
	(new-mod-names new-mod-names))
      (if (null? extractors)
	'()
	(let ((mod (zodiac:syntax-car (car extractors)))
	       (names (zodiac:syntax-cdr (car extractors))))
	  (append
	    (map (lambda (name)
		   (list name
		     (list (car new-mod-names)
		       (list 'quote name))))
	      names)
	    (loop (cdr extractors) (cdr new-mod-names))))))))

(define-term import () Env O C
  ((_ ((mod defs ...) ...) exp ...)
    (andmap identifier? (syntax (mod ...)))
    (let* ((mods (syntax (mod ...)))
	    (new-mods (zodiac:dt-generate-names mods))
	    (extractors (zodiac:dt-import-normalize
			  (syntax ((mod defs ...) ...))
			  new-mods)))
      (let ((expr `(let ,(zodiac:syntax-map (lambda (n-m m) (list n-m m))
			   new-mods mods)
		     ,@(map (lambda (n-m m)
			      `(unless (rs:module? ,n-m)
				 (#%error 'import "~s not a module"
				   (quote ,m))))
			 new-mods mods)
		     (let ,(map (lambda (n-m)
				  `(,n-m (rs:module-exporter ,n-m)))
			     new-mods)
		       (let ,extractors
			 ,@(syntax (exp ...)))))))
	(zodiac:dt-expand expr Env O C)))))

(define-struct zodiac::-form (expr type open close back))

(define-term : () Env O C
  ((_ expr type)
    #t
    (make-zodiac::-form
      (zodiac:dt-expand (syntax expr) Env O C)
      (syntax type)
      O C (zodiac:make-back))))

(define-term define-primitive () Env O C
  ((_ expr ...)
    #t
    (zodiac:dt-expand '(void) Env O C)))

(define-term define-constructor () Env O C
  ((_ expr ...)
    #t
    (zodiac:dt-expand '(void) Env O C)))

(define-term define-type () Env O C
  ((_ expr ...)
    #t
    (zodiac:dt-expand '(void) Env O C)))

(define-term begin-test-case () Env O C
  ((_ exp0 exp1 ...)
    #t
    (zodiac:dt-expand '(void) Env O C)))
