(define zodiac:expand-action eval)

(define-struct zodiac:dt-env-entry (name kind info))

(define-struct zodiac:unbound-varref (var back))
(define-struct zodiac:lexical-varref (var binding back))

(define-struct zodiac:app (fun args open close back))

(define zodiac:dt-not-in-env
  (make-zodiac:dt-env-entry 'dummy-unbound 'global 'dummy))

(define zodiac:dt-extend-global-env-w/-kind
  (lambda (name kind info)
    (set! zodiac:dt-global-env
      (cons (make-zodiac:dt-env-entry name kind info)
	zodiac:dt-global-env))))

(define zodiac:dt-extend-global-env
  (lambda (name action)
    (zodiac:dt-extend-global-env-w/-kind name 'term action)))

(define zodiac:dt-global-env '())

(define zodiac:get-macro-env
  (lambda ()
    zodiac:dt-global-env))

(define zodiac:put-macro-env
  (lambda (env)
    (set! zodiac:dt-global-env env)))

(define zodiac:treat-unchanged
  (lambda (trigger)
    (zodiac:dt-extend-global-env-w/-kind trigger 'unchanged '())))

(zodiac:dt-extend-global-env
  'define-term
  (lambda (body env open close)
    (let ((term-name (cadr body))
	   (guardeds (caddr body))
	   (env-name (cadddr body))
	   (open-loc (car (cddddr body)))
	   (close-loc (car (cdr (cddddr body))))
	   (patterns (cddr (cddddr body))))
      (let ((new-expander
	      (zodiac:expand-action
		`(lambda (body ,env-name ,open-loc ,close-loc)
		   ,(zodiac:sc-expand
		      `(syntax-case body ,guardeds
			 ,@patterns))))))
	(zodiac:dt-extend-global-env term-name new-expander)))))

; ----------------------------------------------------------------------

(define zodiac:make-back (lambda () (box zodiac:box-default-value)))

(define zodiac:dt-expand
  (lambda (body env open close)
    (define (locationize body)
      (if (struct? body) body (zodiac:endow-w/-locs body open close)))
    (let ((open (if (struct? body) (zodiac:determine-start-location body)
		  open))
	   (close (if (struct? body)
		    (zodiac:determine-finish-location body) close)))
      (cond
	((zodiac:syntax-symbol? body)
	  (let ((entry (zodiac:dt-lookup body env)))
	    (let ((body (locationize body)))
	      (case (zodiac:dt-env-entry-kind entry)
		((global) (make-zodiac:unbound-varref body (zodiac:make-back)))
		((lexical) (let ((binding (zodiac:dt-env-entry-info entry)))
			     (make-zodiac:lexical-varref
			       (make-zodiac:symbol
				 (zodiac:symbol-text
				   (locationize
				     (zodiac:bound-var 
				       binding)))
				 open close)
			       (make-zodiac:binding -1 -1
				 (zodiac:dt-env-entry-info entry))
			       (zodiac:make-back))))
		((term) (zodiac:report-static-error "invalid use of keyword"
			  body))
		((unchanged) (make-zodiac:unchanged body open close))))))
	((zodiac:syntax-pair? body)
	  (if (zodiac:improper-list? body)
	    (zodiac:report-static-error "invalid syntax" body))
	  (if (zodiac:syntax-ormap syntax-object? body)
	    (zodiac:dt-expand (zodiac:syntax-map
				(lambda (e) (if (syntax-object? e)
					      (syntax-object-expression e) e))
				body)
	      env open close)
	    (if (zodiac:syntax-symbol? (zodiac:syntax-car body))
	      (let ((entry (zodiac:dt-lookup (zodiac:syntax-car body) env)))
		(case (zodiac:dt-env-entry-kind entry)
		  ((global lexical)
		    (make-zodiac:app
		      (zodiac:dt-expand (zodiac:syntax-car body) env
			open close)
		      (zodiac:syntax-map (lambda (e)
					   (zodiac:dt-expand e env open close))
			(zodiac:syntax-cdr body))
		      open close (zodiac:make-back)))
		  ((term)
		    ((zodiac:dt-env-entry-info entry) body env open close))
		  ((unchanged)
		    (make-zodiac:unchanged body open close))))
	      (make-zodiac:app
		(zodiac:dt-expand (zodiac:syntax-car body) env open close)
		(zodiac:syntax-map (lambda (e)
				     (zodiac:dt-expand e env open close))
		  (zodiac:syntax-cdr body))
		open close (zodiac:make-back)))))
	((or (zodiac:syntax-number? body) (zodiac:syntax-string? body)
	   (zodiac:syntax-char? body) (zodiac:syntax-boolean? body)
	   (zodiac:syntax-vector? body))
	  (make-zodiac:quote-form (locationize body)
	    open close (zodiac:make-back)))
	((syntax-object? body)
	  (zodiac:dt-expand (syntax-object-expression body) env open close))
	(else (zodiac:report-static-error "invalid syntax" body))
	))))

(define zodiac:dt-generate-names
  (lambda (old-names)
    (zodiac:syntax-map zodiac:new-symbol-hook old-names)))

(define zodiac:dt-extend
  (case-lambda
    ((old-env new-bindings)
      (zodiac:dt-extend old-env new-bindings new-bindings))
    ((old-env new-names old-names)
      (append (zodiac:syntax-map (lambda (new-name old-name)
		     (make-zodiac:dt-env-entry (if (syntax-object? old-name)
				       (syntax-object-expression old-name)
				       old-name)
		       'lexical new-name))
		new-names old-names)
	old-env))))

(define zodiac:dt-lookup
  (lambda (name env)
    (let loop ((bindings env))
      (if (zodiac:syntax-null? bindings)
	zodiac:dt-not-in-env
	(let ((first (zodiac:syntax-car bindings)))
	  (if (zodiac:same-name? (zodiac:dt-env-entry-name first) name)
	    first
	    (loop (zodiac:syntax-cdr bindings))))))))

(define zodiac:dt-fake-location (make-zodiac:location -1 -1 -1 'none))

(define (zodiac:dt-determine-start-location e)
  (if (struct? e)
    (zodiac:determine-start-location e)
    zodiac:dt-fake-location))

(define (zodiac:dt-determine-finish-location e)
  (if (struct? e)
    (zodiac:determine-finish-location e)
    zodiac:dt-fake-location))

(define expand
  (lambda (e)
    (zodiac:dt-expand e zodiac:dt-global-env
      (zodiac:dt-determine-start-location e)
      (zodiac:dt-determine-finish-location e))))

(define zodiac:expand expand)

(define zodiac:dt-bind
  (lambda (name/s open close)
    (cond
      ((zodiac:syntax-null? name/s) '())
      ((zodiac:syntax-pair? name/s)
	(cons (zodiac:dt-bind (zodiac:syntax-car name/s) open close)
	  (zodiac:dt-bind (zodiac:syntax-cdr name/s) open close)))
      (else (make-zodiac:bound
	      (if (symbol? name/s)
		(make-zodiac:symbol name/s open close)
		name/s)
	      (zodiac:make-back))))))

(define zodiac:dt-load
  (lambda (filename)
    (with-input-from-file
      filename
      (lambda ()
	(let loop ()
	  (let ((e (read)))
	    (if (eof-object? e)
	      'done
	      (begin
		(if (and (pair? e)
		      (eq? 'define-term (car e)))
		  (expand e)
		  (zodiac:expand-action e))
		(loop)))))))))

; ----------------------------------------------------------------------

;(define (tdt) (pretty-print (expand ((zodiac:read)))))
