(define zodiac:cons #%cons)

(define zodiac:sc-to-monitor #f)

(define zodiac:sc-to-monitor-match #f)

(define zodiac:sc-monitor-match
  (let ((apply apply))
    (lambda args
      (fluid-let ((zodiac:sc-to-monitor zodiac:sc-to-monitor-match))
	(apply zodiac:sc-monitor args)))))

(define zodiac:sc-monitor
  (let ((display display) (newline newline) (null? null?)
	 (pretty-print pretty-print) (cadr cadr) (cddr cddr))
    (lambda (where . args)
      (when zodiac:sc-to-monitor
	(display where) (display ":") (newline)
	(let loop ((args args))
	  (if (null? args)
	    (newline)
	    (begin
	      (display (car args)) (newline)
	      (pretty-print (cadr args))
	      (loop (cddr args)))))))))

(define zodiac:new-symbol-hook
  (let ((count 0))
    (let ((string->symbol string->symbol) (string-append string-append)
	   (number->string number->string) (symbol->string symbol->string)
	   (+ +))
      (lambda (sym)
	(let ((sym (if (syntax-object? sym)
		     (syntax-object-expression sym) sym)))
	  (letrec ((new-name (lambda (sym)
			       (string->symbol
				 (string-append
				   "gen:" (number->string count)
				   ":" (symbol->string sym))))))
	    (set! count (+ 1 count))
	    (if (zodiac:symbol? sym)
	      (make-zodiac:symbol (new-name (zodiac:symbol-text sym))
		(zodiac:symbol-start sym)
		(zodiac:symbol-finish sym))
	      (new-name sym))))))))

(define zodiac:eval-hook eval)

(define zodiac:build-with-source
  (lambda (source text)
    (cond
     ((zodiac:base-type? text) text)
     ((zodiac:list? source)
      (make-zodiac:list
       (if (zodiac:list? text) (zodiac:list-contents text) text)
       (zodiac:syntax-length text)
       (zodiac:list-open source)
       (zodiac:list-close source)))
     ((zodiac:aggregate-type? text) text)
     (else text))))

(define zodiac:build-application
  (lambda (source fun-exp arg-exps)
    (zodiac:build-with-source source
		       (zodiac:syntax-cons
			fun-exp arg-exps))))

(define zodiac:build-conditional
  (lambda (source test-exp then-exp else-exp)
    (zodiac:build-with-source source
		       `(if ,test-exp ,then-exp ,else-exp))))

(define zodiac:build-lexical-reference (lambda (var) var))

(define zodiac:build-lexical-assignment
  (lambda (source var exp)
    (zodiac:build-with-source source
		       `(set! ,var ,exp))))

(define zodiac:build-global-reference (lambda (var) var))

(define zodiac:build-global-assignment
  (lambda (source var exp)
    (zodiac:build-with-source source
		       `(set! ,var ,exp))))

(define zodiac:build-case-lambda
   (lambda (source vars exp)
     (zodiac:build-with-source source
			`(case-lambda ,vars ,exp))))

(define zodiac:build-make-class*
   (lambda (source stuff)
      (zodiac:build-with-source source
	 `(make-class* ,stuff))))

(define zodiac:build-let
  (let ((list list) (map map))
    (lambda (source vars vals body)
      (zodiac:build-with-source source
	`(let ,(map list vars vals) ,body)))))

(define zodiac:build-lambda
   (lambda (source vars exp)
     (zodiac:build-with-source source
			`(lambda ,vars ,exp))))

(define zodiac:build-improper-lambda
   (lambda (source vars var exp)
     (zodiac:build-with-source source
			`(lambda (,@vars . ,var) ,exp))))

(define zodiac:build-data
  (lambda (source exp)
    (zodiac:build-with-source source
		       `(quote ,exp))))

(define zodiac:build-identifier
   (lambda (id)
      `(quote ,id)))

(define zodiac:build-sequence
  (let ((null? null?) (cdr cdr) (car car))
    (lambda (source exps)
      (zodiac:build-with-source
	source
	(if (null? (cdr exps))
	  (car exps)
	  `(begin ,(car exps) ,(zodiac:build-sequence source (cdr exps))))))))

(define zodiac:build-letrec
  (let ((null? null?) (map map) (list list))
    (lambda (source vars val-exps body-exp)
      (zodiac:build-with-source
	source
	(if (null? vars)
	  body-exp
	  `(letrec ,(map list vars val-exps) ,body-exp))))))

(define zodiac:build-global-definition
  (lambda (source var val)
    (zodiac:build-with-source source
		       `(define ,var ,val))))

(define zodiac:expand #f)
(define zodiac:generate-temporaries #f)
(define zodiac:identifier? #f)
(define zodiac:syntax-dispatch #f)
(define zodiac:install-global-transformer #f)
(define zodiac:syntax-error #f)
(define zodiac:bound-identifier=? #f)
(define zodiac:free-identifier=? #f)
(define zodiac:syntax-object->datum #f)
(define zodiac:datum->syntax-object #f)
(define zodiac:get-macro-env #f)
(define zodiac:put-macro-env #f)
