(define zodiac:expr-stripper
  (let ((pair? pair?) (cons cons) (car car) (cdr cdr) (cadr cadr)
	 (list->vector list->vector) (map map) (vector->list vector->list))
    (lambda (expr)
      (cond
	((zodiac:raw-base-type? expr)
	  expr)
	((pair? expr)
	  (cons (zodiac:expr-stripper (car expr))
	    (zodiac:expr-stripper (cdr expr))))
	((vector? expr)
	  (list->vector (map zodiac:expr-stripper (vector->list expr))))

	((zodiac:base-type? expr)
	  (zodiac:base-type-contents expr))
	((zodiac:list? expr)
	  (map zodiac:expr-stripper (zodiac:list-contents expr)))
	((zodiac:vector? expr)
	  (list->vector (map zodiac:expr-stripper
			  (vector->list (zodiac:vector-contents expr)))))
	((zodiac:improper-list? expr)
	  (let loop
	    ((contents (zodiac:improper-list-contents expr)))
	    (cond
	      ((pair? contents)
		(if (zodiac:period? (car contents))
		  (loop (cadr contents))
		  (cons (zodiac:expr-stripper (car contents))
		    (loop (cdr contents)))))
	      (else (zodiac:expr-stripper contents)))))
	  
	((zodiac:delay-form? expr)
	  `(delay ,(zodiac:expr-stripper (zodiac:delay-form-expr expr))))
	((zodiac:if-form? expr)
	  `(if ,(zodiac:expr-stripper (zodiac:if-form-test expr))
	     ,(zodiac:expr-stripper (zodiac:if-form-then expr))
	     ,(zodiac:expr-stripper (zodiac:if-form-else expr))))
	((zodiac:lambda-form? expr)
	  `(lambda ,(zodiac:improper-map zodiac:expr-stripper
		      (zodiac:lambda-form-args expr))
	     ,(zodiac:expr-stripper (zodiac:lambda-form-body expr))))
	((zodiac:set!-form? expr)
	  `(set! ,(zodiac:expr-stripper (zodiac:set!-form-var expr))
	     ,(zodiac:expr-stripper (zodiac:set!-form-val expr))))
	((zodiac:time-form? expr)
	  `(time ,(zodiac:expr-stripper (zodiac:time-form-expr expr))))
	((zodiac:quote-form? expr)
	  `(quote ,(zodiac:expr-stripper (zodiac:quote-form-expr expr))))
	((zodiac:begin-form? expr)
	  `(begin ,(zodiac:expr-stripper (zodiac:begin-form-first expr))
	     ,(zodiac:expr-stripper (zodiac:begin-form-rest expr))))
	((zodiac:let-form? expr)
	  `(let ,(zodiac:syntax-map (lambda (var val)
				      (list (zodiac:expr-stripper var)
					(zodiac:expr-stripper val)))
		   (zodiac:let-form-vars expr)
		   (zodiac:let-form-vals expr))
	     ,(zodiac:expr-stripper (zodiac:let-form-body expr))))
	((zodiac:letrec-form? expr)
	  `(letrec ,(zodiac:syntax-map (lambda (var val)
					 (list (zodiac:expr-stripper var)
					   (zodiac:expr-stripper val)))
		      (zodiac:letrec-form-vars expr)
		      (zodiac:letrec-form-vals expr))
	     ,(zodiac:expr-stripper (zodiac:letrec-form-body expr))))
	((zodiac:define-form? expr)
	  `(define ,(zodiac:expr-stripper (zodiac:define-form-name expr))
	     ,(zodiac:expr-stripper (zodiac:define-form-value expr))))
	((zodiac::-form? expr)
	  `(: ,(zodiac:expr-stripper (zodiac::-form-expr expr))
	     ,(zodiac::-form-type expr)))

	((zodiac:lexical-varref? expr)
	  (zodiac:expr-stripper (zodiac:lexical-varref-var expr)))
	((zodiac:top-level-varref? expr)
	  (zodiac:expr-stripper (zodiac:top-level-varref-var expr)))
	((zodiac:unbound-varref? expr)
	  (zodiac:expr-stripper (zodiac:unbound-varref-var expr)))

	((zodiac:app? expr)
	  `(,(zodiac:expr-stripper (zodiac:app-fun expr))
	     ,@(zodiac:syntax-map zodiac:expr-stripper (zodiac:app-args expr))))

	((zodiac:unchanged? expr)
          (zodiac:strip* (zodiac:unchanged-body expr)))

	((zodiac:bound? expr)
	  (zodiac:expr-stripper (zodiac:bound-var expr)))

	(else expr)))))

(define zodiac:load
  (let ((input-port? input-port?) (open-input-file open-input-file)
	 (current-eval current-eval) (list list))
    (lambda (filename/port file-key interactive? handler)
      (let ([reader (zodiac:read
		      (if (input-port? filename/port)
			filename/port
			(open-input-file filename/port))
		      (make-zodiac:location
			(zodiac:location-line zodiac:default-scan-offset)
			(zodiac:location-column zodiac:default-scan-offset)
			(zodiac:location-offset zodiac:default-scan-offset)
			file-key))]
	     [eva-lu-ator (current-eval)])
	(let read-loop
	  ([eval-values (list (zodiac:void))])
	  (let ((expr (reader)))
	    (if (zodiac:eof? expr)
	      (if interactive?
		(zodiac:report-result file-key eval-values)
		(zodiac:report-loaded file-key))
	      (zodiac:call-with-values
		(lambda ()
		  (let* ((expanded (zodiac:expand expr))
			  (parsed (zodiac:parse-expr
				    expr expanded zodiac:empty-env 0))
			  (annotated (handler parsed))
			  (stripped (zodiac:expr-stripper annotated)))
		    (eva-lu-ator stripped)))
		(lambda vals
		  (read-loop vals))))))))))
