; -*- Scheme -*-

(define zodiac:whitespace-list
  (list #\space #\newline #\tab))

(define zodiac:parens
  (list (cons #\( #\))
    (cons #\[ #\])))

(define zodiac:matched-parens?
  (let ((assq assq) (eq? eq?) (cdr cdr))
    (lambda (open close)
      (let ((entry (assq open zodiac:parens)))
	(and entry (eq? close (cdr entry)))))))

(define zodiac:delimiter-list
  (append zodiac:whitespace-list
    (map car zodiac:parens) (map cdr zodiac:parens)
    (list #\" #\;
      #\' #\` #\,))) ;; this line added to stuff in R4RS

; ----------------------------------------------------------------------

(define char-set-size 256)

(define make-char-table
  (let ((make-vector make-vector))
    (case-lambda
      (()
	(make-vector char-set-size))
      ((default)
	(make-vector char-set-size default)))))

(define set-entries
  (lambda (table entries handler)
    (map (lambda (c) (vector-set! table (char->integer c) handler)) entries)))

(define scan
  (let ((scan-table (make-char-table scan-symbol-or-number)))
    (set-entries scan-table zodiac:whitespace-list scan-whitespace)
    (set-entries scan-table (map car zodiac:parens) scan-open-paren)
    (set-entries scan-table (map cdr zodiac:parens) scan-close-paren)
    (set-entries scan-table (list #\#) scan-pound)
    (set-entries scan-table (list #\") scan-string)
    (set-entries scan-table (list #\.) scan-period)
    (set-entries scan-table (list #\` #\' #\,) scan-quote-family)
    (lambda ()


(define zodiac:string->char
  (let ([string->char-table '(("#\\newline" #\newline)
			      ("#\\nl"      #\newline)
                              ("#\\return"  #\return)
			      ("#\\space"   #\space)
			      ("#\\tab"     #\tab))])
    (let ((null? null?) (=? =?) (char=? char=?) (string-ref string-ref)
	   (string-ci=? string-ci=?) (string-length string-length)
	   (car car) (cadr cadr) (cdr cdr))
      (lambda (s)
	(let loop
	  ([table string->char-table])
	  (if (null? table)
	    (and (=? (string-length s) 3)
	      (char=? (string-ref s 0) #\#)
	      (char=? (string-ref s 1) #\\)
	      (string-ref s 2))
	    (if (string-ci=? s (car (car table)))
	      (cadr (car table))
	      (loop (cdr table)))))))))

(define zodiac:no-file-cookie 'no-file)

(define zodiac:default-scan-offset
  (make-zodiac:location 1 0 -1 zodiac:no-file-cookie))

(define zodiac:scan
    (lambda (port start-at)
	  (lambda (state paren-file text)

		  [(eof-object? c)
		    (if (in-state? 'string)
		      (zodiac:report-static-error
			"EOF within string"
			(start-location))
		      (return 'eof paren-file text))]

		  [(and (in-state? 'string)
		     (char-equal? c #\\))
		    ;; don't need to cons the #\\ on since
		    ;; list->string will take care of it for us;
		    ;; also note we let anything be quoted inside a string
		    (advance-input!)
		    (reader state paren-file (cons c text))]
		  [(and (in-state? 'string)
		     (char-equal? c #\"))
		    (return 'string paren-file text)]
		  [(in-state? 'string)
		    (reader state paren-file (cons c text))]

		  ;; list/vector
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c zodiac:open-parens))
		    (advance-input!)
		    (return 'vector
		      (cons (make-zodiac:open-paren
			      c
			      (start-location))
			paren-file)
		      text)]
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c zodiac:digit-list))
		    (reader 'vector-w/-size paren-file text)]
		  [(and (in-state? 'vector-w/-size)
		     (char-contained? c zodiac:digit-list))
		    (reader 'vector-w/-size paren-file (cons c text))]
		  [(and (in-state? 'vector-w/-size)
		     (char-contained? c zodiac:open-parens))
		    (return 'vector
		      (cons (make-zodiac:open-paren
			      c
			      (start-location))
			paren-file)
		      text)]
		  [(in-state? 'vector-w/-size)
		    (zodiac:report-static-error
		      "Non-digit character in vector size prefix"
		      (current-location))]
		  [(and (in-state? 'ether)
		     (char-contained? c zodiac:close-parens))
		    (if (null? paren-file)
		      (zodiac:report-static-error
			"No matching opening parenthesis"
			(start-location))
		      (let* ([to-match (car paren-file)]
			      [open (zodiac:open-paren-text to-match)])
			(if (zodiac:matched-parens? open c)
			  (return 'endseq (cdr paren-file) (cons c text))
			  (zodiac:report-static-error
			    "Mismatched parenthesis"
			    (start-location)))))]

		  ;; boolean
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (or (char-equal? next-c #\t)
		       (char-equal? next-c #\f)))
		    (advance-input!)
		    ;; remember: consing must be done in reverse order
		    (return 'boolean paren-file (cons c (cons #\# text)))]

		  ;; period
		  [(and (in-state? 'ether)
		     (char-equal? c #\.)
		     (char-contained? next-c zodiac:delimiter-list))
		    (return 'period paren-file (cons c text))]

		  ;; char/number/symbol
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-equal? next-c #\\))
		    (advance-input!)
		    (if (char-contained? next-c zodiac:alphabet-list)
		      (reader 'until-delim paren-file (cons c (cons #\# text)))
		      (begin
			(advance-input!)
			(return 'char paren-file
			  (cons c (cons #\\ (cons #\# text))))))]
		  [(in-state? 'ether)
		    (reader 'until-delim paren-file (cons c text))]
		  [(in-state? 'until-delim)
		    (if (char-contained? c zodiac:delimiter-list)
		      (let ([text-string (list->string (reverse text))])
			(cond
			  [(string->number text-string)
			    (return-unread 'number paren-file text)]
			  [(zodiac:string->char text-string)
			    (return-unread 'char paren-file text)]
			  [(and (>=? (string-length text-string) 2)
			     (char-equal? (string-ref text-string 0) #\#)
			     (char-equal? (string-ref text-string 1) #\\))
			    (zodiac:report-static-error
			      "Invalid character syntax"
			      (start-location))]
			  [(string->symbol text-string)
			    (return-unread 'symbol paren-file text)]
			  [else
			    (error 'zodiac:scan
			      "Should not get here ~s" text-string)]))
		      (reader 'until-delim paren-file (cons c text)))]

		  [else
		    (error 'zodiac:scan
		      "Should not get here ~s ~s" state c)])))))))))

