(define zodiac:raw-base-type?
  (let ((symbol? symbol?) (number? number?) (string? string?)
	 (char? char?) (boolean? boolean?))
    (lambda (object)
      (or (symbol? object)
	(number? object)
	(string? object)
	(char? object)
	(boolean? object)))))

(define zodiac:unannotated-type?
  (let ((pair? pair?))
    (lambda (object)
      (or (zodiac:raw-base-type? object)
	(pair? object)))))

(define (zodiac:base-type? object)
  (or (zodiac:number? object)
      (zodiac:symbol? object)
      (zodiac:string? object)
      (zodiac:char? object)
      (zodiac:boolean? object)))

(define (zodiac:aggregate-type? object)
  (or (zodiac:list? object)
      (zodiac:vector? object)
      (zodiac:improper-list? object)))

(define (zodiac:base-type-contents object)
  ((cond ((zodiac:number? object) zodiac:number-text)
	 ((zodiac:symbol? object) zodiac:symbol-text)
	 ((zodiac:string? object) zodiac:string-text)
	 ((zodiac:char? object) zodiac:char-text)
	 ((zodiac:boolean? object) zodiac:boolean-text))
   object))

(define zodiac:strip
  (let ((pair? pair?) (car car) (cadr cadr) (cdr cdr))
    (lambda (object)
      (cond
	((zodiac:base-type? object) (zodiac:base-type-contents object))
	((zodiac:list? object) (zodiac:list-contents object))
	((zodiac:vector? object) (zodiac:vector-contents object))
	((zodiac:improper-list? object)
	  (let loop
	    ((contents (zodiac:improper-list-contents object)))
	    (cond
	      ((pair? contents)
		(if (zodiac:period? (car contents))
		  (loop (cadr contents))
		  (cons (car contents) (loop (cdr contents)))))
	      (else contents))))
	(else object)))))

(define zodiac:strip* 
  (let ((pair? pair?) (car car) (cadr cadr) (cdr cdr) (null? null?)
	 (vector? vector?) (cons cons) (list->vector list->vector)
	 (vector->list vector->list))
    (lambda (object)
      (cond
	((zodiac:base-type? object) (zodiac:base-type-contents object))
	((zodiac:list? object) (zodiac:strip* (zodiac:list-contents object)))
	((zodiac:vector? object)
	  (zodiac:strip* (zodiac:vector-contents object)))
	((zodiac:improper-list? object)
	  (let loop
	    ((contents (zodiac:improper-list-contents object)))
	    (cond
	      ((pair? contents)
		(if (zodiac:period? (car contents))
		  (loop (cadr contents))
		  (cons (zodiac:strip* (car contents))
		    (loop (cdr contents)))))
	      (else (zodiac:strip* contents)))))
	((null? object) object)
	((pair? object) (cons (zodiac:strip* (car object))
			  (zodiac:strip* (cdr object))))
	((vector? object) (list->vector (map zodiac:strip*
					  (vector->list object))))
	((zodiac:raw-base-type? object) object)
	(else object)))))

(define (zodiac:location->list location)
  (list (zodiac:location-line location)
	(zodiac:location-column location)
	(zodiac:location-offset location)
	(zodiac:location-file location)))

(define zodiac:syntax-pair?
  (let ((pair? pair?) (>= >=) (> >))
    (lambda (object)
      (or (pair? object)
	(and (zodiac:list? object)
	  (>= (zodiac:list-length object) 1))
	(and (zodiac:improper-list? object)
	  (> (zodiac:improper-list-length object) 1))))))

(define zodiac:syntax-null?
  (let ((null? null?) (= =))
    (lambda (object)
      (or (null? object)
	(or (and (zodiac:list? object)
	      (= 0 (zodiac:list-length object))))))))

(define zodiac:syntax-list?
  (let ((list? list?))
    (lambda (object)
      (or (zodiac:list? object)
	(list? object)))))

(define zodiac:syntax-improper-list?
  (let ((pair? pair?) (list? list?) (not not))
    (lambda (object)
      (or (zodiac:improper-list? object)
	(and (pair? object) (not (list? object)))))))

(define zodiac:syntax-vector?
  (let ((vector? vector?))
    (lambda (object)
      (or (zodiac:vector? object)
	(vector? object)))))

(define zodiac:syntax-symbol?
  (let ((symbol? symbol?))
    (lambda (object)
      (or (symbol? object) (zodiac:symbol? object)))))

(define zodiac:syntax-number?
  (let ((number? number?))
    (lambda (object)
      (or (number? object) (zodiac:number? object)))))

(define zodiac:syntax-boolean?
  (let ((boolean? boolean?))
    (lambda (object)
      (or (boolean? object) (zodiac:boolean? object)))))

(define zodiac:syntax-string?
  (let ((string? string?))
    (lambda (object)
      (or (string? object) (zodiac:string? object)))))

(define zodiac:syntax-char?
  (let ((char? char?))
    (lambda (object)
      (or (char? object) (zodiac:char? object)))))

(define zodiac:syntax-length
  (let ((null? null?) (pair? pair?) (length length))
    (lambda (object)
      (cond
	((null? object) 0)
	((pair? object) (length object))
	((zodiac:list? object) (zodiac:list-length object))
	(else (error 'syntax-length "not a list ~s" object))))))

(define zodiac:endow-w/-locs
  (let ((symbol? symbol?) (null? null?) (char? char?) (boolean? boolean?)
	 (number? number?) (string? string?) (vector? vector?)
	 (list? list?) (list->vector list->vector) (vector->list vector->list)
	 (vector-length vector-length) (length length))
    (lambda (expr start finish)
      (cond
	((structure? expr) expr)
	((symbol? expr) (make-zodiac:symbol expr start finish))
	((null? expr) (make-zodiac:list expr 0 start finish))
	((char? expr) (make-zodiac:char expr start finish))
	((boolean? expr) (make-zodiac:boolean expr start finish))
	((number? expr) (make-zodiac:number expr start finish))
	((string? expr) (make-zodiac:string expr start finish))
	((vector? expr) (make-zodiac:vector (list->vector
					      (map (lambda (e)
						     (zodiac:endow-w/-locs
						       e start finish))
						(vector->list expr)))
			  (vector-length expr)
			  start finish))
	((list? expr) (make-zodiac:list (map (lambda (e)
					       (zodiac:endow-w/-locs
						 e start finish))
					  expr)
			(length expr)
			start finish))
	(else (error 'zodiac:endow-w/-locs "unknown object ~s" expr))))))

(define zodiac:endow
    (lambda (expr source)
      (let ((start (zodiac:determine-start-location source))
	     (finish (zodiac:determine-finish-location source)))
	(zodiac:endow-w/-locs expr start finish))))

(define zodiac:andmap
  (let ((null? null?) (error error) (map map) (apply apply) 
	 (car car) (cdr cdr))
    (lambda (predicate? . lists)
      (if (null? lists)
	(error 'zodiac:andmap "no lists given"))
      (let loop
	((lists (map (lambda (l)
		       (if (zodiac:list? l)
			 (zodiac:list-contents l)
			 l))
		  lists)))
	(if (null? (car lists))
	  #t
	  (and (apply predicate? (map car lists))
	    (loop (map cdr lists))))))))

(define zodiac:ormap
  (let ((null? null?) (car car) (cdr cdr))
    (lambda (predicate? expr)
      (if (null? expr)
	#f
	(or (predicate? (car expr))
	  (zodiac:ormap predicate? (cdr expr)))))))

(define zodiac:syntax-ormap
  (let ((apply apply))
    (lambda (predicate? . lists)
      (apply zodiac:ormap
	predicate?
	(map (lambda (l)
	       (if (zodiac:list? l)
		 (zodiac:list-contents l)
		 l))
	  lists)))))

(define zodiac:syntax-andmap
  (let ((apply apply))
    (lambda (predicate? . lists)
      (apply zodiac:andmap
	predicate?
	(map (lambda (l)
	       (if (zodiac:list? l)
		 (zodiac:list-contents l)
		 l))
	  lists)))))

(define zodiac:syntax-map
  (let ((null? null?) (cdr cdr) (car car) (pair? pair?) (map map)
	 (cons cons) (apply apply))
    (lambda (function . lists)
      (if (null? (cdr lists))
	(let ((list (car lists)))
	  (cond
	    ((null? list) list)
	    ((pair? list) (map function list))
	    ((zodiac:list? list)
	      (make-zodiac:list
		(map function (zodiac:list-contents list))
		(zodiac:list-length list)
		(zodiac:list-open list)
		(zodiac:list-close list)))
	    (else
	      (error 'zodiac:syntax-map "not a list ~s" list))))
	(let loop ((lists lists))
	  (cond
	    ((zodiac:syntax-null? (car lists))
	      (if (zodiac:andmap zodiac:syntax-null? (cdr lists))
		'()
		(error 'zodiac:syntax-map "not all empty ~s" lists)))
	    (else
	      (cons
		(apply function (map zodiac:syntax-car lists))
		(loop (map zodiac:syntax-cdr lists))))))))))

(define zodiac:improper-map
  (let ((cons #%cons))
    (lambda (function list)
      (cond
	((zodiac:syntax-null? list) '())
	((zodiac:syntax-pair? list)
	  (cons (function (zodiac:syntax-car list))
	    (zodiac:improper-map function (zodiac:syntax-cdr list))))
	(else
	  (function list))))))

(define zodiac:syntax-list->vector
  (let ((list->vector list->vector))
    (lambda (list)
      (cond
	((zodiac:list? list)
	  (list->vector (zodiac:list-contents list)))
	(else
	  (list->vector list))))))

(define zodiac:syntax-vector->list
  (let ((vector->list vector->list))
    (lambda (vector)
      (cond
	((zodiac:vector? vector)
	  (vector->list (zodiac:vector-contents vector)))
	(else
	  (vector->list vector))))))

(define zodiac:syntax-append
  (let ((null? null?) (car car) (cdr cdr))
    (lambda lists
      (let loop ((lists lists))
	(if (null? lists)
	  '()
	  (let ((first (car lists)))
	    (append (cond ((zodiac:list? first) (zodiac:list-contents first))
		      (else first))
	      (loop (cdr lists)))))))))

(define zodiac:syntax-car
  (let ((pair? pair?) (car car) (null? null?) (error error))
    (lambda (expr)
      (cond
	((pair? expr) (car expr))
	((zodiac:list? expr)
	  (let ((contents (zodiac:list-contents expr)))
	    (cond
	      ((null? contents) (error 'zodiac:syntax-car "non-pair ~s" expr))
	      (else (car contents)))))
	((zodiac:improper-list? expr)
	  (let ((contents (zodiac:improper-list-contents expr)))
	    (cond
	      ((pair? contents) (car contents))
	      (else (error 'zodiac:syntax-car "not a pair ~s" expr)))))
	(else (error 'zodiac:syntax-car "not a pair ~s" expr))))))

(define zodiac:syntax-cdr
  (let ((pair? pair?) (cdr cdr) (null? null?) 
	 (cadr cadr) (caddr caddr) (- -))
    (lambda (expr)
      (cond
	((pair? expr) (cdr expr))
	((zodiac:list? expr)
	  (let ((contents (zodiac:list-contents expr)))
	    (cond
	      ((null? contents) (error 'zodiac:syntax-cdr "non-pair ~s" expr))
	      (else (cdr contents)))))
	((zodiac:improper-list? expr)
	  (let ((contents (zodiac:improper-list-contents expr)))
	    (cond
	      ((pair? contents)
		(if (zodiac:period? (cadr contents))
		  (caddr contents)
		  (make-zodiac:improper-list (cdr contents)
		    (- (zodiac:improper-list-length expr)
		      1)
		    (zodiac:determine-start-location expr)
		    (zodiac:improper-list-close expr))))
	      (else (error 'zodiac:syntax-cdr "not a pair ~s" expr)))))))))

; CHECK: if unquote, let primitives outside lambda

'(define zodiac:syntax-cdr
  (lambda (expr)
    (cond
     ((pair? expr) (cdr expr))
     ((zodiac:list? expr)
      (let ((contents (zodiac:list-contents expr)))
	(cond
	 ((null? contents) (error 'zodiac:syntax-cdr "not a pair ~s" expr))
	 (else (cdr contents)))))
     ((zodiac:improper-list? expr)
      (let ((contents (zodiac:improper-list-contents expr)))
	(cond
	 ((pair? contents)
	  (cdr (let loop
		   ((contents contents))
		 (if (zodiac:period? (car contents))
		     (cadr contents)
		     (cons (car contents) (loop (cdr contents)))))))
	 (else (error 'zodiac:syntax-cdr "not a pair ~s" expr))))))))

; CHECK: if unquote, let primitives outside lambda

'(define zodiac:syntax-cdr
  (lambda (expr)
    (cond
     ((pair? expr) (cdr expr))
     ((zodiac:list? expr)
      (let ((contents (zodiac:list-contents expr))
	    (close (zodiac:list-close expr))
	    (length (zodiac:list-length expr)))
	(cond
	 ((null? contents) (error 'zodiac:syntax-cdr "not a pair ~s" expr))
	 (else
	  (let ((rest (cdr contents)))
	    (make-zodiac:list rest (- length 1)
			     (make-zodiac:phantom-paren
			      'open
			      (if (null? rest)
				  (zodiac:close-paren-location close)
				  (zodiac:determine-start-location rest)))
			     close))))))
     ((zodiac:improper-list? expr)
      (let ((contents (zodiac:improper-list-contents expr))
	    (close (zodiac:improper-list-close expr))
	    (length (zodiac:improper-list-length expr)))
	(cond
	 ((pair? contents)
	  (let ((rest (cdr contents)))
	    ;; CHECK: shouldn't this make an improper list?
	    (make-zodiac:list rest (- length 1)
			     (make-zodiac:phantom-paren
			      'open
			      (if (null? rest)
				  (zodiac:close-paren-location close)
				  (zodiac:determine-start-location rest)))
			     close)))
	 (else (error 'zodiac:syntax-cdr "not a pair ~s" expr)))))
     (else (error 'zodiac:syntax-cdr "not a pair ~s" expr)))))

;; CHECK: should also deal with improper-lists!

(define zodiac:syntax-cons
  (let ((cons cons) (+ +))
    (lambda (a d)
      (cond
	((and (or (zodiac:base-type? a) (zodiac:list? a)) (zodiac:list? d))
	  (make-zodiac:list
	    (cons a (zodiac:list-contents d))
	    (+ 1 (zodiac:list-length d))
	    (zodiac:determine-start-location a)
	    (zodiac:determine-finish-location d)))
	((and (zodiac:base-type? a) (zodiac:improper-list? d))
	  (make-zodiac:improper-list
	    (cons a (zodiac:improper-list-contents d))
	    (+ 1 (zodiac:improper-list-length d))
	    (zodiac:determine-start-location a)
	    (zodiac:determine-finish-location d)))
	(else (cons a d))))))

(define zodiac:same-name?
  (let ((eq? eq?))
    (lambda (s1 s2)
      (let ((real-s1 (if (zodiac:symbol? s1) (zodiac:symbol-text s1) s1))
	     (real-s2 (if (zodiac:symbol? s2) (zodiac:symbol-text s2) s2)))
	(eq? real-s1 real-s2)))))

(define zodiac:syntax-symbol-name
  (let ((symbol? symbol?) (error error))
    (lambda (s)
      (cond
	[(zodiac:symbol? s) (zodiac:symbol-text s)]
	[(symbol? s) s]
	[(zodiac:lexical-varref? s)
	  (zodiac:syntax-symbol-name (zodiac:lexical-varref-var s))]
	[(zodiac:top-level-varref? s)
	  (zodiac:syntax-symbol-name (zodiac:top-level-varref-var s))]
	[(zodiac:unbound-varref? s)
	  (zodiac:syntax-symbol-name (zodiac:unbound-varref-var s))]
	[else (error 'zodiac:syntax-symbol-name "invalid symbol ~s" s)]))))

(define zodiac:same-location?
  (let ((equal? equal?))
    (lambda (l1 l2)
      (equal? (zodiac:location->list l1) (zodiac:location->list l2)))))

; ----------------------------------------------------------------------

(define zodiac:union
  (let ((apply apply) (append append))
    (lambda sets
      (apply append sets))))

(define (zodiac:filter-vars-of expr lexical-k top-level-k unbound-k
			      quote-k accumulate)

  (define (filter expr)
    (cond
      ((zodiac:if-form? expr)
	(accumulate (filter (zodiac:if-form-test expr))
	  (filter (zodiac:if-form-then expr))
	  (filter (zodiac:if-form-else expr))))
      ((zodiac:set!-form? expr)
	(filter (zodiac:set!-form-val expr)))
      ((zodiac:quote-form? expr)
	(quote-k expr))
      ((zodiac:begin-form? expr)
	(accumulate (filter (zodiac:begin-form-first expr))
	  (filter (zodiac:begin-form-rest expr))))
      ((zodiac:define-form? expr)
	(filter (zodiac:define-form-value expr)))
      ((zodiac:time-form? expr)
	(filter (zodiac:time-form-expr expr)))
      ((zodiac:lambda-form? expr)
	(filter (zodiac:lambda-form-body expr)))
      ((zodiac:letrec-form? expr)
	(accumulate
	  (apply accumulate (map filter
			      (zodiac:letrec-form-vals expr)))
	  (filter (zodiac:letrec-form-body expr))))

      ((zodiac:app? expr)
	(accumulate (filter (zodiac:app-fun expr))
	  (apply accumulate
	    (map filter (zodiac:app-args expr)))))

      ((zodiac:lexical-varref? expr)
	(lexical-k expr))
      ((zodiac:top-level-varref? expr)
	(top-level-k expr))
      ((zodiac:unbound-varref? expr)
	(unbound-k expr))))

  (filter expr))

(define zodiac:unique-vars
  (let ((null? null?) (car car) (cdr cdr))
    (lambda (vars)
      (let loop
	((vars vars))
	(if (null? vars)
	  '()
	  (let ((first (car vars))
		 (rest (cdr vars)))
	    (if (present-in? first rest)
	      (loop rest)
	      (cons first (loop rest)))))))))

(define zodiac:top-level&unbound-vars-of
  (let ((list list))
    (lambda (expr)
      (zodiac:filter-vars-of expr
	(lambda (lexical-var) '())
	(lambda (top-level-var) (list top-level-var))
	(lambda (unbound-var) (list unbound-var))
	(lambda (quote-form) '())
	zodiac:union))))

(define zodiac:lexical-vars-of
  (let ((list list))
    (lambda (expr)
      (zodiac:filter-vars-of expr
	(lambda (lexical-var) (list lexical-var))
	(lambda (top-level-var) '())
	(lambda (unbound-var) '())
	(lambda (quote-form) '())
	zodiac:union))))
