; -*- Scheme -*-

(define zodiac:alphabet-list
  (list #\a #\b #\c #\d #\e #\f #\g #\h #\i #\j #\k #\l #\m
	#\n #\o #\p #\q #\r #\s #\t #\u #\v #\w #\x #\y #\z))

(define zodiac:digit-list
  (list #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9))

(define zodiac:identifier-initial-list
  (append zodiac:alphabet-list
	  (list #\! #\$ #\% #\& #\* #\/ #\: #\< #\= #\> #\? #\~ #\_ #\^)))

(define zodiac:identifier-subsequent-list
  (append zodiac:identifier-initial-list
	  zodiac:digit-list
	  (list #\. #\+ #\-)))

(define zodiac:whitespace-list
  (list #\space #\newline #\tab #\return))

(define zodiac:open-parens
  (list #\( #\[ #\{)) ;; we're being generous compared to R4RS

(define zodiac:close-parens
  (list #\) #\] #\})) ;; even more generosity

(define zodiac:delimiter-list
  (append zodiac:whitespace-list
	  zodiac:open-parens zodiac:close-parens
	  (list #\" #\;
		#\' #\` #\,))) ;; this line added to stuff in R4RS

(define zodiac:quote-prefix-list
  (list 'quote 'quasiquote 'unquote 'unquote-splicing))

(define zodiac:default-vector-value
  (let ([default-value 0])
    (let ((length length))
      (lambda args
	(case (length args)
	  [(0) default-value]
	  [(1) (set! default-value (car args))]
	  [else (error 'zodiac:default-vector-value "too many arguments")])))))

(define zodiac:void
  (let ([void-object (if #f #f)])
    (lambda ()
      void-object)))

(define zodiac:matched-parens?
  (let ((char=? char=?))
    (lambda (open close)
      (or (and (char=? #\( open) (char=? #\) close))
	(and (char=? #\[ open) (char=? #\] close))
	(and (char=? #\{ open) (char=? #\} close))))))

(define zodiac:string->char
  (let ([string->char-table '(("#\\newline" #\newline)
			      ("#\\nl"      #\newline)
                              ("#\\return"  #\return)
			      ("#\\space"   #\space)
			      ("#\\tab"     #\tab))])
    (let ((null? null?) (=? =?) (char=? char=?) (string-ref string-ref)
	   (string-ci=? string-ci=?) (string-length string-length)
	   (car car) (cadr cadr) (cdr cdr))
      (lambda (s)
	(let loop
	  ([table string->char-table])
	  (if (null? table)
	    (and (=? (string-length s) 3)
	      (char=? (string-ref s 0) #\#)
	      (char=? (string-ref s 1) #\\)
	      (string-ref s 2))
	    (if (string-ci=? s (car (car table)))
	      (cadr (car table))
	      (loop (cdr table)))))))))

(define zodiac:no-file-cookie 'no-file)

(define zodiac:default-scan-offset
  (make-zodiac:location 1 0 -1 zodiac:no-file-cookie))

(define zodiac:scan
  (let ((read-char read-char) (peek-char peek-char) (+ +)
	 (eof-object? eof-object?) (char=? char=?) (char-ci=? char-ci=?)
	 (eq? eq?) (string->list string->list) (null? null?) (car car)
	 (cdr cdr) (cons cons) (list->string list->string)
	 (string->number string->number) (reverse reverse)
	 (>=? >=?) (string->symbol string->symbol) (error error))
    (lambda (port start-at)
      (let ([current-line-number   (zodiac:location-line start-at)]
	     [current-column-number (zodiac:location-column start-at)]
	     [current-offset        (zodiac:location-offset start-at)]
	     [current-file          (zodiac:location-file start-at)]
	     [start-line-number     'error]
	     [start-column-number   'error]
	     [start-offset          'error]
	     [c                     'error]
	     [next-c                'error]
	     [first-time            #f])
	(letrec
	  ([advance-input!
	     (lambda ()
	       (set! c (read-char port))
	       (set! next-c (peek-char port))
	       (set! current-offset (+ 1 current-offset))
	       (if (not (eof-object? c))
		 (cond
		   [(char=? c #\newline)
		     (set! current-line-number (+ 1 current-line-number))
		     (set! current-column-number 0)]
		   [(char=? c #\tab)
		     (set! current-column-number
		       (+ current-column-number
			 (- 8 (remainder current-column-number 8))))]
		   [else
		     (set! current-column-number
		       (+ 1 current-column-number))])
		 'endif))]
	    [start-location
	      (lambda ()
		(make-zodiac:location start-line-number
		  start-column-number
		  start-offset
		  current-file))]
	    [current-location
	      (lambda ()
		(make-zodiac:location current-line-number
		  current-column-number
		  current-offset
		  current-file))]
	    [record-start-position!
	      (lambda ()
		(set! start-line-number current-line-number)
		(set! start-column-number current-column-number)
		(set! start-offset current-offset))]
	    [char-equal?
	      (lambda (c1 c2)
		(if (or (eof-object? c1) (eof-object? c2))
		  #f
		  (char-ci=? c1 c2)))]
	    [char-contained?
	      (lambda (c l)
		(zodiac:ormap (lambda (elt) (char-equal? c elt)) l))]
	    [skip-comment!
	      (lambda ()
		(let loop
		  ([in-comment #f])
		  (if in-comment
		    (begin
		      (advance-input!)
		      (if (or (eof-object? c) (char-equal? c #\newline))
			(zodiac:void)
			(loop #t)))
		    (if (char-equal? c #\;)
		      (loop #t)
		      (zodiac:void)))))]
	    [skip-atmosphere!
	      (lambda ()
		(skip-comment!)
		(if (char-contained? c zodiac:whitespace-list)
		  (begin
		    (advance-input!)
		    (skip-atmosphere!))
		  'endif))])
	  (advance-input!)
	  (lambda (state paren-file text)
	    (skip-atmosphere!)
	    (record-start-position!)
	    (set! first-time #t)
	    (let reader
	      ([state      state]
		[paren-file paren-file]
		[text       text])
	      (letrec
		([return-unread
		   (lambda (type paren-file text)
		     (zodiac:values
		       (make-zodiac:token
			 text type
			 (start-location)
			 (current-location))
		       paren-file))]
		  [return
		    (lambda (type paren-file text)
		      (advance-input!)
		      (return-unread type paren-file text))]
		  [in-state?
		    (lambda (check-state)
		      (eq? state check-state))]
		  [match-input
		    (lambda (input)
		      (let matcher
			([input (string->list input)])
			(if (null? input)
			  #t
			  (if (char-equal? c (car input))
			    (begin
			      (advance-input!)
			      (matcher (cdr input)))
			    #f))))])
		(if first-time
		  (set! first-time #f)
		  (advance-input!))
		(cond

		  ;; eof
		  [(and (eof-object? c)
		     (not (in-state? 'until-delim)))
		    (if (in-state? 'string)
		      (zodiac:report-static-error
			"EOF within string"
			(start-location))
		      (return 'eof paren-file text))]
		  ;; string
		  [(and (in-state? 'ether)
		     (char-equal? c #\"))
		    (reader 'string paren-file text)]
		  [(and (in-state? 'string)
		     (char-equal? c #\\))
		    ;; don't need to cons the #\\ on since
		    ;; list->string will take care of it for us;
		    ;; also note we let anything be quoted inside a string
		    (advance-input!)
		    (reader state paren-file (cons c text))]
		  [(and (in-state? 'string)
		     (char-equal? c #\"))
		    (return 'string paren-file text)]
		  [(in-state? 'string)
		    (reader state paren-file (cons c text))]

		  ;; list/vector
		  [(and (in-state? 'ether)
		     (char-contained? c zodiac:open-parens))
		    (return 'list
		      (cons (make-zodiac:open-paren
			      c
			      (start-location))
			paren-file)
		      text)]
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c zodiac:open-parens))
		    (advance-input!)
		    (return 'vector
		      (cons (make-zodiac:open-paren
			      c
			      (start-location))
			paren-file)
		      text)]
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-contained? next-c zodiac:digit-list))
		    (reader 'vector-w/-size paren-file text)]
		  [(and (in-state? 'vector-w/-size)
		     (char-contained? c zodiac:digit-list))
		    (reader 'vector-w/-size paren-file (cons c text))]
		  [(and (in-state? 'vector-w/-size)
		     (char-contained? c zodiac:open-parens))
		    (return 'vector
		      (cons (make-zodiac:open-paren
			      c
			      (start-location))
			paren-file)
		      text)]
		  [(in-state? 'vector-w/-size)
		    (zodiac:report-static-error
		      "Non-digit character in vector size prefix"
		      (current-location))]
		  [(and (in-state? 'ether)
		     (char-contained? c zodiac:close-parens))
		    (if (null? paren-file)
		      (zodiac:report-static-error
			"No matching opening parenthesis"
			(start-location))
		      (let* ([to-match (car paren-file)]
			      [open (zodiac:open-paren-text to-match)])
			(if (zodiac:matched-parens? open c)
			  (return 'endseq (cdr paren-file) (cons c text))
			  (zodiac:report-static-error
			    "Mismatched parenthesis"
			    (start-location)))))]

		  ;; boolean
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (or (char-equal? next-c #\t)
		       (char-equal? next-c #\f)))
		    (advance-input!)
		    ;; remember: consing must be done in reverse order
		    (return 'boolean paren-file (cons c (cons #\# text)))]

		  ;; period
		  [(and (in-state? 'ether)
		     (char-equal? c #\.)
		     (char-contained? next-c zodiac:delimiter-list))
		    (return 'period paren-file (cons c text))]

		  ;; quote/quasiquote/unquote/unquote-splicing
		  [(and (in-state? 'ether)
		     (or (char-equal? c #\')
		       (char-equal? c #\`)
		       (char-equal? c #\,)))
		    (case c
		      [(#\') (return 'quote paren-file (cons c text))]
		      [(#\`) (return 'quasiquote paren-file (cons c text))]
		      [(#\,) (return
			       (if (char-equal? next-c #\@)
				 (begin
				   (advance-input!)
				   'unquote-splicing)
				 'unquote)
			       paren-file
			       (cons c text))])]

		  ;; char/number/symbol
		  [(and (in-state? 'ether)
		     (char-equal? c #\#)
		     (char-equal? next-c #\\))
		    (advance-input!)
		    (if (char-contained? next-c zodiac:alphabet-list)
		      (reader 'until-delim paren-file (cons c (cons #\# text)))
		      (begin
			(advance-input!)
			(return 'char paren-file
			  (cons c (cons #\\ (cons #\# text))))))]
		  [(in-state? 'ether)
		    (reader 'until-delim paren-file (cons c text))]
		  [(in-state? 'until-delim)
		    (if (or (char-contained? c zodiac:delimiter-list)
			  (eof-object? c))
		      (let ([text-string (list->string (reverse text))])
			(cond
			  [(string->number text-string)
			    (return-unread 'number paren-file text)]
			  [(zodiac:string->char text-string)
			    (return-unread 'char paren-file text)]
			  [(and (>=? (string-length text-string) 2)
			     (char-equal? (string-ref text-string 0) #\#)
			     (char-equal? (string-ref text-string 1) #\\))
			    (zodiac:report-static-error
			      "Invalid character syntax"
			      (start-location))]
			  [(string->symbol text-string)
			    (return-unread 'symbol paren-file text)]
			  [else
			    (error 'zodiac:scan
			      "Should not get here ~s" text-string)]))
		      (reader 'until-delim paren-file (cons c text)))]

		  [else
		    (error 'zodiac:scan
		      "Should not get here ~s ~s" state c)])))))))))

(define zodiac:read
  (let ((null? null?) (car car) (string->number string->number)
	 (list->string list->string) (reverse reverse) (list-ref list-ref)
	 (+ +) (vector-set! vector-set!) (cdr cdr) (apply apply)
	 (list list) (error error) (eq? eq?)
	 (current-input-port current-input-port))
  (lambda port+offset
    (let ((getter #f))

      (lambda ()

	(define (read-one paren-file sequence-type)
	  (zodiac:call-with-values
	    (lambda () (getter 'ether paren-file '()))
	    (lambda (token paren-file)
	      (case (zodiac:token-type token)
		((eof)
		  (if (null? paren-file)
		    (zodiac:values (make-zodiac:eof (zodiac:token-start token))
		      paren-file)
		    (zodiac:report-static-error
		      "Unclosed parenthesis" (car paren-file))))
		((list)
		  (zodiac:call-with-values
		    (lambda () (read-sequence paren-file 'list 0))
		    (lambda (sequence last-paren new-paren-file length proper?)
		      (zodiac:values
			((if proper? make-zodiac:list make-zodiac:improper-list)
			  sequence length (car paren-file) last-paren)
			new-paren-file))))
		((vector)
		  (zodiac:call-with-values
		    (lambda () (read-sequence paren-file 'vector 0))
		    (lambda (sequence last-paren new-paren-file length last-item)
		      (let ((length-specified
			      (string->number
				(list->string
				  (reverse (zodiac:token-text token))))))
			(cond
			  (length-specified
			    (if (> length length-specified)
			      (zodiac:report-static-error
				"Vector longer than specified length"
				(list-ref sequence length-specified)))
			    (let ((v (make-vector length-specified last-item)))
			      (zodiac:values
				(make-zodiac:vector
				  (let loop
				    ((index 0) (sequence sequence))
				    (if (null? sequence)
				      v
				      (begin (vector-set! v index (car sequence))
					(loop (+ 1 index) (cdr sequence)))))
				  length-specified (car paren-file) last-paren)
				new-paren-file)))
			  (else
			    (zodiac:values
			      (make-zodiac:vector
				(apply vector sequence)
				length (car paren-file) last-paren)
			      new-paren-file)))))))
		((endseq)
		  (if sequence-type
		    (zodiac:values
		      (make-zodiac:close-paren (car (zodiac:token-text token))
			(zodiac:token-start token))
		      paren-file)
		    (zodiac:report-static-error "Extra parenthesis" token)))
		((period)
		  (case sequence-type
		    ((list)
		      (zodiac:values
			(make-zodiac:period (zodiac:token-start token))
			paren-file))
		    ((vector) (zodiac:report-static-error ". within vector"
				token))
		    (else (zodiac:report-static-error ". outside list" token))))
		((string boolean char symbol number)
		  (zodiac:call-with-values
		    (lambda ()
		      (let ((text
			      (list->string (reverse (zodiac:token-text token)))))
			(case (zodiac:token-type token)
			  ((string)
			    (zodiac:values make-zodiac:string text))
			  ((boolean)
			    (zodiac:values make-zodiac:boolean
			      (string-ci=? text "#t")))
			  ((char)
			    (zodiac:values make-zodiac:char
			      (zodiac:string->char text)))
			  ((symbol)
			    (zodiac:values make-zodiac:symbol
			      (string->symbol text)))
			  ((number)
			    (zodiac:values make-zodiac:number
			      (string->number text))))))
		    (lambda (maker text)
		      (zodiac:values
			(maker text
			  (zodiac:token-start token)
			  (zodiac:token-finish token))
			paren-file))))
		((quote quasiquote unquote unquote-splicing)
		  (zodiac:call-with-values
		    (lambda () (read-one paren-file sequence-type))
		    (lambda (result paren-file)
		      (zodiac:values
			(make-zodiac:list
			  (list (make-zodiac:symbol (zodiac:token-type token)
				  (zodiac:token-start token)
				  (zodiac:token-finish token))
			    result)
			  2
			  (make-zodiac:phantom-paren
			    'open
			    (zodiac:determine-start-location token))
			  (make-zodiac:phantom-paren
			    'close
			    (zodiac:determine-finish-location result)))
			paren-file))))
		(else (error 'read-one "Shouldn't get here ~s" token))))))

	(define (read-sequence paren-file sequence-type length)
	  (let loop
	    ((paren-file paren-file)
	      (length length)
	      (accum '()))
	    (zodiac:call-with-values
	      (lambda () (read-one paren-file sequence-type))
	      (lambda (object paren-file)
		(cond
		  ((zodiac:close-paren? object)
		    (zodiac:values (reverse accum) object
		      paren-file length
		      (if (eq? sequence-type 'vector)
			(if (null? accum)
			  (zodiac:default-vector-value)
			  (car accum))
			#t)))
		  ((or (zodiac:base-type? object)
		     (zodiac:aggregate-type? object))
		    (loop paren-file (+ 1 length) (cons object accum)))
		  ((zodiac:eof? object)
		    (zodiac:report-static-error "Unclosed parenthesis"
		      (car paren-file)))
		  ((zodiac:period? object)
		    (if (not (eq? sequence-type 'list))
		      (zodiac:report-static-error ".'s only in lists" object))
		    (if (null? accum)
		      (zodiac:report-static-error "No datum before ." object))
		    (zodiac:call-with-values
		      (lambda () (read-one paren-file sequence-type))
		      (lambda (dot+1 paren-file)
			(cond
			  ((or (zodiac:base-type? dot+1)
			     (zodiac:aggregate-type? dot+1))
			    (zodiac:call-with-values
			      (lambda () (read-one paren-file sequence-type))
			      (lambda (dot+2 paren-file)
				(cond
				  ((zodiac:close-paren? dot+2)
				    (zodiac:values
				      (reverse (cons dot+1 (cons object accum)))
				      dot+2 paren-file (+ 1 length) #f))
				  (else
				    (zodiac:report-static-error
				      "Not a closing parenthesis"
				      dot+2))))))
			  (else
			    (zodiac:report-static-error "Invalid object after ."
			      dot+1))))))
		  (else
		    (error 'read-sequence "Shouldn't get here ~s" object)))))))

	(if (not getter)
	  (set! getter (zodiac:scan (if (null? port+offset)
				      (current-input-port)
				      (car port+offset))
			 (if (or (null? port+offset)
			       (null? (cdr port+offset)))
			   zodiac:default-scan-offset
			   (cadr port+offset)))))
	(zodiac:call-with-values
	  (lambda () (read-one '() #f))
	  (lambda (expr parens)
	    (if (null? parens)
	      expr
	      (error 'zodiac:read "Shouldn't get here ~s" parens)))))))))

(define zodiac:determine-start-location
  (let ((pair? pair?) (car car) (error error))
    (lambda (object)
      (let ((start-location-extractor
	      (cond
		((zodiac:location? object) (lambda (x) x))
		((zodiac:token? object) zodiac:token-start)
		((zodiac:open-paren? object) zodiac:open-paren-location)
		((zodiac:close-paren? object) zodiac:close-paren-location)
		((zodiac:phantom-paren? object) zodiac:phantom-paren-location)
		((zodiac:period? object) zodiac:period-location)
		((zodiac:eof? object) zodiac:eof-location)
		((zodiac:string? object) zodiac:string-start)
		((zodiac:boolean? object) zodiac:boolean-start)
		((zodiac:char? object) zodiac:char-start)
		((zodiac:symbol? object) zodiac:symbol-start)
		((zodiac:number? object) zodiac:number-start)
		((zodiac:list? object) zodiac:list-open)
		((zodiac:improper-list? object) zodiac:improper-list-open)
		((zodiac:vector? object) zodiac:vector-open)
		((zodiac:if-form? object) zodiac:if-form-open)
		((zodiac:lambda-form? object) zodiac:lambda-form-open)
		((zodiac:set!-form? object) zodiac:set!-form-open)
		((zodiac:quote-form? object) zodiac:quote-form-open)
		((zodiac:time-form? object) zodiac:time-form-open)
		((zodiac:begin-form? object) zodiac:begin-form-open)
		((zodiac:let-form? object) zodiac:let-form-open)
		((zodiac:letrec-form? object) zodiac:letrec-form-open)
		((zodiac:define-form? object) zodiac:define-form-open)
		((zodiac:delay-form? object) zodiac:delay-form-open)
		((zodiac::-form? object) zodiac::-form-open)
		((zodiac:app? object) zodiac:app-open)
		((zodiac:unchanged? object) zodiac:unchanged-open)
		((zodiac:lexical-varref? object) zodiac:lexical-varref-var)
		((zodiac:top-level-varref? object) zodiac:top-level-varref-var)
		((zodiac:unbound-varref? object) zodiac:unbound-varref-var)
		((zodiac:bound? object) zodiac:bound-var)
		((pair? object) car)
		(else (error 'zodiac
			"syntax error at ~s" object)))))
	(let ((location (start-location-extractor object)))
	  (if (zodiac:location? location)
	    location
	    (zodiac:determine-start-location location)))))))

(define zodiac:determine-finish-location
  (let ((pair? pair?) (null? null?) (car car) (cdr cdr) (error error))
    (lambda (object)
      (let ((finish-location-extractor
	      (cond
		((zodiac:location? object) (lambda (x) x))
		((zodiac:token? object) zodiac:token-finish)
		((zodiac:open-paren? object) zodiac:open-paren-location)
		((zodiac:close-paren? object) zodiac:close-paren-location)
		((zodiac:phantom-paren? object) zodiac:phantom-paren-location)
		((zodiac:period? object) zodiac:period-location)
		((zodiac:eof? object) zodiac:eof-location)
		((zodiac:string? object) zodiac:string-finish)
		((zodiac:boolean? object) zodiac:boolean-finish)
		((zodiac:char? object) zodiac:char-finish)
		((zodiac:symbol? object) zodiac:symbol-finish)
		((zodiac:number? object) zodiac:number-finish)
		((zodiac:list? object) zodiac:list-close)
		((zodiac:improper-list? object) zodiac:improper-list-close)
		((zodiac:vector? object) zodiac:vector-close)
		((zodiac:if-form? object) zodiac:if-form-close)
		((zodiac:lambda-form? object) zodiac:lambda-form-close)
		((zodiac:set!-form? object) zodiac:set!-form-close)
		((zodiac:quote-form? object) zodiac:quote-form-close)
		((zodiac:time-form? object) zodiac:time-form-close)
		((zodiac:begin-form? object) zodiac:begin-form-close)
		((zodiac:let-form? object) zodiac:let-form-close)
		((zodiac:letrec-form? object) zodiac:letrec-form-close)
		((zodiac:define-form? object) zodiac:define-form-close)
		((zodiac:delay-form? object) zodiac:delay-form-close)
		((zodiac::-form? object) zodiac::-form-close)
		((zodiac:app? object) zodiac:app-close)
		((zodiac:unchanged? object) zodiac:unchanged-close)
		((zodiac:lexical-varref? object) zodiac:lexical-varref-var)
		((zodiac:top-level-varref? object) zodiac:top-level-varref-var)
		((zodiac:unbound-varref? object) zodiac:unbound-varref-var)
		((zodiac:bound? object) zodiac:bound-var)
		((pair? object) (if (null? (cdr object)) car cdr))
		(else (error 'zodiac
			"syntax error at ~s" object)))))
	(let ((location (finish-location-extractor object)))
	  (if (zodiac:location? location)
	    location
	    (zodiac:determine-finish-location location)))))))

(if zodiac:print-on-startup
  (printf "Scanner and reader loaded~n"))
